//////////////////////////////////////////////////////////////////////////
//	"getopt.h"
//------------------------------------------------------------------------
//	$Id: getopt.h 4 2006-03-01 22:26:13Z taro $
//////////////////////////////////////////////////////////////////////////
#ifndef	GETOPT_H_INCLUDED
#define	GETOPT_H_INCLUDED

#include	<stdlib.h>
#if defined(WIN32)
# include	<tchar.h>
#endif
#include	<string.h>
#include	<list>

#if	defined(LINUX)
# ifndef	BOOL
#  define	BOOL	int
# endif
# ifndef	TRUE
#  define	TRUE	1
# endif
# ifndef	FALSE
#  define	FALSE	0
# endif
# ifndef	TCHAR
#   define	TCHAR	char
# endif
# ifndef	_TCHAR
#   define	_TCHAR	char
# endif
# ifndef	TEXT
#   define	TEXT(s)	s
# endif
# ifndef	_TEXT
#   define	_TEXT(s)	s
# endif
#endif

class	GetOpt
{
public:
	enum	VarType
	{
		VAR_UNKNOWN = -1,
		VAR_NOTHING = 0,
		VAR_FLAG   ,
		VAR_TEXT   ,
		VAR_INT    ,
		VAR_INT64  ,
		VAR_FUNC   ,
	};

#if defined(WIN32)
	typedef __int64	Int64;
#elif defined(LINUX)
	typedef long long	Int64;
#endif

	struct	VarVal
	{
		VarType				vartype;
		int					selno;
		union
		{
			bool			flag;
			const _TCHAR	*text;
			int				n;
			Int64			n64;
		};
	};
	struct	VarPtr
	{
		VarType				vartype;
		union
		{
			bool			*flag;
			const _TCHAR	**text;
			int				*n;
			Int64			*n64;
		};
	};
	enum	SelType
	{
		SEL_OFF   ,
		SEL_MINMAX,
		SEL_TABLE  ,
	};
	struct Select
	{
		SelType	seltype;
		int		tblnum;
		VarVal	*tbl;
	};
	struct Item
	{
		int				usertype;
		const _TCHAR	*name;		// long switch name
		unsigned		name_len;
		const _TCHAR	*sw;		// short switch charactors
		VarType			vartype;
		VarVal			std;		// default
		Select			sel;
		VarVal			var;
		VarPtr			ptr;
	};
	struct AddItemParam
	{
		int					usertype;
		const _TCHAR		*name;
		const _TCHAR		*sw;
		VarType				vartype;
		bool				omit;
		struct std_tag
		{
			const _TCHAR	*text;
			int				n;
			Int64			n64;
		};
		std_tag				std;
	};

	typedef int (* GetOpt::GetVarFunc)(Item *, const _TCHAR *);
private:
	struct FuncItem
	{
		int			vartype;
		GetVarFunc	func;
	};
	std::list<FuncItem *>			funcList;

public:
	typedef std::list<Item  *>		List;
	typedef List::iterator			iterator;


private:
	List							list;
public:
	void		clear(void) { iterator it; for ( it = this->begin(); it != this->end(); it++ ) { delete *it; } this->list.clear(); }
	void		push_back(Item * item) { this->list.push_back(item); }
	iterator	begin(void) { return this->list.begin(); }
	iterator	end(void) { return this->list.end(); }
	int			size(void) const { return (int) this->list.size(); }

private:
	int			last_pos;
	int			cur_pos;
	int			next_pos;
	int			swpos;
public:
	const _TCHAR	*last_param;
	const _TCHAR	*err_option;
	const _TCHAR	*err_param;

public:
	GetOpt();
	virtual ~GetOpt();

	void			SetCurPos(int pos) { this->cur_pos = pos; }
	int				GetCurPos(void) const { return this->cur_pos; }
	int				GetLastPos(void) const { return this->last_pos; }
	int				AddFunc(GetVarFunc func);
	GetVarFunc		GetFunc(int vartype);
	Item *			AddItem(int usertype, const TCHAR * name, const _TCHAR * sw, VarType vartype);
	int				AddItem(unsigned n, GetOpt::AddItemParam * tbl);
	bool			IsSwitch(const TCHAR *s);
	const _TCHAR *	GetCurArg(int argc, const _TCHAR * argv[]);
	Item *			GetItem(int usertype, const _TCHAR * name);
	int				GetVar(GetOpt::Item * item, const _TCHAR * s);
	Item *			ParseArg(int & pos, int argc, const _TCHAR * argv[]);
	Item *			ParseSwitch(int & pos, int argc, const _TCHAR * argv[]);
	static Int64	GetOpt::GetInt64(const _TCHAR * s);

public:
	struct HelpItem
	{
		Item				*item;
		int					usertype;
		const _TCHAR		*param;
		const _TCHAR		*text;
	};
	enum	ColTbl
	{
		COL_1,
		COL_2,
		COL_3,
		COL_4,
		COL_MAX
	};
	int					coltbl[COL_MAX];

	void				Help(unsigned tblnum, const GetOpt::HelpItem * tbl);
};

#endif
