//////////////////////////////////////////////////////////////////////////
//	"pglib.h"
//------------------------------------------------------------------------
//	$Id: pglib.h 28 2006-03-04 17:23:05Z taro $
//////////////////////////////////////////////////////////////////////////
#ifndef	PGLIB_H_INCLUDED
#define	PGLIB_H_INCLUDED

#include	<libpq-fe.h>
#define	libpq

class	PgLib
{
public:
	enum PgLib_enum
	{
		PGTYPE_HASH_MAX = 16,
	};

	typedef	libpq::Oid				Oid;
	enum ConnStatus
	{
		CONNECTION_OK                = libpq::CONNECTION_OK     ,
		CONNECTION_BAD               = libpq::CONNECTION_BAD    ,
		CONNECTION_STARTED           = libpq::CONNECTION_STARTED,
		CONNECTION_MADE              = libpq::CONNECTION_MADE   ,
		CONNECTION_AWAITING_RESPONSE = libpq::CONNECTION_AWAITING_RESPONSE,
		CONNECTION_AUTH_OK           = libpq::CONNECTION_AUTH_OK    ,
		CONNECTION_SETENV            = libpq::CONNECTION_SETENV     ,
		CONNECTION_SSL_STARTUP       = libpq::CONNECTION_SSL_STARTUP,
		CONNECTION_NEEDED            = libpq::CONNECTION_NEEDED     ,
	};

	enum  ResultStatus
	{
		PGRES_EMPTY_QUERY    = libpq::PGRES_EMPTY_QUERY   ,
		PGRES_COMMAND_OK     = libpq::PGRES_COMMAND_OK    ,
		PGRES_TUPLES_OK      = libpq::PGRES_TUPLES_OK     ,
		PGRES_COPY_OUT       = libpq::PGRES_COPY_OUT      ,
		PGRES_COPY_IN        = libpq::PGRES_COPY_IN       ,
		PGRES_BAD_RESPONSE   = libpq::PGRES_BAD_RESPONSE  ,
		PGRES_NONFATAL_ERROR = libpq::PGRES_NONFATAL_ERROR,
		PGRES_FATAL_ERROR    = libpq::PGRES_FATAL_ERROR   ,
	};
	enum PgType
	{
		PGLIB_NONE        ,
		PGLIB_BOOL        ,
		PGLIB_BYTEA       ,
		PGLIB_CHAR        ,
		PGLIB_NAME        ,
		PGLIB_INT8        ,
		PGLIB_INT2        ,
		PGLIB_INT2VECTOR  ,
		PGLIB_INT4        ,
		PGLIB_REGPROC     ,
		PGLIB_TEXT        ,
		PGLIB_OID         ,
		PGLIB_TID         ,
		PGLIB_XID         ,
		PGLIB_CID         ,
		PGLIB_OIDVECTOR   ,
		PGLIB_POINT       ,
		PGLIB_LSEG        ,
		PGLIB_PATH        ,
		PGLIB_BOX         ,
		PGLIB_POLYGON     ,
		PGLIB_LINE        ,
		PGLIB_FLOAT4      ,
		PGLIB_FLOAT8      ,
		PGLIB_ABSTIME     ,
		PGLIB_RELTIME     ,
		PGLIB_TINTERVAL   ,
		PGLIB_UNKNOWN     ,
		PGLIB_CIRCLE      ,
		PGLIB_CASH        ,
		PGLIB_MACADDR     ,
		PGLIB_INET        ,
		PGLIB_CIDR        ,
		PGLIB_ACLITEM     ,
		PGLIB_BPCHAR      ,
		PGLIB_VARCHAR     ,
		PGLIB_DATE        ,
		PGLIB_TIME        ,
		PGLIB_TIMESTAMP   ,
		PGLIB_TIMESTAMPTZ ,
		PGLIB_INTERVAL    ,
		PGLIB_TIMETZ      ,
		PGLIB_BIT         ,
		PGLIB_VARBIT      ,
		PGLIB_NUMERIC     ,
		PGLIB_REFCURSOR   ,
		PGLIB_REGPROCEDURE,
		PGLIB_REGOPER     ,
		PGLIB_REGOPERATOR ,
		PGLIB_REGCLASS    ,
		PGLIB_REGTYPE     ,
		PGLIB_RECORD      ,
		PGLIB_CSTRING     ,
		PGLIB_ANY         ,
		PGLIB_ANYARRAY    ,
		PGLIB_VOID        ,
		PGLIB_TRIGGER     ,
		PGLIB_LANGUAGE_HANDLER,
		PGLIB_INTERNAL    ,
		PGLIB_OPAQUE      ,
		PGLIB_ANYELEMENT  ,
	};
	struct PgTypeItem
	{
		Oid				oid;
		PgType			type;
		const char *	name;
	};

	class	Result
	{
	friend class	PgLib;
	private:
		class srep
		{
		public:
			libpq::PGresult	*	pg_result;
			int					ref;
			srep ()
			{
				this->pg_result=  NULL;
				this->ref = 1;
			};
		};
		srep *				p;
	private:
		libpq::PGresult *	GetResult(void) const { return this->p != NULL ? this->p->pg_result : NULL; }

	public:
		Result();
		virtual ~Result();

		void	Clear(void);

		Result &		Result::operator = (Result & res);
		Result &		Result::operator = (libpq::PGresult * pgres);

		friend bool		operator == (const Result & res1, const Result & res2);
		friend bool		operator != (const Result & res1, const Result & res2) { return !(res1 == res2); }
		friend bool		operator == (const Result & res1, const libpq::PGresult * pgres2);

		ResultStatus	GetStatus(void) { return (ResultStatus) (this->p!=NULL && this->p->pg_result!=NULL ? libpq::PQresultStatus(this->p->pg_result) : libpq::PGRES_FATAL_ERROR); }

		int		GetTuples(void)   { return this->p!=NULL && this->p->pg_result!=NULL ? libpq::PQntuples(this->p->pg_result) : -1; }
		int		GetFields(void)   { return this->p!=NULL && this->p->pg_result!=NULL ? libpq::PQnfields(this->p->pg_result) : -1; }
		char *	GetFname(int idx) { return this->p!=NULL && this->p->pg_result!=NULL ? libpq::PQfname(  this->p->pg_result,idx) : NULL; }
		int		GetFnumber(const char * fname)
		                          { return this->p!=NULL && this->p->pg_result!=NULL ? libpq::PQfnumber(this->p->pg_result,fname) : -1; }
		Oid		GetFtype(int idx) { return this->p!=NULL && this->p->pg_result!=NULL ? libpq::PQftype(  this->p->pg_result,idx) : (Oid)-1; }
		int		GetFmod(int idx)  { return this->p!=NULL && this->p->pg_result!=NULL ? libpq::PQfmod(   this->p->pg_result,idx) : -1; }
		int		GetFsize(int idx) { return this->p!=NULL && this->p->pg_result!=NULL ? libpq::PQfsize(  this->p->pg_result,idx) : -1; }
		int		GetBinaryTuples(void)
		                          { return this->p!=NULL && this->p->pg_result!=NULL ? libpq::PQbinaryTuples(this->p->pg_result) : -1; }

		char *	GetValue(int y, int x)  { return this->p!=NULL && this->p->pg_result!=NULL ? libpq::PQgetvalue( this->p->pg_result,y,x) : NULL; }
		int		GetIsNull(int y, int x) { return this->p!=NULL && this->p->pg_result!=NULL ? libpq::PQgetisnull(this->p->pg_result,y,x) : -1; }
		int		GetLength(int y, int x) { return this->p!=NULL && this->p->pg_result!=NULL ? libpq::PQgetlength(this->p->pg_result,y,x) : -1; }

		char *	GetCmdStatus(void) { return this->p!=NULL && this->p->pg_result!=NULL ? libpq::PQcmdStatus(this->p->pg_result) : NULL; }
		char *	GetCmdTuples(void) { return this->p!=NULL && this->p->pg_result!=NULL ? libpq::PQcmdTuples(this->p->pg_result) : NULL; }
		Oid		GetOidValue(void)  { return this->p!=NULL && this->p->pg_result!=NULL ? libpq::PQoidValue( this->p->pg_result) : (Oid)-1; }
	};
	friend class		Result;

private:
	libpq::PGconn *		conn;

	class	ResultItem : public	Result
	{
	friend class	PgLib;
	private:
		ResultItem			*next, *prev;
		PgLib *				pglib;
	public:
		ResultItem();
		virtual ~ResultItem();
	};
	friend class		ResultItem;
	typedef ResultItem *	PresultPtr;

	ResultItem			*lastResult;
	unsigned			numResult;

	void			attach_result(PgLib::ResultItem * result);
	void			detach_result(PgLib::ResultItem * result);
	ResultItem *	create_result(libpq::PGresult * pg_result);
	int				destroy_result(PgLib::ResultItem * result);
	void			clear_result(void);

	struct PgTypeHash : public PgTypeItem
	{
		PgTypeHash	*next, *prev;
	};
	struct pgtype_t
	{
		PgTypeHash *	tbl[PGTYPE_HASH_MAX];
		unsigned		num;
	} pgtype;
	int			init_pgtype(void);
	void		clear_pgtype(void);
	int			add_pgtype(PgLib::PgTypeItem & pgtype);
	static unsigned	pgtype_hash(PgLib::Oid oid);

	int			work_level;
	bool		work_rollback;

public:
	PgLib();
	~PgLib();

	int			Connectdb(const char * info);
	int			Finish(void);

	ConnStatus	GetStatus(void)       { if (this->conn==NULL) return CONNECTION_BAD; return (ConnStatus)libpq::PQstatus(this->conn); }
	char *		GetDb(void)           { if (this->conn==NULL) return NULL; return libpq::PQdb(this->conn); }
	char *		GetUser(void)         { if (this->conn==NULL) return NULL; return libpq::PQuser(this->conn); }
	char *		GetPass(void)         { if (this->conn==NULL) return NULL; return libpq::PQpass(this->conn); }
	char *		GetHost(void)         { if (this->conn==NULL) return NULL; return libpq::PQhost(this->conn); }
	char *		GetPort(void)         { if (this->conn==NULL) return NULL; return libpq::PQport(this->conn); }
	char *		GetTty(void)          { if (this->conn==NULL) return NULL; return libpq::PQtty(this->conn); }
	char *		GetOptions(void)      { if (this->conn==NULL) return NULL; return libpq::PQoptions(this->conn); }
	char *		GetErrorMessage(void) { if (this->conn==NULL) return NULL; return libpq::PQerrorMessage(this->conn); }
	int			GetbackendPID(void)   { if (this->conn==NULL) return -1; return libpq::PQbackendPID(this->conn); }

	int			SetClientEncoding(const char *enc) { if (this->conn==NULL) return -1; return libpq::PQsetClientEncoding(this->conn, enc); }

	int			Exec(PgLib::Result & result, const char * query);
	Result  *	Exec(const char * query);
	int			GetResult(PgLib::Result & result);
	Result *	GetResult(void);
	void		ClearResult(PgLib::Result * result = NULL);

	int			SendQuery(const char * query);
	int			SetNonBlocking(int arg) { return this->conn == NULL ? -1 : libpq::PQsetnonblocking(this->conn,arg); }
	int			IsNonBlocking(void)     { return this->conn == NULL ? -1 : libpq::PQisnonblocking(this->conn); }
	int			ConsumeInput(void)      { return this->conn == NULL ? -1 : libpq::PQconsumeInput(this->conn); }
	int			IsBusy(void)            { return this->conn == NULL ? -1 : libpq::PQisBusy(this->conn); }
	int			Flush(void)             { return this->conn == NULL ? -1 : libpq::PQflush(this->conn); }
	int			RequestCancel(void)     { return this->conn == NULL ? -1 : libpq::PQrequestCancel(this->conn); }

	PgTypeItem *	GetPgType(PgLib::Oid oid);

	int			BeginWork(void);
	int			CommitWork(void);
	int			RollbackWork(void);

	static void	NoticeReceiver(void * arg, const libpq::PGresult *res);
	static void NoticeProcessor(void *arg, const char * message);

	virtual	void	NoticeReceiver(const libpq::PGresult *res);
	virtual	void	NoticeProcessor(const char * message);

protected:
	bool				ExecStart(void);
	libpq::PGresult *	ExecFinish(void);

public:
	//////////////////////////////////////////////////////////////////////
	//	VarBuffer
	class	VarBuffer
	{
	public:
		enum varbuffer_enum
		{
			LINE_BUFFER_SIZE =  1024,
		};
	private:
		unsigned	buffersize;
		unsigned	validsize;
		void	   *buffer;

	public:
		VarBuffer();
		virtual ~VarBuffer();
		void		Clear(void);
		int			Add(unsigned size, const void * data);
		int			AddChar(int ch);
		void *		GetBuffer(void) const { return this->buffer; }
		unsigned	GetBufferSize(void) const { return this->validsize; }
		unsigned	GetTextLength(void) { return this->validsize > 0 && ((char *)this->buffer)[this->validsize - 1] == '\0' ? this->validsize - 1 : this->validsize; }
	};
};

#endif
