//////////////////////////////////////////////////////////////////////////
//	"pgtype.cpp"
//------------------------------------------------------------------------
//	$Id: pgtype.cpp 4 2006-03-01 22:26:13Z taro $
//////////////////////////////////////////////////////////////////////////

#include	<stdlib.h>
#include	"pglib.h"

#include	<postgres.h>
#include	<catalog/pg_type.h>

#define	PGOID(mac, name)	mac##OID, PgLib::PGLIB_##mac, name

#if 0
static struct OidTbl
{
	PgLib::Oid			oid;
	PgLib::PgType		type;
	const char *		name;
}
#endif
static  PgLib::PgTypeItem oid_init_tbl[] =
{
	{                      0, PgLib::PGLIB_NONE, "" },
	{ PGOID(BOOL            , "bool"            ) },
	{ PGOID(BYTEA           , "bytea"           ) },
	{ PGOID(CHAR            , "char"            ) },
	{ PGOID(NAME            , "name"            ) },
	{ PGOID(INT8            , "int8"            ) },
	{ PGOID(INT2            , "int2"            ) },
	{ PGOID(INT2VECTOR      , "int2vector"      ) },
	{ PGOID(INT4            , "int4"            ) },
	{ PGOID(REGPROC         , "regproc"         ) },
	{ PGOID(TEXT            , "text"            ) },
	{ PGOID(OID             , "oid"             ) },
	{ PGOID(TID             , "tid"             ) },
	{ PGOID(XID             , "xid"             ) },
	{ PGOID(CID             , "cid"             ) },
	{ PGOID(OIDVECTOR       , "oidvector"       ) },
	{ PGOID(POINT           , "point"           ) },
	{ PGOID(LSEG            , "lseg"            ) },
	{ PGOID(PATH            , "path"            ) },
	{ PGOID(BOX             , "box"             ) },
	{ PGOID(POLYGON         , "polygon"         ) },
	{ PGOID(LINE            , "line"            ) },
	{ PGOID(FLOAT4          , "float4"          ) },
	{ PGOID(FLOAT8          , "float8"          ) },
	{ PGOID(ABSTIME         , "abstime"         ) },
	{ PGOID(RELTIME         , "reltime"         ) },
	{ PGOID(TINTERVAL       , "tinternal"       ) },
	{ PGOID(UNKNOWN         , "unknown"         ) },
	{ PGOID(CIRCLE          , "circle"          ) },
	{ PGOID(CASH            , "cash"            ) },
	{ PGOID(MACADDR         , "macaddr"         ) },
	{ PGOID(INET            , "inet"            ) },
	{ PGOID(CIDR            , "cidr"            ) },
	{ PGOID(ACLITEM         , "aclitem"         ) },
	{ PGOID(BPCHAR          , "bpchar"          ) },
	{ PGOID(VARCHAR         , "varchar"         ) },
	{ PGOID(DATE            , "date"            ) },
	{ PGOID(TIME            , "time"            ) },
	{ PGOID(TIMESTAMP       , "timestamp"       ) },
	{ PGOID(TIMESTAMPTZ     , "timestamptz"     ) },
	{ PGOID(INTERVAL        , "inerval"         ) },
	{ PGOID(TIMETZ          , "timetz"          ) },
	{ PGOID(BIT             , "bit"             ) },
	{ PGOID(VARBIT          , "varbit"          ) },
	{ PGOID(NUMERIC         , "numeric"         ) },
	{ PGOID(REFCURSOR       , "refcuror"        ) },
	{ PGOID(REGPROCEDURE    , "regprocedure"    ) },
	{ PGOID(REGOPER         , "regoper"         ) },
	{ PGOID(REGOPERATOR     , "regoperator"     ) },
	{ PGOID(REGCLASS        , "regclass"        ) },
	{ PGOID(REGTYPE         , "regtype"         ) },
	{ PGOID(RECORD          , "record"          ) },
	{ PGOID(CSTRING         , "cstring"         ) },
	{ PGOID(ANY             , "any"             ) },
	{ PGOID(ANYARRAY        , "anyarray"        ) },
	{ PGOID(VOID            , "void"            ) },
	{ PGOID(TRIGGER         , "trigger"         ) },
	{ PGOID(LANGUAGE_HANDLER, "language_handler") },
	{ PGOID(INTERNAL        , "internal"        ) },
	{ PGOID(OPAQUE          , "opaque"          ) },
	{ PGOID(ANYELEMENT      , "anyelement"      ) },
};


int		PgLib::init_pgtype(void)
{
	unsigned		n;

	for ( n = 0; n < sizeof(oid_init_tbl)/sizeof(*oid_init_tbl); n++ )
	{
		if (  this->add_pgtype(oid_init_tbl[n]) < 0 )
			return -1;
	}

	return 0;
}

void		PgLib::clear_pgtype(void)
{
	unsigned		n;

	for ( n = 0; n < PGTYPE_HASH_MAX; n++ )
	{
		PgLib::PgTypeHash	*pghash, *pghash_next;

		pghash = this->pgtype.tbl[n];
		while( pghash != NULL )
		{
			pghash_next = pghash->next;
			delete pghash;
			pghash = pghash_next;
		}
		this->pgtype.tbl[n] = NULL;
	}
	this->pgtype.num = 0;
}

int		PgLib::add_pgtype(PgLib::PgTypeItem & pgtype)
{
	PgLib::PgTypeHash	**pgbase, *pghash;

	if ( (pghash = new PgLib::PgTypeHash) == NULL )
		return -1;
	(PgLib::PgTypeItem &)*pghash = pgtype;
	pgbase = &this->pgtype.tbl[PgLib::pgtype_hash(pgtype.oid)];

	pghash->next = *pgbase;
	if ( pghash->next != NULL )
		pghash->next->prev = pghash;
	pghash->prev = NULL;
	*pgbase = pghash;
	this->pgtype.num++;

	return 0;
}

unsigned	PgLib::pgtype_hash(PgLib::Oid oid)
{
	return ((unsigned)oid * 31) % PGTYPE_HASH_MAX;
}


PgLib::PgTypeItem *	PgLib::GetPgType(PgLib::Oid oid)
{
	PgLib::PgTypeHash	**pgbase, *pghash;

	pgbase = &this->pgtype.tbl[PgLib::pgtype_hash(oid)];
	for ( pghash = *pgbase; pghash != NULL; pghash = pghash->next )
	{
		if ( pghash->oid == oid )
			break;
	}
	return pghash;
}
