//////////////////////////////////////////////////////////////////////////
//	"scrlib.h"
//------------------------------------------------------------------------
//	$Id: scrlib.h 17 2006-03-04 03:33:26Z taro $
//////////////////////////////////////////////////////////////////////////

#ifndef	SCRLIB_H_INCLUDED
#define	SCRLIB_H_INCLUDED

#if	defined(SDBCTL) || defined(LINUX)
# include	"sqstr.h"
#else
# include	<sqstr.h>
#endif
#include	"namedata.h"
#if	defined(WIN32)
# include	<windows.h>
#else
# include	<stdlib.h>
# include	<stdio.h>
#endif

class	ScrLib;

class	ScrLib
{
public:
	class	ScType;
	class	ScValue;
	friend class	ScType;
	friend class	ScValue;

	enum scrlib_enum
	{
		DEFAULT_IOBUFSIZ = 4096,
	};
	enum reserved_enum
	{
		T_NOTHING  ,
		T_BREAK    ,
		T_CASE     ,
		T_CHAR     ,
		T_CLASS    ,
		T_CONST    ,
		T_CONTINUE ,
		T_DELETE   ,
		T_DO       ,
		T_DOUBLE   ,
		T_ELSE     ,
		T_ENUM     ,
		T_FLOAT    ,
		T_FOR      ,
		T_IF       ,
		T_INT      ,
		T_LONG     ,
		T_NEW      ,
		T_PRIVATE  ,
		T_PROTECTED,
		T_PUBLIC   ,
		T_RETURN   ,
		T_SIGNED   ,
		T_SIZEOF   ,
		T_SHORT    ,
		T_STATIC   ,
		T_STRING   ,
		T_STRUCT   ,
		T_SWITCH   ,
		T_THIS     ,
		T_THROW    ,
		T_TRY      ,
		T_TYPEDEF  ,
		T_UNION    ,
		T_UNSIGNED ,
		T_VIRTUAL  ,
		T_VOID     ,
		T_WHILE    ,

		T_USERTYPE ,
	};
	enum type_enum
	{
		T_NAMEMASK  = 0x0000FFFF,
		T_FLAGTYPE  = 0x00010000,
		T_FLAGVALUE = 0x00020000,
		T_FLAGARRAY = 0x00040000,
	//	T_FLAGMEMB  = 0x00040000,
	};

	class ScType : public NameItem
	{
	public:
	//	SQStr			name;			// ^
		unsigned		type;
		unsigned		size;
		union tagDef
		{
			NameData   *memb;			// T_STRUCT		\̗po
		} def;
	public:
		ScType();
		virtual ~ScType();

		BOOL				IsValue(void) { return this->type & T_FLAGVALUE ? TRUE : FALSE; }

		virtual ScValue *	CreateValue(SQStr & name);
		virtual ScValue *	ParseCreateValue(ScrLib & scrlib, SQStr & name);
		ScValue *			ParseCreateValueArray(ScrLib & scrlib, SQStr & name);

	private:
		ScValue *			ParseCreateValueStruct(ScrLib & scrlib, SQStr & name);
		ScValue *			ParseCreateValueInt(ScrLib & scrlib, SQStr & name);
		ScValue *			ParseCreateValueString(ScrLib & scrlib, SQStr & name);
	public:
		int					AddMember(SQStr & name, ScType * scType);
	};
	class	ScValue : public ScType	// ϐ
	{
	public:
		ScValue();
		virtual ~ScValue();

		SQStr			type_name;		// ̕ϐ̌^
		union tagVal
		{
			int			si;				// T_INT
			void	   *data;
			SQStr	   *str;			// T_STRING
			NameData   *memb;			// T_STRUCT		\̗po
			NameData   *a;				// T_FLAGARRAY	ziDj(ScType.type ɂ T_FLAGARRAY ͂Ȃ
		} val;

		void				Destroy(void);
		static ScValue *	CreateValue(SQStr & name, ScType & scType);
		BOOL				IsArray(void) { return this->type & T_FLAGARRAY ? TRUE : FALSE; }
		BOOL				IsStruct(void) { return (this->type & T_NAMEMASK) == T_STRUCT ? TRUE : FALSE; }
		ScValue *			EnumMemb(ScValue * scValue);
		ScValue *			GetMemb(SQStr & name);
		ScValue *			GetArrayValue(int n);
	};

private:
	NameData		global;

	struct tagIo
	{
#if	defined(WIN32)
		HANDLE			hFile;
#else
		FILE			*fp;
#endif
		unsigned		bufsize;
		BYTE		   *buffer;
		unsigned		validsize;
		unsigned		pos;
		SQStr			str;
		BOOL			unget_flag;
		int				unget_ch;
		unsigned		lno;		// sԍ
		unsigned		cno;		// J
	} io;

public:
	ScrLib();
	~ScrLib();

public:
	int			Init(void);
	int			OpenFile(const TCHAR * path);
	int			CloseFile(void);
	int			CreateIoBuf(unsigned size = DEFAULT_IOBUFSIZ);
	int			ReadBuffer(void);
	int			UnGetChar(int ch);
	int			GetChar(void);
	void		SkipSpace(void);
	void		SkipComment(int ch);
	int			GetCharCheck(int check);
	int			GetEscChar(void);
	int			GetWord(SQStr & word);
	int			IsNameTopChar(int ch);
	int			IsNameChar(int ch);
	int			AddType(NameData & ns, ScType * scType);
	ScValue *	GetValue(SQStr & name);
	ScType *	EnumType(ScType * scType);
	int			Parse(void);

private:
	int			parse_struct(void);
	int			parse_type(ScrLib::ScType & scType);
	int			parse_value(ScrLib::ScValue & scValue);
};

#endif
