//////////////////////////////////////////////////////////////////////////f_QP
//	"smtplib.h"
//------------------------------------------------------------------------
//	$Id: smtplib.cpp 67 2006-03-27 20:33:53Z taro $
//////////////////////////////////////////////////////////////////////////

#include	"smtplib.h"
#include	<stdio.h>
#include	<stdlib.h>
#include	<time.h>
#if	defined(WIN32)
# include	<mbctype.h>
#elif	defined(LINUX)
# include	<sys/time.h>
# include	<unistd.h>
# include	"../linux/defs.h"
#endif

#define	EOLCODE	"\r\n"

#if	! defined(SMTPLIB_VARBUF_BASE64)
const unsigned		SmtpLib::base64tbl[64+1] =
{
	'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
	'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
	'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
	'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
	'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
	'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
	'w', 'x', 'y', 'z', '0', '1', '2', '3',
	'4', '5', '6', '7', '8', '9', '+', '/',
	'='
};
#endif

const char *		SmtpLib::MimeStart = "=?iso-2022-jp?B?";
const char *		SmtpLib::MimeEnd   = "=?=";
const char *		SmtpLib::EscIn     = "\x1B$B";	// {0x1B, 0x24, 0x42};
const char *		SmtpLib::EscKana   = "\x1B(I";	// {0x1B, 0x28, 0x49};
const char *		SmtpLib::EscOut    = "\x1B(J";	// {0x1B, 0x28, 0x4A};
const char *		SmtpLib::SO        = "\x0E";
const char *		SmtpLib::SI        = "\x0F";
const unsigned		SmtpLib::MimeStartSize = 16;
const unsigned		SmtpLib::MimeEndSize   = 3;
const unsigned		SmtpLib::EscInSize     = 3;
const unsigned		SmtpLib::EscKanaSize   = 3;
const unsigned		SmtpLib::EscOutSize    = 3;
const unsigned		SmtpLib::SOSize        = 1;
const unsigned		SmtpLib::SISize        = 1;

#if	defined(SMTPLIB_DEBUG)
#	define	DBG_PRINTF(msg_)	printf msg_
#else
#	define	DBG_PRINTF(msg_)
#endif

SmtpLib::SmtpLib()
{
	this->timeout = TIMEOUT;
	this->sendwait = SENDWAIT;
	this->smtp_addr.sockaddr.sin_family = AF_INET;
	this->smtp_addr.SetPort(PORT_SMTP);
#if	defined(WIN32)
	this->event = NULL;
#endif

	return;
}

SmtpLib::~SmtpLib()
{
	return;
}

int			SmtpLib::CheckAbort(void)
{
#if	defined(WIN32)
	int ret;

	if ( this->event == NULL )
		return 0;
	ret = ::WaitForSingleObject(this->event, 0);
	switch ( ret )
	{
		case WAIT_OBJECT_0:	// VOi
			return 1;

		case WAIT_TIMEOUT:	// VOi
			return 0;

		case WAIT_FAILED:
		case WAIT_ABANDONED:
		default:
			return -1;
	}
#else
	return 0;
#endif
}

int 	SmtpLib::Create(void)
{
	if ( this->IsCreate() == FALSE )
	{
		if ( this->SockLib::Create(SOCK_STREAM, IPPROTO_IP) < 0 )	// TCP
			return -1;
	}
	{
		int		tm;

		tm = TIMEOUT;
		this->SetSockOpt(SOL_SOCKET, SO_SNDTIMEO, &tm, sizeof(tm));
		this->SetSockOpt(SOL_SOCKET, SO_RCVTIMEO, &tm, sizeof(tm));
	}

	return 0;
}

int		SmtpLib::Connect(void)
{
	if ( this->SockLib::Connect(this->smtp_addr) < 0 )
		return -1;
	SmtpLib::Sleep(SYSWAIT);
	return 0;
}

// ڑmF
int		SmtpLib::CheckConnect(void)
{
	VarBuffer	lbuf;
	int			ret;

	this->rescode = 0;
	ret = this->CheckRecv(this->timeout, (LineCheckFunc)&SmtpLib::linecheck_func);
	if ( ret <= 0 )
		return -1;
	if ( this->rescode >= 200 && this->rescode <= 299 )
		return 1;
	return 0;
}

//////////////////////////////////////////////////////////////////////////
//	[M

int		SmtpLib::SendMail(const TCHAR * from, const TCHAR * to, const TCHAR * text, const TCHAR * header, BOOL connect)
{
	int		ret = 0, res;

	if ( connect )
	{
		int		retry;

		for ( retry = 0; retry < CONNECT_RETRY; retry++ )
		{
			DBG_PRINTF(("Connect retry %d\n", retry));
			if ( this->IsCreate() == FALSE )
				this->Create();
			if ( (ret = this->Connect()) < 0 )
			{
				DBG_PRINTF(("Server Connect Error!\n"));
				this->Close();
				SmtpLib::Sleep(CONNECT_WAIT);
				continue;
			}

			if ( this->CheckConnect() > 0 )
				break;	// OK 
			DBG_PRINTF(("Check Connect Error!\n"));
			ret = -1;
			this->Close();
			SmtpLib::Sleep(CONNECT_WAIT);
		}
		if ( ret < 0 )
			goto EXIT;
	}


	// HELO or EHLO
	{
		char	client[256], tmp[512];
		int		type;

		if ( SockLib::GetHostName(sizeof(client), client) < 0 )
			strcpy(client, "localhost");
	//	strcpy(client, "192.168.1.36");

		for ( type = 0; type < 2; type ++ )
		{
			if ( type == 0 )
				sprintf(tmp, "EHLO %s" EOLCODE,  client);
			else
				sprintf(tmp, "HELO %s" EOLCODE,  client);
			DBG_PRINTF(("%s", tmp));
#if	defined(SMTPLIB_LOG_SEND)
			this->vbuf.AddText(tmp);
#endif

			if ( this->CheckSend( (unsigned)strlen(tmp), tmp, this->timeout) <= 0 )
			{
				ret = -1;
				break;
			}
			res = this->CheckRecv(this->timeout * 10);
			if ( res == 250 )
			{
				ret = 0;
				break;
			}
			ret = -1;
		}
		if ( ret < 0 )
			goto EXIT;
	}

	// MAIL FROM
	{
		char	tmp[1024];

		sprintf(tmp, "MAIL FROM:<%s>" EOLCODE, from);
		DBG_PRINTF(("%s", tmp));
#if	defined(SMTPLIB_LOG_SEND)
		this->vbuf.AddText(tmp);
#endif
		if ( this->CheckSend( (unsigned)strlen(tmp), tmp, this->timeout) <= 0 )
		{
			ret = -1;
			goto EXIT;
		}
		res = this->CheckRecv(this->timeout);
		if ( res != 250 )
		{
			ret = -1;
			goto EXIT;
		}
	}

	// RCPT TO
	{
		char	tmp[1024];

		sprintf(tmp, "RCPT TO:<%s>" EOLCODE, to);
		DBG_PRINTF(("%s", tmp));
#if	defined(SMTPLIB_LOG_SEND)
		this->vbuf.AddText(tmp);
#endif
		if ( this->CheckSend( (unsigned)strlen(tmp), tmp, this->timeout) <= 0 )
		{
			ret = -1;
			goto EXIT;
		}
		res = this->CheckRecv(this->timeout);
		if ( res != 250 )
		{
			ret = -1;
			goto EXIT;
		}
	}

	// DATA
	{
		char		*cmd;
		VarBuffer	ebuf;

		cmd = "DATA" EOLCODE;
		DBG_PRINTF(("%s", cmd));
#if	defined(SMTPLIB_LOG_SEND)
		this->vbuf.AddText(cmd);
#endif
		if ( this->CheckSend( (unsigned)strlen(cmd), cmd, this->timeout) <= 0 )
		{
			ret = -1;
			goto EXIT;
		}
		res = this->CheckRecv(this->timeout);
		if ( res != 354 )
		{
			ret = -1;
			goto EXIT;
		}

		// {
#if	defined(SMTPLIB_DEBUG)
		fwrite(text, strlen(text), 1, stdout);
#endif
	//	if ( this->SendText(strlen(text), text, this->timeout) <= 0 )
		if ( header != NULL && strlen(header) > 0 )
		{
			if ( this->SendText((unsigned)strlen(header), header, this->timeout) <= 0 )
			{
				ret = -1;
				goto EXIT;
			}
			if ( this->CheckSend((unsigned)strlen(EOLCODE), EOLCODE, this->timeout) <= 0 )
			{
				ret = -1;
				goto EXIT;
			}
		}

		Sjis2JisText(text, ebuf);
	//	printf("Jis encode size %u\n", ebuf.GetSize());
	//	fwrite(ebuf.GetText(), ebuf.GetSize(), 1, stdout);
		if ( this->SendText(ebuf.GetSize(), ebuf.GetText(), this->timeout) <= 0 )
		{
			ret = -1;
			goto EXIT;
		}
		cmd = "." EOLCODE;
#if	defined(SMTPLIB_LOG_SEND)
		this->vbuf.AddText(cmd);
#endif
		if ( this->CheckSend( (unsigned)strlen(cmd), cmd, this->timeout) <= 0 )
		{
			ret = -1;
			goto EXIT;
		}
		res = this->CheckRecv(this->timeout);
		if ( res != 250 )
		{
			ret = -1;
			goto EXIT;
		}
	}

	// QUIT
	if ( connect )
	{
		if ( this->Quit(this->timeout) < 0 )
		{
			ret = -1;
			goto EXIT;
		}
	}
	ret = 1;	// OK

EXIT:
	if ( connect )
		this->Close();
	return ret;
}

int		SmtpLib::Quit(unsigned msec)
{
	char	*cmd = "QUIT" EOLCODE;
	int		res;

	if ( this->CheckSend( (unsigned)strlen(cmd), cmd, this->timeout) <= 0 )
		return -1;
	res = this->CheckRecv(this->timeout);
	if ( res < 200 || res > 299 )
		return -1;
	return 0;
}

int		SmtpLib::Wait(unsigned msec)
{
	TickTime		tick_start;

	tick_start = SmtpLib::GetTick();
	while ( SmtpLib::GetTick() - tick_start < (TICK_PER_SECOND * msec / 1000) )
	{
		if ( this->CheckAbort() > 0 )
			return 1;	// f
		SmtpLib::Sleep(LOOP_TIMEOUT);
	}
	return 0;
}

//////////////////////////////////////////////////////////////////////////
//	Ytt@Ct̃wb_^eLXg쐬

int		SmtpLib::CreateMutipartMessage(VarBuffer & bdbuf, VarBuffer & body, const char * text, unsigned partnum, SmtpLib::PartData * parttbl)
{
	static unsigned	bid = 0;
	char			boundary[BOUNDARY_MAX], tmp[1024];
	unsigned		len;
	unsigned		pi;

	SmtpLib::GetBoundary(boundary, bid++);
	bdbuf.Clear();
	bdbuf.AddText(boundary);
	bdbuf.AddChar(0);
#if 0
	{
		sprintf(tmp, "Content-Type: multipart/mixed;" EOLCODE);
		header.AddText(tmp);
		sprintf(tmp, "\tboundary=\"%s\"" EOLCODE, boundary);
	}
#endif

	if ( text != NULL && (len = (unsigned)strlen(text)) > 0 )
	{
		body.Add(len, text);
		if ( text[len-1] != '\n' )
			body.AddText(EOLCODE);
		body.AddText(EOLCODE);
	} else if ( text == NULL )
	{
		sprintf(tmp, "This is a multi-part message in MIME format." EOLCODE);
		body.AddText(tmp);
		body.AddText(EOLCODE);
	}

	for ( pi = 0; pi < partnum; pi++ )
	{
		PartData	*pd;

		pd = &parttbl[pi];
		sprintf(tmp, "--%s" EOLCODE, boundary);
		body.AddText(tmp);
		if ( pd->content_type != NULL && strlen(pd->content_type) > 0 )
		{
			sprintf(tmp, "Content-Type: %s", pd->content_type);
			body.AddText(tmp);

			if ( pd->content_name != NULL && strlen(pd->content_name) > 0 )
			{
				VarBuffer	tbuf;

				body.AddText(";" EOLCODE);
				tbuf.Clear();
				SmtpLib::MimeEncode(pd->content_name, tbuf, TRUE);
				tbuf.AddChar(0);
				sprintf(tmp, "\tname=\"%s\"", tbuf.GetText());
				body.AddText(tmp);
			}
			body.AddText(EOLCODE);
		}
		if ( pd->content_id != NULL && strlen(pd->content_id) > 0 )
		{
			sprintf(tmp, "Content-ID: <%s>" EOLCODE, pd->content_id);
			body.AddText(tmp);
		}

		{
			const char	*encode_str;

			switch ( pd->encode )
			{
				case ENCODE_7BIT  : encode_str = "7bit"            ; break;
				case ENCODE_8BIT  : encode_str = "8bit"            ; break;
				case ENCODE_QP    : encode_str = "quoted-printable"; break;
				case ENCODE_BASE64: encode_str = "base64"          ; break;
				default           : encode_str = NULL; break;
			}
			if ( encode_str != NULL )
			{
				sprintf(tmp, "Content-Transfer-Encoding: %s" EOLCODE, encode_str);
				body.AddText(tmp);
			}
		}
		if ( pd->filename != NULL && strlen(pd->filename) > 0 )
		{
			VarBuffer	tbuf;

			sprintf(tmp, "Content-Disposition: attachment;" EOLCODE);
			body.AddText(tmp);
			tbuf.Clear();
			SmtpLib::MimeEncode(pd->filename, tbuf, TRUE);
			tbuf.AddChar(0);
			sprintf(tmp, "\tfilename=\"%s\"" EOLCODE, tbuf.GetText());
			body.AddText(tmp);
		}

		body.AddText(EOLCODE);
		if ( pd->datasize > 0 && pd->data != NULL )
		{
			if ( pd->encode == ENCODE_BASE64 )
			{
				VarBuffer	tbuf;
				char		*p;

				tbuf.Clear();
				SmtpLib::Base64Encode(pd->datasize, pd->data, tbuf, BASE64_MAXLEN);
				len = tbuf.GetSize();
				p = tbuf.GetText();
				body.Add(len, p);
				if ( len > 0 && p[len-1] != '\n' )
					body.AddText(EOLCODE);
			} else if ( pd->encode == ENCODE_QP )
			{
				VarBuffer	tbuf;
				char		*p;

				tbuf.Clear();
				SmtpLib::QuotedPrintableEncode(pd->datasize, pd->data, tbuf, QP_MAXLEN);
				len = tbuf.GetSize();
				p = tbuf.GetText();
				body.Add(len, p);
				if ( len > 0 && p[len-1] != '\n' )
					body.AddText(EOLCODE);
			} else
			{
				char	*p;

				body.Add(pd->datasize, pd->data);
				p = (char *)pd->data;
				if ( pd->datasize > 0 && p[pd->datasize-1] != '\n' )
					body.AddText(EOLCODE);
			}
			body.AddText(EOLCODE);
		}
	}
	sprintf(tmp, "--%s--" EOLCODE, boundary);
	body.AddText(tmp);
	body.AddText(EOLCODE);

	return 0;
}

int		SmtpLib::GetBoundary(char * buffer, unsigned id) // static
{
	time_t		tm;
	TickTime	tick;
	struct tm	*tp;

	::time(&tm);
	tp = ::localtime(&tm);
	tick = SmtpLib::GetTick();
	sprintf( buffer, "----=_NextPart_%04u%02u%02u%02u%02u%02u_%08X_%u"
	, (unsigned) tp->tm_year + 1900
	, (unsigned) tp->tm_mon + 1
	, (unsigned) tp->tm_mday
	, (unsigned) tp->tm_hour
	, (unsigned) tp->tm_min
	, (unsigned) tp->tm_sec
	, (((unsigned)tm & 0xFFFF) << 16) | (((unsigned)tick) & 0xFFFF)
	, id
	);
	return 0;
}

int		SmtpLib::CheckRecv(unsigned msec)
{
	int		ret;

	this->rescode = 0;
	ret = this->CheckRecv(this->timeout, (LineCheckFunc)&SmtpLib::linecheck_func);
	if ( ret <= 0 )
		return -1;
	return this->rescode;
}

int		SmtpLib::CheckRecv(unsigned msec, LineCheckFunc func)
{
	TickTime		tick_start;
	int				ret, result = 0;
	unsigned		size = 0;
	VarBuffer		lbuf;

	if ( this->IsCreate() == FALSE )
		return -1;

	tick_start = SmtpLib::GetTick();
	while ( SmtpLib::GetTick() - tick_start < (TICK_PER_SECOND * msec / 1000) )
	{
		BYTE				buffer[512];

		if ( this->CheckAbort() > 0 )
		{
			result = 0;
			break;
		}
		ret = this->Recv(sizeof(buffer), buffer, 0);
		if ( ret > 0 )
		{
			size += ret;
			lbuf.Add(ret, buffer);
			this->vbuf.Add(ret, buffer);
		}
		{
			TCHAR		*p;
			unsigned	len;

			if ( (len = lbuf.GetLength()) > 0 )
			{
				p = lbuf.GetText();
				if ( func != 0 )
				{
					unsigned	spos, linelen;
					TCHAR		*linetop;
					int			ch;

					spos = 0;
					linelen = 0;
					linetop = p;
					while ( spos < len )
					{
						ch = p[spos++] & 0xFF;
						if ( ch == '\r' )
							continue;
						linelen++;
						if ( ch == '\n' )
						{
							if ( ((*this).*(func))(linelen, linetop) > 0 )
								result = 1;
							linelen = 0;
							linetop = &p[spos];
						}
					}
					if ( linelen > 0 )
					{
						if ( ((*this).*(func))(linelen, linetop) > 0 )
							result = 1;
					}
				}
				if ( p[len-1] == TEXT('\n') )	// s
				{
					lbuf.AddChar(0);
					if ( result > 0 )
						break;
					lbuf.Clear();
				}
			}
		}
	}
	return result;
}

int		SmtpLib::linecheck_func(unsigned len, const char * text)
{
	unsigned	spos, dpos;
	int			ch;
	unsigned	linelen = 0;
	TCHAR		resbuf[256];

	spos = 0;
	dpos = 0;
	while ( spos < len )
	{
		ch = text[spos++] & 0xFF;
		if ( ch >= '0' && ch <= '9' )
		{
			if ( dpos >= sizeof(resbuf)/sizeof(*resbuf) - 1 )
				break;
			resbuf[dpos++] = ch;
		} else
		{
			spos--;
			break;
		}
	}
	if ( dpos == 0 || spos >= len )
		return 0;
	resbuf[dpos] = 0;
	this->rescode = atoi(resbuf);
	ch = text[spos++] & 0xFF;
	if ( ch == ' ' || ch == '\n' )
		return 1;
	return 0;
}

int		SmtpLib::CheckSend(unsigned size, const TCHAR * buffer, unsigned msec)
{
	TickTime		tick_start;
	int				ret, result = 0;
	const TCHAR		*send_ptr = buffer;
	unsigned		send_size = 0, send_remain = size;

	tick_start = SmtpLib::GetTick();
	while ( SmtpLib::GetTick() - tick_start < (TICK_PER_SECOND * msec / 1000) )
	{
		if ( this->CheckAbort() > 0 )
		{
			result = 0;
			break;
		}
		ret = this->Send(send_remain, send_ptr, 0);
		if ( ret > 0 )
		{
			send_size += ret;
			send_ptr += ret;
			if ( ret > (int)send_remain )
				ret = send_remain;
			send_remain -= ret;
		}
		if ( send_size >= (int)size )
		{
			if ( this->sendwait > 0 )
				SmtpLib::Sleep(this->sendwait);
			result = 1;
			break;
		}
	}
	return result;
}

int		SmtpLib::SendText(unsigned size, const TCHAR * buffer, unsigned msec)
{
	const TCHAR	*top = buffer;
	unsigned	send_size = 0;
	unsigned 	spos = 0;
	int			ret = 0;
	BOOL		tol = TRUE;	// top of line

	while ( spos < size )
	{
		int		ch;

		ch = buffer[spos++] & 0xFF;
		if ( ch == '\r' )
			continue;
		if ( ch == '.' && tol )	// 擪sIh
		{
			if ( (ret = this->CheckSend(1, ".", msec)) <= 0 )
				return ret;
		}
		tol = FALSE;
		send_size++;
		if ( ch == '\n' )
		{
			send_size--;
			if ( send_size > 0 )
			{
				if ( (ret = this->CheckSend(send_size, top, msec)) <= 0 )
					return ret;
			}
			if ( (ret = this->CheckSend((unsigned)strlen(EOLCODE), EOLCODE, msec)) <= 0 )
				return ret;
			top = buffer + spos;
			tol = TRUE;
			send_size = 0;
		}
	}
	if ( send_size > 0 )
	{
		if ( (ret = this->CheckSend(send_size, top, msec)) <= 0 )
			return ret;
	}
	if ( buffer[size-1] != '\n' )
	{
		if ( (ret = this->CheckSend((unsigned)strlen(EOLCODE), EOLCODE, msec)) <= 0 )
			return ret;
	}
	return ret;
}

int		SmtpLib::EncodeHeaderText(const TCHAR * name, VarBuffer & wbuf)
{
	unsigned		spos = 0;
	BOOL			useKan = FALSE;
	BOOL			useSpc = FALSE;
	VarBuffer		tbuf;

	spos = 0;
	for (;;)
	{
		int			ch, ch2;

		ch = name[spos] & 0xFF;
		if ( ch == 0 )
			break;
		ch2 = name[spos+1] & 0xFF;
		if ( _ismbblead(ch) && _ismbbtrail(ch2) )
		{
			tbuf.AddChar(ch);
			tbuf.AddChar(ch2);
			useKan = TRUE;
			spos += 2;
		} else
		{
			if ( _ismbbkprint(ch) )	// pJ^Ji
				useKan = TRUE;
			if ( IsSpecialChar(ch) )
			{
		//		tbuf.AddChar('\\');
				useSpc = TRUE;
			}
			tbuf.AddChar(ch);
			spos++;
		}
	}
	tbuf.AddChar(0);

	if ( useSpc )
		wbuf.AddChar('"');
#if 0
	if ( useSpc && useKan )
	{
		VarBuffer	encbuf;

		Sjis2JisText(tbuf.GetText(), encbuf);
		wbuf.Add(MimeStartSize, MimeStart);
		Base64Encode(encbuf.GetSize(), encbuf.GetText(), wbuf);
		wbuf.Add(MimeEndSize  , MimeEnd  );
	} else
#endif
		MimeEncode(tbuf.GetText(), wbuf, useSpc);
	if ( useSpc )
		wbuf.AddChar('"');
	wbuf.AddChar(0);

	return 0;
}

int		SmtpLib::IsSpecialChar(int ch)	// static
{
	switch ( ch )
	{
		case '(':
		case ')':
		case '<':
		case '>':
		case ':':
		case ';':
		case '@':
		case '\\':
		case ',':
	//	case '.':
		case '\"':
			return 1;
	}
	return 0;
}

//#define	USE_SOSI

int		SmtpLib::MimeEncode(const TCHAR * src, VarBuffer & wbuf, BOOL dqenc)	// static
{
	unsigned	spos = 0;
	VarBuffer	encbuf;
	enum encmode_enum
	{
		ENC_OFF,
		ENC_JIS,
		ENC_KANA,
	};
	int			encmode = 0;

	encbuf.Clear();
	for (;;)
	{
		int			ch, ch2;

		if ( (ch = src[spos] & 0xFF) == 0 )
			break;
		ch2 = src[spos+1] & 0xFF;
		if ( _ismbblead(ch) && _ismbbtrail(ch2) )
		{
			if ( encmode != ENC_JIS )
			{
#if	defined(USE_SOSI)
				if ( encmode == ENC_KANA )
					encbuf.Add(SISize, SI);
#endif
				encbuf.Add(EscInSize, EscIn);
				encmode = ENC_JIS;
			}
			ch = Sjis2Jis((ch<<8) | ch2);
			encbuf.AddChar( (ch >> 8) & 0xFF );
			encbuf.AddChar(  ch       & 0xFF );
			spos += 2;
		} else if ( _ismbbkprint(ch) )	// pJ^Ji
		{
			if ( encmode != ENC_KANA )
			{
#if	defined(USE_SOSI)
				if ( encmode == ENC_JIS )
					encbuf.Add(EscOutSize, EscOut);
				encbuf.Add(SOSize, SO);
#else
				encbuf.Add(EscKanaSize, EscKana);
#endif
				encmode = ENC_KANA;
			}
			encbuf.AddChar(ch & 0x7F);
			spos++;
		} else
		{
			if ( encbuf.GetSize() > 0 )
			{
#if	defined(USE_SOSI)
				switch ( encmode )
				{
					case ENC_JIS:
#endif
						encbuf.Add(EscOutSize, EscOut);
#if	defined(USE_SOSI)
						break;
					case ENC_KANA:
						encbuf.Add(SISize, SI);
						break;
				}
#endif
				wbuf.Add(MimeStartSize, MimeStart);
				Base64Encode(encbuf.GetSize(), encbuf.GetText(), wbuf);
				wbuf.Add(MimeEndSize  , MimeEnd  );
				encbuf.Clear();
				encmode = ENC_OFF;
			}
			if ( dqenc && IsSpecialChar(ch) )
				wbuf.AddChar('\\');
			wbuf.AddChar(ch);
			spos++;
		}
	}
	if ( encbuf.GetSize() > 0 )
	{
#if	defined(USE_SOSI)
		switch ( encmode )
		{
			case ENC_JIS:
#endif
				encbuf.Add(EscOutSize, EscOut);
#if	defined(USE_SOSI)
				break;
			case ENC_KANA:
				encbuf.Add(SISize, SI);
				break;
		}
#endif
		wbuf.Add(MimeStartSize, MimeStart);
		Base64Encode(encbuf.GetSize(), encbuf.GetText(), wbuf);
		wbuf.Add(MimeEndSize  , MimeEnd  );
		encbuf.Clear();
		encmode = ENC_OFF;
	}

	return 0;
}

#if	! defined(SMTPLIB_VARBUF_BASE64)
int		SmtpLib::Base64Encode(unsigned size, const void * buffer, VarBuffer & wbuf, unsigned maxlen)	// static
{
	unsigned	spos = 0, pos = 0;
	BYTE		*src = (BYTE *)buffer;
	BOOL		end = FALSE;

	for (;;)
	{
		unsigned	code, prev;
		int			ch;

		// 1st byte
		if ( spos >= size )
			break;
		code = src[spos++] & 0xFF;
		ch = GetBase64Char( code >> 2 );
		prev = code;
		wbuf.AddChar(ch); pos++;

		// 2nd
		if ( spos >= size )
		{
			code = 0;
			end = TRUE;
		} else
			code = src[spos++] & 0xFF;
		ch = GetBase64Char( (prev<<4) | (code>>4) );
		prev =  code;
		wbuf.AddChar(ch); pos++;

		// 3rd
		if ( end )
		{
			ch = '=';
		} else
		{
			if ( spos >= size )
			{
				code = 0;
				end = TRUE;
			} else
				code = src[spos++] & 0xFF;
			ch = GetBase64Char((prev<<2) | (code>>6) );
			prev = code;
		}
		wbuf.AddChar(ch); pos++;

		// 4th
		if ( end )
			ch = '=';
		else
			ch = GetBase64Char(prev);
		wbuf.AddChar(ch); pos++;
		if ( maxlen > 0 && pos >= maxlen )
		{
			wbuf.Add(strlen(EOLCODE), EOLCODE);
			pos = 0;
		}
	}

	return 0;
}
#endif

static int		QuotedPrintableChar(int ch)
{
	return ch != '=' &&
	 (   ch == ' '
	  || ch == '\t'
	  || (ch >= 0x21 && ch <= 0x5E)
	  || (ch >= 0x60 && ch <= 0x7E)
	) ? FALSE : TRUE;
}

int		SmtpLib::QuotedPrintableEncode(unsigned size, const void * buffer, VarBuffer & wbuf, unsigned maxlen)	// static
{
	unsigned	spos = 0, pos = 0;
	BYTE		*src = (BYTE *)buffer;
	BOOL		end = FALSE;

	while ( spos < size )
	{
		int			ch;

		ch = src[spos++] & 0xFF;
		if ( ch == '\r' )
			continue;
		if ( ch == '\n' )
		{
			wbuf.Add((unsigned)strlen(EOLCODE), EOLCODE);
			pos = 0;
		} else if ( QuotedPrintableChar(ch) )
		{
			TCHAR	tmp[16];

			sprintf(tmp, "=%02X", ch);
			wbuf.Add(3, tmp); pos += 3;
		} else
		{
			wbuf.AddChar(ch); pos++;
		}
		if ( maxlen > 0 && pos >= maxlen )
		{
			static const TCHAR	*nc = "=" EOLCODE;

			wbuf.Add((unsigned)strlen(nc), nc);
			pos = 0;
		}
	}
	return 0;
}

int		SmtpLib::Sjis2JisText(const TCHAR * src, VarBuffer & wbuf) // static
{
	unsigned	spos = 0;
	enum encmode_enum
	{
		ESC_OFF,
		ESC_JIS,
		ESC_KANA,
	};
	int			esc = 0;

	for (;;)
	{
		int			ch, ch2;

		if ( (ch = src[spos] & 0xFF) == 0 )
			break;
		if ( ch == '\r' )
		{
			spos++;
			continue;
		}
		ch2 = src[spos+1] & 0xFF;
		if ( _ismbblead(ch) && _ismbbtrail(ch2) )
		{
			if ( esc != ESC_JIS )
			{
				wbuf.Add(EscInSize, EscIn);
				esc = ESC_JIS;
			}
			ch = Sjis2Jis((ch<<8) | ch2);
			wbuf.AddChar( (ch >> 8) & 0xFF );
			wbuf.AddChar(  ch       & 0xFF );
			spos += 2;
		} else if ( _ismbbkprint(ch) )	// pJ^Ji
		{
			if ( esc != ESC_KANA )
			{
				wbuf.Add(EscKanaSize, EscKana);
				esc = ESC_KANA;
			}
			wbuf.AddChar(ch & 0x7F);
			spos++;
		} else
		{
			if ( esc != ESC_OFF  )
			{
				wbuf.Add(EscOutSize, EscOut);
				esc = ESC_OFF;
			}
			if ( ch == '\n' )
				wbuf.AddChar('\r');
			wbuf.AddChar(ch);
			spos++;
		}
	}
	if ( esc != ESC_OFF  )
	{
		wbuf.Add(EscOutSize, EscOut);
		esc = ESC_OFF;
	}
	return 0;
}

// Shift-JISR[hJISR[hɕϊ
int		SmtpLib::Sjis2Jis(int ch)	// static
{
	WORD		x0, x1, x2, y0;
	BYTE		b = LOBYTE(ch);

	if ( (b >= 0x40) && (b <= 0x7F) )
	{
		x0 = 0x8140;
		y0 = 0x2121;
	} else if ( (b >= 0x80) && (b <= 0x9E) )
	{
		x0 = 0x8180;
		y0 = 0x2160;
	} else
	{
		x0 = 0x819f;
		y0 = 0x2221;
	}
	if ( HIBYTE(ch) >= 0xE0 )
	{
		x0 = x0 + 0x5F00;
		y0 = y0 + 0x3E00;
	}
	x1 = (ch - x0) / 0x100;
	x2 = (ch - x0) % 0x100;
	return (y0 + x1*0x200 + x2);
}

int		SmtpLib::Jis2Sjis(int ch)	// static
{
	WORD		n1, n2, sjis;

	n1 = (ch-0x2121) / 0x200;
	n2 = (ch-0x2121) % 0x200;

	if ( n1 <= 0x1E )
		sjis = 0x8100 + n1*256;
	else
		sjis = 0xC100 + n1*256;

	if ( n2 <= 0x3E )
		return (sjis + n2 + 0x40);
	else if ( ( n2>= 0x3F) && (n2 <= 0x5D) )
		return (sjis + n2 + 0x41);
	else
		return (sjis + n2 - 0x61);
}

SmtpLib::TickTime	SmtpLib::GetTick(void) // static
{
#if	defined(WIN32)
	return (TickTime)::GetTickCount();
#elif	defined(LINUX)
	struct timeval	tv;

	gettimeofday(&tv, NULL);
	return (TickTime)tv.tv_sec * CLOCKS_PER_SEC + (TickTime)tv.tv_usec * CLOCKS_PER_SEC / I64(1000000);
#endif
}

void		SmtpLib::Sleep(unsigned msec) // static
{
#if	defined(WIN32)
	::Sleep(msec);
#elif	defined(LINUX)
	usleep( (unsigned)(msec * I64(1000) / I64(1000000)) );
#endif
	return;
}


#if 0
const TCHAR *	SmtpLib::GetDayTime(const time_t * tm, VarBuffer & vbuf) // sttaic
{
	static const char	*montbl[12] =
	{
		"Jan", "Feb", "Mar", "Apr", "May", "Jun",
		"Jul", "Aug", "Sep", "Oct", "Nov", "Dec",
	};
	struct tm	*tp;
	char		tmp[80];

	tp = gmtime(tm);
	strftime( tmp, sizeof(tmp), "%z", tp);

	sprintf(tmp, "%u %s %u %02u:%02u:%02u"
	  , (unsigned)tp->tm_mday
	  , montbl[tp->tm_mon]
	  , (unsigned)tp->tm_year + 1900
	  , (unsigned)tp->tm_hour
	  , (unsigned)tp->tm_min
	  , (unsigned)tp->tm_sec
	);
	vbuf.Clear();
	vbuf.AddText(tmp);
	vbuf.AddChar(0);
	return vbuf.GetText();
}
#endif

#if	defined(SMTPLIB_DEBUG)

int		load_data(VarBuffer & dbuf, SmtpLib::PartData & partdata, const char * path)
{
	FILE		*fp;
	char		*buf;
	unsigned	size;

	if ( (fp = fopen(path, "rb")) == NULL )
		return -1;
	fseek(fp, 0, SEEK_END);
	if ( (size = ftell(fp)) >  0 )
	{
		if ( (buf = new char [size]) != NULL )
		{
			fseek(fp, 0, SEEK_SET);
			fread(buf, size, 1, fp);
			dbuf.Add(size, buf);
			delete [] buf;
		}
	}
	fclose(fp);
	partdata.data     = dbuf.GetText();
	partdata.datasize = dbuf.GetSize();
	partdata.datasize = 16; // dbuf.GetSize();
	return 0;
}

int		parttest(VarBuffer & head, VarBuffer & body, const char * text)
{
	SmtpLib				smtplib;
	VarBuffer			databuf, bodybuf, bd;
	SmtpLib::PartData	partdata[2], *pd;
	int					len;
	unsigned			partnum = 0;
	char				tmp[1024];

	if ( text != NULL && (len = strlen(text)) > 0 )
	{
		pd = &partdata[partnum++];

		pd->content_type = "text/plain; charset=\"iso-2022-jp\"";
		pd->content_name = NULL;
		pd->encode       = SmtpLib::ENCODE_7BIT;
		pd->data         = text;
		pd->datasize     = len;
		pd->filename     = NULL;
	}

	{
		pd = &partdata[partnum++];

		pd->content_type = "text/x-vcard";
		pd->content_name = "vcard.vcf";
		pd->encode       = SmtpLib::ENCODE_QP;
		pd->filename     = "vcard.vcf";
		load_data(databuf, *pd, "vcard.vcf");
	}

	smtplib.CreateMutipartMessage(bd, body, "MultiPart Message" EOLCODE, partnum, partdata);
	sprintf(tmp, "Content-Type: multipart/mixed;" EOLCODE "\tboundary=\"%s\"" EOLCODE, bd.GetText());
	head.AddText(tmp);
	head.AddChar(0);
	body.AddChar(0);

	fwrite(head.GetText(), head.GetSize() - 1, 1, stdout);
	fwrite(body.GetText(), body.GetSize() - 1, 1, stdout);

	{
		char	tmp[256];

		fprintf(stderr, "\n*** stop create multi message *****\n");
		gets(tmp);
	}

	return 0;
}

int		main(int argc, char ** argv)
{
	SmtpLib			smtplib;
	int				i, ret;
	char			*server = "intruder";
	char			*from   = "runser@runser.jp";
	char			*to     = "taro@runser.jp";
	char			*text   =
//	  "[<t.k>]Y" EOLCODE
//	  "" EOLCODE
	  "." EOLCODE
	  ".." EOLCODE
	  "..." EOLCODE
	  "eXg" EOLCODE
	  "test" EOLCODE
//	  "eXg" EOLCODE;
	  ;
//	char			*from_name = " ёY"; //"ёY";
	char			*from_name = "Taro Kobayashi"; //"ёY";
//	char			*subject = " test eXg title ^Cg" ; // "eXg";
	char			*subject = "vcard send test" ; // "eXg";
	char			*filename = NULL;
	VarBuffer		hbuf, bbuf;

#if 0
	{
		char		*src = "ёY";
		VarBuffer	tbuf;

		SmtpLib::MimeEncode(src, tbuf);
		tbuf.AddChar(0);
		printf("%s", tbuf.GetText());
		goto EXIT;
	}
#endif

	for ( i = 1; i < argc; i++ )
	{
		const char	*s;

		s = argv[i];
	}

	if ( smtplib.smtp_addr.SetAddr(server) < 0 )
	{
		fprintf(stderr, "Smtp Serfer Address Error! (%s)\n", server);
		goto EXIT;
	}
	smtplib.smtp_addr.SetPort(SmtpLib::PORT_SMTP);
	if ( smtplib.Create() < 0 )
	{
		fprintf(stderr, "smtplib.Create Error!\n");
		goto EXIT;
	}


	// wb_
	{
		char		tmp[1024];
		VarBuffer	tbuf;

		// From
		tbuf.Clear();
		SmtpLib::EncodeHeaderText(from_name, tbuf);
		sprintf(tmp, "From: %s <%s>" EOLCODE, tbuf.GetText(), from );
		hbuf.Add(strlen(tmp), tmp);

		// To
		tbuf.Clear();
		sprintf(tmp, "To: %s" EOLCODE, to);
		hbuf.Add(strlen(tmp), tmp);

		// Subject
		tbuf.Clear();
		SmtpLib::MimeEncode(subject, tbuf);
		tbuf.AddChar(0);
		sprintf(tmp, "Subject: %s" EOLCODE, tbuf.GetText());
		hbuf.Add(strlen(tmp), tmp);

#if 0
		parttest(hbuf, bbuf, text);
		text = bbuf.GetText();
#else
		hbuf.AddChar(0);
#endif
	}
	ret = smtplib.SendMail(from, to, text, hbuf.GetText());

EXIT:
	if ( smtplib.vbuf.GetSize() > 0 )
	{
		smtplib.vbuf.AddChar(0);
		fputs(smtplib.vbuf.GetText(), stdout);
	}
	{
		char	tmp[256];

		fprintf(stderr, "\n*** progtram terminate *****\n");
		gets(tmp);
	}
	return 0;
}

#endif
