//////////////////////////////////////////////////////////////////////////
//	"socklib.h"
//------------------------------------------------------------------------
//	$Id: socklib.h 51 2006-03-12 15:42:50Z taro $
//////////////////////////////////////////////////////////////////////////
#ifndef	SOCKLIB_H_INCLUDED
#define	SOCKLIB_H_INCLUDED

#if	defined(WIN32)
# ifndef	SOCKLIB_WINSOCK_VER
#  define	SOCKLIB_WINSOCK_VER	2
# endif

# ifndef	WIN32_LEAN_AND_MEAN
#  define WIN32_LEAN_AND_MEAN
# endif

# if	SOCKLIB_WINSOCK_VER == 1
#  include <winsock.h>
# endif
# if	SOCKLIB_WINSOCK_VER == 2
#  include <winsock2.h>
# endif
typedef int		socklen_t;
#elif	defined(LINUX)
# include <sys/socket.h>
# include <arpa/inet.h>
# include <netinet/in.h>
# include <netinet/tcp.h>
# include <netdb.h>
# include <unistd.h>
# ifndef SOCKET
#  define	SOCKET			int
# endif
# ifndef SOCKET_ERROR
#  define	SOCKET_ERROR	(-1)
# endif
# ifndef	BOOL
#	define	BOOL	int
# endif
# ifndef	FALSE
#	define	FALSE	0
# endif
# ifndef	TRUE
#	define	TRUE	1
# endif
#endif

class	SockLib
{
public:
	class SockAddr
	{
	public:
		struct sockaddr_in	sockaddr;
		SockAddr();

#if	defined(WIN32)
		void		SetAddr(unsigned long addr) { this->sockaddr.sin_addr.s_addr = ::htonl(addr); }
		void		SetPort(unsigned port) { this->sockaddr.sin_port = ::htons(port); }
#else
		void		SetAddr(unsigned long addr) { this->sockaddr.sin_addr.s_addr = htonl(addr); }
		void		SetPort(unsigned port) { this->sockaddr.sin_port = htons(port); }
#endif
		int			SetAddr(unsigned long addr, unsigned port);
		int			SetAddr(const char * s);
		const char	*GetAddrText(void);
		const char	*GetAddrText(unsigned length, char * buffer);
		struct hostent	*GetHostent(void);
	};
	enum	socklib_flag_enum
	{
		F_SOCKET     = 0x0001,
	};
	friend class		SockAddr;

private:
	static BOOL			sockInit;
	static int			sockCount;
#if defined(WIN32)
	static WORD			sockVer;
	static WSADATA		wsaData;
#endif

 	unsigned			flags;
	SOCKET				conn_socket;
	int					socket_type;
	int					socket_proto;

public:
	SockLib();
	virtual ~SockLib();

#if	defined(WIN32)
	static unsigned	GetSockVer(void) { return SockLib::sockVer; }
#endif

	static int		Init(void);
	static void		Term(void);
	int			IsCreate(void) { return (this->flags & F_SOCKET) ? TRUE : FALSE; }
	SOCKET		GetSocket(void) const { return this->conn_socket; }
	int			Create(int type, int proto = IPPROTO_IP);
	int			Close(void);
	int			SetSocket(SOCKET socket);
	int			GetSockOpt(int level, int optname, void * optval, socklen_t * optlen);
	int			SetSockOpt(int level, int optname, void * optval, socklen_t optlen);
	int			Connect(SockAddr & addr);
	int			Bind(SockAddr & addr);
	int			Listen(int backlog);
	int			Accept(SockLib * socklib, SockAddr & addr);
	int			Send(unsigned size, const void * buffer, int flag);
	int			SendTo(unsigned size, const void * buffer, int flag, SockLib::SockAddr & toaddr);
	int			Recv(unsigned size, void * buffer, int flag);
	int			RecvFrom(unsigned size, void * buffer, int flag, SockAddr & from);

	static int	GetHostName(unsigned size, char * buffer);
};


#endif
