//////////////////////////////////////////////////////////////////////////
//	"sqllib.cpp"
//------------------------------------------------------------------------
//	$Id: sqllib.cpp 149 2006-04-30 04:51:21Z taro $
//////////////////////////////////////////////////////////////////////////

#include	"sqllib.h"
#if	defined(SQLLIB_SUPPORT_PGLIB)
# include	<pglib.h>
#endif
#include	<string.h>

#if	!defined(HAVE_MEMSET) && !defined(WIN32)
void *	memset(void * s, int c, size_t n)
{
	BYTE	*p = (BYTE *)s;

	while ( n-- > 0 )
		*p++ = c;
	return s;
}
#endif

SqlLib::SqlLib()
{
	this->dbtype = DBTYPE_NONE;
	this->db.work = NULL;
	this->work_level = 0;

	memset(&this->ftbl, 0, sizeof(this->ftbl));
}

SqlLib::~SqlLib()
{
	switch ( this->dbtype )
	{
#if	defined(SQLLIB_SUPPORT_PGLIB)
	case DBTYPE_PGLIB:
		this->pglib_term();
		break;
#endif
	default:
		break;
	}
}

int		SqlLib::SetDBType(SqlLib::DBType dbtype)
{
	if ( this->dbtype == dbtype )
		return 0;
	if ( this->dbtype != DBTYPE_NONE )
		return -1;

	switch ( dbtype )
	{
#if	defined(SQLLIB_SUPPORT_PGLIB)
	case DBTYPE_PGLIB:
		if ( this->pglib_init() < 0 )
			return -1;
		break;
#endif
	default:
		memset(&this->ftbl, 0, sizeof(this->ftbl));
		return -1;
	}
	this->dbtype = dbtype;

	return 0;
}

int		SqlLib::GetDateTimeValue(unsigned size, int * ar, const TCHAR * s) // static
{
	unsigned	an = 0;

	if ( ar == NULL || s == NULL )
		return -1;
	if ( *s == '\'' )
		++s;
	while ( *s && an < size )
	{
		if ( *s >= TEXT('0') && *s <= TEXT('9') )
		{
			int			n, ch;

			n = 0;
			do
			{
				ch = *s++;
				n = n * 10 + (ch - '0');
			} while ( *s >= TEXT('0') && *s <= TEXT('9') );
			ar[an++] = n;
		} else
			s++;
	}
	return an;
}


// wbuf̓NAȂAŌNUL͒ǉȂ
int		SqlLib::EscapeString(SqlLib::VarBuffer & wbuf, unsigned size, const void * buffer, int quat) // static
{
	const unsigned char	*p = (const unsigned char *) buffer;
	unsigned			pos, len = 0;

	if ( quat > 0 )
	{
		wbuf.AddChar(quat);
		len++;
	}
	pos = 0;
	while ( pos < size )
	{
#if 0
		if ( _ismbblead(p[0]) && pos + 1 < size && _ismbbtrail(p[1]) )
		{	// 
			wbuf.AddChar(*p);	len++;
			p++;
			pos++;

			wbuf.AddChar(*p);	len++;
			p++;
			pos++;
		} else
#endif
		{
			switch ( *p )
			{
			case '\'':
			case '\\':
				wbuf.AddChar('\\');		len++;
				break;
			}
			wbuf.AddChar(*p);	len++;
			p++;
			pos++;
		}
	}
	if ( quat > 0 )
	{
		wbuf.AddChar(quat);
		len++;
	}
	return (int)len;
}
// wbuf̓NAȂAŌNUL͒ǉȂ
int		SqlLib::EscapeString(SqlLib::VarBuffer & wbuf, const TCHAR * text, int quat) // static
{
	return SqlLib::EscapeString(wbuf, (unsigned)_tcslen(text), text, quat);
}

// wbuf̓NAȂAŌNUL͒ǉȂ
int		SqlLib::EscapeBinary(SqlLib::VarBuffer & wbuf, unsigned size, const void * buffer, int quat) // static
{
	const unsigned char	*p = (const unsigned char *) buffer;
	unsigned			pos, len = 0;

	if ( quat > 0 )
	{
		wbuf.AddChar(quat);
		len++;
	}
	pos = 0;
	while ( pos < size )
	{
		switch ( *p )
		{
		case '\'':
			wbuf.AddChar('\\');	len++;
			wbuf.AddChar('\'');	len++;
			break;
		case '\\':
			wbuf.AddChar('\\');	len++;
			wbuf.AddChar('\\');	len++;
			wbuf.AddChar('\\');	len++;
			wbuf.AddChar('\\');	len++;
			break;
		default:
			if ( *p < 0x20 || *p > 0x7E )
			{
				char	tmp[32];

				sprintf(tmp, "\\\\%03o", *p);
				wbuf.AddText(tmp);	len += 5;
			} else
			{
				wbuf.AddChar(*p);len++;
			}
			break;
		}
		p++;
		pos++;
	}
	if ( quat > 0 )
	{
		wbuf.AddChar(quat);	len++;
	}

	return (int)len;
}

int		SqlLib::DeEscapeString(SqlLib::VarBuffer & wbuf, unsigned size, const void * buffer) // static
{
	const unsigned char	*p = (const unsigned char *) buffer;
	unsigned			pos, len = 0;

	pos = 0;
	while ( pos < size )
	{
#if 0
		if ( _ismbblead(p[0]) && pos + 1 < size && _ismbbtrail(p[1]) )
		{	// 
			wbuf.AddChar(*p);	len++;
			p++;
			pos++;

			wbuf.AddChar(*p);	len++;
			p++;
			pos++;
		} else
#endif
		if ( *p == '\\' && pos + 1 < size /*&& (p[1] == '\'' || p[1] == '\\')*/ )
		{
			p++;
			pos++;
		}
		wbuf.AddChar(*p);	len++;
		p++;
		pos++;
	}

	return (int)len;
}

int		SqlLib::DeEscapeString(SqlLib::VarBuffer & wbuf, const TCHAR * text, int quat) // static
{
	int		ch, len = 0;

//	_ftprintf(stderr, TEXT("DEBUG: DeEscapeString : %s \n"), text);
	if ( text == NULL )
		return -1;
	if ( quat > 0 )
	{
		while ( (ch = *text++) != TEXT('\0') )
		{
			if ( ch == quat )
				break;
		}
		if ( ch != quat )
			return -1;
	}
	len = 0;
	for (;;)
	{
		ch = *text++;
		if ( ch == TEXT('\0') || (quat > 0 && ch == quat) )
			break;
		if ( ch == TEXT('\\') && *text != TEXT('\0') /*&& (*text == TEXT('\'') || *text == TEXT('\\'))*/ )
			ch = *text++;
		wbuf.AddChar(ch);	len++;
	}
	return (int) len;
}

int		SqlLib::GetTimeParts(unsigned tblnum, const TCHAR * tbl[], const TCHAR * text)
{
	int			ch;
	unsigned	n = 0;

	if ( tblnum == 0 || tbl == NULL || text == NULL )
		return -1;
	tbl[n++] = text;
	while ( (ch = *text++) != TEXT('\0') && n < tblnum )
	{
		switch ( ch )
		{
		case TEXT('\t'):
		case TEXT('\n'):
		case TEXT(' '):
		case TEXT('-'):
		case TEXT(':'):
		case TEXT('.'):
			if ( n < tblnum )
				tbl[n++] = text;
			break;
		}
	}

	return (int)n;
}

int		SqlLib::GetTimeStamp(SqlLib::TimeStamp & ts, const TCHAR * text)
{
	const TCHAR	*tbl[6];
	int			i, n, ret;

	if ( (ret = SqlLib::GetTimeParts(6, tbl, text)) < 3 )
		return -1;
	for ( i = 0; i < 6; i++ )
	{
		n = i < ret ? _ttoi(tbl[i]) : 0;
		switch ( i )
		{
		case 0: ts.year   = n; break;
		case 1: ts.month  = n; break;
		case 2: ts.day    = n; break;
		case 3: ts.hour   = n; break;
		case 4: ts.minute = n; break;
		case 5: ts.second = n; break;
		}
	}
	return 0;
}

int		SqlLib::Begin(void)
{
	if ( this->work_level == 0 )
	{
		SqlLib::Result	result;

		this->work_rollback = FALSE;
		if ( this->Exec(result, TEXT("BEGIN")) < 0
		  || this->GetStatus(result) != SqlLib::RES_OK )
			return -1;
		this->ClearResult(result);
	}
	this->work_level++;
	return 0;
}

int		SqlLib::Commit(void)
{
	if ( this->work_level > 0 )
	{
		this->work_level--;
		if ( this->work_level == 0 )
		{
			SqlLib::Result	result;

			if ( this->work_rollback == false )
			{
				if ( this->Exec(result, TEXT("COMMIT")) < 0 )
					return -1;
			} else
			{
				if ( this->Exec(result,TEXT("ROLLBACK")) < 0 )
					return -1;
			}
			if ( this->GetStatus(result) != SqlLib::RES_OK )
				return -1;
			this->ClearResult(result);
		}
	}
	return 0;
}

int		SqlLib::Rollback(void)
{
	if ( this->work_level > 0 )
	{
		this->work_level--;
		if ( this->work_level == 0  )
		{
			SqlLib::Result	result;

			if ( this->Exec(result, TEXT("ROLLBACK")) < 0
			  || this->GetStatus(result) != SqlLib::RES_OK )
				return -1;
		} else
			this->work_rollback = TRUE;
	}
	return 0;
}

int			SqlLib::GetLastVal(SqlLib::DataValue & val)
{
	SqlLib::Result		result;
	int					ret;

	if ( this->Exec(result, TEXT("SELECT lastval()")) < 0
	  || this->GetStatus(result) != SqlLib::RES_OK )
	{
		return -1;
	}
	ret = this->GetDataValue(result, 0, 0, val);
	this->ClearResult(result);
	return ret;
}

int			SqlLib::GetLastVal(SqlLib::BigInt & bi)
{
	DataValue			val;
	int					ret;

	if ( (ret = this->GetLastVal(val)) < 0 )
		return ret;
	switch ( val.GetDataType() )
	{
	case DTYPE_SMALLINT: bi = val.GetSmallInt(); break;
	case DTYPE_INTEGER : bi = val.GetInteger() ; break;
	case DTYPE_BIGINT  : bi = val.GetBigInt()  ; break;
	default :
		ret = -1;
		break;
	}
	return ret;
}

SqlLib::Int64	SqlLib::atoi64(const TCHAR *s)	// static
{
	Int64	n = 0;
	int		sign = 1;	

	if ( *s == '-' )
	{
		sign = -1;
		s++;
	}
	while ( *s )
	{
		if ( *s < '0' || *s > '9' )
			break;
		n = n * I64(10) + (Int64)(*s - '0');
		s++;
	}
	return sign < 0 ? -n : n;
}
