//////////////////////////////////////////////////////////////////////////
//	"sqllib.h"
//------------------------------------------------------------------------
//	$Id: sqllib.h 149 2006-04-30 04:51:21Z taro $
//////////////////////////////////////////////////////////////////////////
#ifndef	SQLLIB_H_INCLUDED
#define	SQLLIB_H_INCLUDED

#if	defined(WIN32)
# include	<windows.h>
# include	<tchar.h>
#elif	defined(LINUX)
# include	<stdlib.h>
#endif
#include	"sql_defs.h"

#define	SQLLIB_SUPPORT_PGLIB

#if	defined(SQLLIB_SUPPORT_PGLIB)
class	PgLib;
#endif

class	SqlLib 
{
public:
	class	Result;

	enum	DBType
	{
		DBTYPE_NONE ,
#if	defined(SQLLIB_SUPPORT_PGLIB)
		DBTYPE_PGLIB,
#endif
	};

	enum	DataType
	{
		DTYPE_ERROR    = -1,
		DTYPE_NONE     = 0,
		DTYPE_NUL      ,
		DTYPE_OID      ,

		DTYPE_BOOLEAN  ,

		DTYPE_SMALLINT ,
		DTYPE_INTEGER  ,
		DTYPE_BIGINT   ,
		DTYPE_FLOAT    ,
		DTYPE_DOUBLE   ,

		DTYPE_CHAR     ,
		DTYPE_TEXT     ,
		DTYPE_VARCHAR  ,
		DTYPE_BINARY   ,
		DTYPE_VARBINARY,

		DTYPE_DATE     ,
		DTYPE_TIME     ,
		DTYPE_TIMESTAMP,

		DTYPE_SERIAL   ,
		DTYPE_BIGSERIAL,

		DTYPE_USERTYPE ,
		DTYPE_UNKNOWN  ,
	};
#if	defined(WIN32)
	typedef __int64			Int64;
#else
	typedef long long		Int64;
#endif
	typedef BOOL			Boolean;
	typedef int				Oid;
	typedef short int		SmallInt;
	typedef int				Integer;
	typedef Int64			BigInt;
	typedef float			Float;
	typedef double			Double;

#if	defined(_MSC_VER)
#	pragma pack(push,1)
#endif
	struct Date
	{
		unsigned short	year;
		unsigned char	month;
		unsigned char	day;
	};
	struct Time
	{
		unsigned char	hour;
		unsigned char	minute;
		unsigned char	second;
	};
	struct TimeStamp
	{
		unsigned short	year;
		unsigned char	month;
		unsigned char	day;
		unsigned char	hour;
		unsigned char	minute;
		unsigned char	second;
	};
#if	defined(_MSC_VER)
#	pragma pack(pop)
#endif

	enum	ResultStatus
	{
		RES_ERROR = -1,
		RES_OK    = 0,
	};

	//////////////////////////////////////////////////////////////////////
	//	VarBuffer
public:
	class	VarBuffer
	{
	public:
		enum varbuffer_enum
		{
			LINE_BUFFER_SIZE =  1024,
		};
	private:
		unsigned	buffersize;
		unsigned	validsize;
		void	   *buffer;

	public:
		VarBuffer();
		virtual ~VarBuffer();
		void		Clear(void);
		int			Add(unsigned size, const void * data);
		int			AddText(const TCHAR * text);
		int			AddChar(int ch);
		void *		GetBuffer(void) const { return this->buffer; }
		TCHAR *		GetText(void) const { return (TCHAR *) this->buffer; }
		unsigned	GetBufferSize(void) const { return this->validsize; }
		unsigned	GetTextLength(void) { return this->validsize > 0 && ((char *)this->buffer)[this->validsize - 1] == '\0' ? this->validsize - 1 : this->validsize; }
		int			AddTChar(TCHAR ch) { return this->Add(sizeof(ch), &ch); }
		int			AddNulChar(void) { return this->AddTChar(0); }
	};

public:
	class DataValue
	{
	protected:
		class Srep
		{
		friend class	DataValue;
		private:
			DataType			dtype;
			union v_tag
			{
				Boolean			b;
				Oid				oid;
				SmallInt		si;
				Integer			i;
				BigInt			bi;
				Float			f;
				Double			d;
				Date			date;
				Time			time;
				TimeStamp		ts;
				void			*p;
				TCHAR			*text;
			} v;
			int					size;
			int					ref;
		public:
			Srep();
			~Srep();
			void	Clear(void);
			int		SetTextBuffer(DataType dtype, const TCHAR *s, int size);
			int		SetBinaryBuffer(DataType dtype, const void * p, int size);
		};
		Srep			*p;

	public:
		DataValue();
		virtual ~DataValue();
		void	Clear(void);

		DataValue & 		operator = (const DataValue & src);

	protected:
		int		check_srep(void)  { if ( ! this->p ) { if ( !(this->p = new Srep) ) { return -1; } } return 0; }
		int		check_clear(void) { this->Clear(); return this->check_srep(); }
		void	SetTextBuffer(DataType dtype, const TCHAR *s, int size);
		void	SetBinaryBuffer(DataType dtype, const void *p, int size);

	public:
		DataType	GetDataType(void) const { return ! this->p ? DTYPE_NONE : this->p->dtype; }
		unsigned	GetSize(void) const { return ! this->p ? 0 : this->p->size; }

		void	SetSerial   (Integer  i    ) { this->check_clear(); this->p->dtype = DTYPE_SERIAL   ; this->p->v.i    = i   ; }
		void	SetBigSerial(BigInt   bi   ) { this->check_clear(); this->p->dtype = DTYPE_BIGSERIAL; this->p->v.bi   = bi  ; }

		void	SetBoolean  (Boolean  b    ) { this->check_clear(); this->p->dtype = DTYPE_BOOLEAN  ; this->p->v.b    = b   ; }
		void	SetOid      (Oid      oid  ) { this->check_clear(); this->p->dtype = DTYPE_OID      ; this->p->v.oid  = oid ; }
		void	SetSmallInt (SmallInt si   ) { this->check_clear(); this->p->dtype = DTYPE_SMALLINT ; this->p->v.si   = si  ; }
		void	SetInteger  (Integer  i    ) { this->check_clear(); this->p->dtype = DTYPE_INTEGER  ; this->p->v.i    = i   ; }
		void	SetBigInt   (BigInt   bi   ) { this->check_clear(); this->p->dtype = DTYPE_BIGINT   ; this->p->v.bi   = bi  ; }
		void	SetFloat    (Float    f    ) { this->check_clear(); this->p->dtype = DTYPE_FLOAT    ; this->p->v.f    = f   ; }
		void	SetDouble   (Double   d    ) { this->check_clear(); this->p->dtype = DTYPE_DOUBLE   ; this->p->v.d    = d   ; }

		void	SetChar     (const TCHAR *s, int size = -1) { this->check_clear(); this->p->SetTextBuffer(DTYPE_CHAR       , s, size); }
		void	SetText     (const TCHAR *s, int size = -1) { this->check_clear(); this->p->SetTextBuffer(DTYPE_TEXT       , s, size); }
		void	SetVarChar  (const TCHAR *s, int size = -1) { this->check_clear(); this->p->SetTextBuffer(DTYPE_VARCHAR    , s, size); }
		void	SetBinary   (const void * p, int size     ) { this->check_clear(); this->p->SetBinaryBuffer(DTYPE_BINARY   , p, size); }
		void	SetVarBinary(const void * p, int size     ) { this->check_clear(); this->p->SetBinaryBuffer(DTYPE_VARBINARY, p, size); }

		void	SetDate     (const Date   & date ) { this->check_clear(); this->p->dtype = DTYPE_DATE     ; this->p->v.date = date; }
		void	SetTime     (const Time   & time ) { this->check_clear(); this->p->dtype = DTYPE_TIME     ; this->p->v.time = time; }
		void	SetTimeStamp(const TimeStamp & ts) { this->check_clear(); this->p->dtype = DTYPE_TIMESTAMP; this->p->v.ts   = ts  ; }

	public:
		Integer		GetSerial   (void) const { return ! this->p ?     0 : this->p->v.i  ; }
		BigInt		GetBigSerial(void) const { return ! this->p ? I64(0): this->p->v.bi ; }

		Boolean		GetBoolean (void) const { return ! this->p ? FALSE : this->p->v.b  ; }
		Oid			GetOid     (void) const { return ! this->p ?     0 : this->p->v.oid; }
		SmallInt	GetSmallInt(void) const { return ! this->p ?     0 : this->p->v.si ; }
		Integer		GetInteger (void) const { return ! this->p ?     0 : this->p->v.i  ; }
		BigInt		GetBigInt  (void) const { return ! this->p ? I64(0): this->p->v.bi ; }
		Float		GetFloat   (void) const { return ! this->p ? (Float )0: this->p->v.f  ; }
		Double		GetDouble  (void) const { return ! this->p ? (Double)0: this->p->v.d  ; }

		const TCHAR *	GetChar     (void) const { return ! this->p ? NULL : this->p->v.text; }
		const TCHAR *	GetText     (void) const { return ! this->p ? NULL : this->p->v.text; }
		const TCHAR *	GetVarChar  (void) const { return ! this->p ? NULL : this->p->v.text; }
		const void *	GetBinary   (void) const { return ! this->p ? NULL : this->p->v.p   ; }
		const void *	GetVarBinary(void) const { return ! this->p ? NULL : this->p->v.p   ; }

		Date &			GetDate     (void) { this->check_srep(); return this->p->v.date; }
		Time &			GetTime     (void) { this->check_srep(); return this->p->v.time; }
		TimeStamp &		GetTimeStamp(void) { this->check_srep(); return this->p->v.ts  ; }
	};

	//////////////////////////////////////////////////////////////////////
	//	Result
public:
	class	Result
	{
	friend class	SqlLib;
	private:
		struct Rep
		{
			int					ref;
			DBType				dbtype;
			union db_tag
			{
				void *			work;
#if	defined(SQLLIB_SUPPORT_PGLIB)
				void *			pg_res;
#endif
			} db;
			Rep() { this->ref = 1; this->dbtype = DBTYPE_NONE; this->db.work = NULL; };
		};
		Rep *		rep;
	private:
		int			check_rep(void) { if ( this->rep == NULL ) { if ( (this->rep = new Rep) == NULL ) { return -1; } } return 0; }
		int			create(DBType type);
		void		destroy(void);
	public:
		Result();
		Result(DBType dbtype);
		virtual ~Result();

		Result &	operator = (const Result & right);
	protected:
#if	defined(SQLLIB_SUPPORT_PGLIB)
		void *		pglib_getres(void) { return this->rep != NULL && this->rep->dbtype == DBTYPE_PGLIB ? this->rep->db.pg_res : NULL; }
#endif
	};

	//////////////////////////////////////////////////////////////////////
	//	Table Lib
public:
	enum FieldFlag
	{
		FF_SERIAL  = 0x00000001,	// VA^
		FF_PRIMARY = 0x00000002,	// vC}L[
		FF_UNIQUE  = 0x00000004,	// j[N
		FF_INDEX   = 0x00000008,	// CfbNX쐬
	};
	struct	TableItem
	{
		int				fno;
		const TCHAR *	fname;
		DataType		dtype;
		const TCHAR *	tname;
		int				size;	// option (varchar)
		unsigned		flags;
		const TCHAR *	text;	// remark
	};
	class	TableDef
	{
		friend class	SqlLib;
		enum
		{
			DEFAULT_HASHMAX = 8,
		};
	private:
		class	Item : public TableItem
		{
		friend class	TableDef;
		friend class	SqlLib;
		private:
			struct list_tag
			{
				Item		*next;
			} list;
			struct hash_tag
			{
				Item		*next;
			} hash;
		};
		struct list_tag
		{
			unsigned	num;
			Item		*top, *last;
		} list;
		struct hash_tag
		{
			unsigned	num;
			unsigned	tblmax;
			Item **		tbl;
		} hash;
	public:
		TableDef();
		virtual ~TableDef();

	private:
		int			create_hashtbl(void);
		void		destroy(void);
		unsigned	get_hash(Item * item);
		int			add_item(Item * item);
	public:
		unsigned	GetCount(void) const { return this->list.num; }
		int			Add(TableItem & item);
		TableItem *	GetDef(const TCHAR * fname);
		int			GetFno(const TCHAR * fname) { TableItem * item = this->GetDef(fname); return item != NULL ? item->fno : -1; }
	};
	class	FieldsTbl
	{
	friend class	SqlLib;
	private:
		class Data
		{
		friend class	FieldsTbl;
		private:
			Data *			next;
			TCHAR *			fname;
		 	DataValue		val;
		public:
			Data() { this->fname = NULL; }
			~Data() { if ( this->fname != NULL ) delete [] this->fname; }
		};
		int					num;
		Data				*top;
		Data				*last;
		Data				**ar;
	public:
		FieldsTbl();
		FieldsTbl(int num);
		virtual ~FieldsTbl();

		void		Clear(void);
		int			Add(const TCHAR * fname, DataValue & val);
		TCHAR *		GetValue(int n, DataValue & val);
		int			GetCount(void) const { return this->num; }

		int		AddBoolean  (const TCHAR * fname, Boolean  b  ) { DataValue val; val.SetBoolean  (b  ); return this->Add(fname, val); }
		int		AddOid      (const TCHAR * fname, Oid      oid) { DataValue val; val.SetOid      (oid); return this->Add(fname, val); }
		int		AddSmallInt (const TCHAR * fname, SmallInt si ) { DataValue val; val.SetSmallInt (si ); return this->Add(fname, val); }
		int		AddInteger  (const TCHAR * fname, Integer  i  ) { DataValue val; val.SetInteger  (i  ); return this->Add(fname, val); }
		int		AddBigInt   (const TCHAR * fname, BigInt   bi ) { DataValue val; val.SetBigInt   (bi ); return this->Add(fname, val); }
		int		AddFloat    (const TCHAR * fname, Float    f  ) { DataValue val; val.SetFloat    (f  ); return this->Add(fname, val); }
		int		AddDouble   (const TCHAR * fname, Double   d  ) { DataValue val; val.SetDouble   (d  ); return this->Add(fname, val); }

		int		AddChar     (const TCHAR * fname, const TCHAR * s, int size = -1) { DataValue val; val.SetChar     (s,size); return this->Add(fname, val); }
		int		AddText     (const TCHAR * fname, const TCHAR * s, int size = -1);// { DataValue val; val.SetText     (s,size); return this->Add(fname, val); }
		int		AddVarChar  (const TCHAR * fname, const TCHAR * s, int size = -1);// { DataValue val; val.SetVarChar  (s,size); return this->Add(fname, val); }
		int		AddBinary   (const TCHAR * fname, const TCHAR * s, int size     ) { DataValue val; val.SetBinary   (s,size); return this->Add(fname, val); }
		int		AddVarBinary(const TCHAR * fname, const TCHAR * s, int size     ) { DataValue val; val.SetVarBinary(s,size); return this->Add(fname, val); }

		int		AddDate     (const TCHAR * fname, const Date    & date) { DataValue val; val.SetDate(date)   ; return this->Add(fname, val); }
		int		AddTime     (const TCHAR * fname, const Time    & time) { DataValue val; val.SetTime(time)   ; return this->Add(fname, val); }
		int		AddTimeStamp(const TCHAR * fname, const TimeStamp & ts) { DataValue val; val.SetTimeStamp(ts); return this->Add(fname, val); }
	};

public:
	int			CreateTable(const TCHAR * tblname, SqlLib::TableDef & tbldef, const TCHAR * addtbl = NULL);
	int			DropTable(const TCHAR * tblname, SqlLib::TableDef & tbldef);
	int			GetSql_Insert(SqlLib::VarBuffer & sql, const TCHAR * tblname, SqlLib::TableDef & tbldef, SqlLib::FieldsTbl & data);
	int			GetSql_Update(SqlLib::VarBuffer & sql, const TCHAR * tblname, SqlLib::TableDef & tbldef, SqlLib::FieldsTbl & data, const TCHAR * where);

	int			Insert(const TCHAR * tblname, SqlLib::TableDef & tbldef, SqlLib::FieldsTbl & data);
	int			Update(const TCHAR * tblname, SqlLib::TableDef & tbldef, SqlLib::FieldsTbl & data, const TCHAR * where);
	int			Select(SqlLib::Result & result, SqlLib::TableDef & tbldef, const TCHAR * sql);

public:
	struct ConnectInfo
	{
		TCHAR *		host;
		TCHAR *		port;
		TCHAR *		dbname;
		TCHAR *		user;
		TCHAR *		pass;
		TCHAR *		option;
	};
	typedef BOOL   (SqlLib::* Func_IsConnect   )(void);
	typedef int    (SqlLib::* Func_Connect     )(const ConnectInfo * info);
	typedef int    (SqlLib::* Func_Finish      )(void);
	typedef int    (SqlLib::* Func_Exec        )(SqlLib::Result & result, const TCHAR * query);
//	typedef Result * (SqlLib::* Func_ExecRes   )(const TCHAR * query);
	typedef void   (SqlLib::* Func_ClearResult )(Result & result);

	typedef TCHAR *(SqlLib::* Func_GetErrorMessage)(void);

	typedef TCHAR *(SqlLib::* Func_GetTypeName )(SqlLib::VarBuffer & type, SqlLib::TableItem & item);
	typedef int    (SqlLib::* Func_SetSqlValue )(SqlLib::VarBuffer & str, SqlLib::TableItem & item, SqlLib::DataValue & val);

	typedef ResultStatus (SqlLib::* Func_GetStatus)(SqlLib::Result & res);
	typedef int    (SqlLib::* Func_GetTuples   )(SqlLib::Result & res);
	typedef int    (SqlLib::* Func_GetFields   )(SqlLib::Result & res);
	typedef TCHAR *(SqlLib::* Func_GetFname    )(SqlLib::Result & res, int idx);
	typedef int    (SqlLib::* Func_GetFnumber  )(SqlLib::Result & res, const TCHAR * fname);
	typedef DataType (SqlLib::* Func_GetFtype  )(SqlLib::Result & res, int idx);
	typedef int    (SqlLib::* Func_GetFmod     )(SqlLib::Result & res, int idx);
	typedef int    (SqlLib::* Func_GetFsize    )(SqlLib::Result & res, int idx);
	typedef TCHAR *(SqlLib::* Func_GetValue    )(SqlLib::Result & res, int y, int x);
	typedef int    (SqlLib::* Func_GetDataValue)(SqlLib::Result & res, int y, int x, SqlLib::DataValue & val);
	typedef int    (SqlLib::* Func_IsNull      )(SqlLib::Result & res, int y, int x);
	typedef int    (SqlLib::* Func_GetLength   )(SqlLib::Result & res, int y, int x);
	typedef TCHAR *(SqlLib::* Func_GetCmdStatus)(SqlLib::Result & res);
	typedef Oid    (SqlLib::* Func_GetOidValue )(SqlLib::Result & res);

	struct FuncTable
	{
		Func_IsConnect			IsConnect;
		Func_Connect			Connect;
		Func_Finish				Finish;

		Func_Exec				Exec;
//		Func_ExecRes			ExecRes;
		Func_ClearResult		ClearResult;

		Func_GetErrorMessage	GetErrorMessage;

		Func_GetTypeName		GetTypeName;
		Func_SetSqlValue		SetSqlValue;

		Func_GetStatus			GetStatus;
		Func_GetTuples			GetTuples;
		Func_GetFields			GetFields;
		Func_GetFname			GetFname;
		Func_GetFnumber			GetFnumber;
		Func_GetFtype			GetFtype;
		Func_GetFmod			GetFmod;
		Func_GetFsize			GetFsize;
		Func_GetValue			GetValue;
		Func_GetDataValue		GetDataValue;
		Func_IsNull				IsNull;
		Func_GetLength			GetLength;
		Func_GetCmdStatus		GetCmdStatus;
		Func_GetOidValue		GetOidValue;
	};
private:
	FuncTable	ftbl;

private:
	DBType				dbtype;
	union db_tag
	{
		void *			work;
#if	defined(SQLLIB_SUPPORT_PGLIB)
		PgLib *			pglib;
#endif
	} db;
	int			work_level;
	BOOL		work_rollback;

public:
	SqlLib();
	virtual	~SqlLib();

	int			SetDBType(SqlLib::DBType dbtype);

public:
	BOOL		IsConnect(void) { return this->ftbl.IsConnect ? (this->*ftbl.IsConnect)() : FALSE; }
	int			Connect(const ConnectInfo * info) { return this->ftbl.Connect ? (this->*ftbl.Connect)(info) : -1; }
	int			Finish(void) { return this->ftbl.Finish ? (this->*ftbl.Finish)() : -1; }

	int			Exec(Result & result, const TCHAR * query) { return this->ftbl.Exec ? (this->*ftbl.Exec)(result, query) : -1; }
//	Result *	ExecRes(const TCHAR * query) { return this->ftbl.Exec ? (this->*ftbl.ExecRes)(query) : NULL; }
	void		ClearResult(Result & result) { if ( this->ftbl.ClearResult ) (this->*ftbl.ClearResult)(result); }

	TCHAR *		GetErrorMessage(void) { return this->ftbl.GetErrorMessage ? (this->*ftbl.GetErrorMessage)() : (TCHAR *)TEXT(""); }

	TCHAR *		GetTypeName(VarBuffer & type, TableItem & item) { return this->ftbl.GetTypeName ? (this->*ftbl.GetTypeName)(type, item) : NULL; }
	int			SetSqlValue(VarBuffer & str, TableItem & item, DataValue & val) { return this->ftbl.SetSqlValue ? (this->*ftbl.SetSqlValue)(str, item,val) : -1; }

	ResultStatus GetStatus(Result & res) { return this->ftbl.GetStatus ? (this->*ftbl.GetStatus)(res) : RES_ERROR; }
	int			GetTuples(Result & res) { return this->ftbl.GetTuples ? (this->*ftbl.GetTuples)(res) : -1; }
	int			GetFields(Result & res) { return this->ftbl.GetFields ? (this->*ftbl.GetFields)(res) : -1; }
	TCHAR *		GetFname(Result & res, int idx) { return this->ftbl.GetFname ? (this->*ftbl.GetFname)(res,idx) : NULL; }
	int			GetFnumber(Result & res, const TCHAR * fname) { return this->ftbl.GetFnumber ? (this->*ftbl.GetFnumber)(res,fname) : -1; }
	DataType	GetFtype(Result & res, int idx) { return this->ftbl.GetFtype ? (this->*ftbl.GetFtype)(res,idx) : DTYPE_ERROR; }
	int			GetFmod(Result & res, int idx) { return this->ftbl.GetFmod ? (this->*ftbl.GetFmod)(res,idx) : -1; }
	int			GetFsize(Result & res, int idx) { return this->ftbl.GetFsize ? (this->*ftbl.GetFsize)(res,idx) : -1; }
	TCHAR *		GetValue(Result & res, int y, int x) { return this->ftbl.GetValue ? (this->*ftbl.GetValue)(res,y,x) : NULL; }
	int			GetDataValue(Result & res, int y, int x, DataValue & val) { return this->ftbl.GetValue ? (this->*ftbl.GetDataValue)(res,y,x,val) : -1; }
	int			IsNull(Result & res, int y, int x) { return this->ftbl.IsNull ? (this->*ftbl.IsNull)(res,y,x) : -1; }
	int			GetLength(Result & res, int y, int x) { return this->ftbl.GetLength ? (this->*ftbl.GetLength)(res,y,x) : -1; }
	TCHAR *		GetCmdStatus(Result & res) { return this->ftbl.GetCmdStatus ? (this->*ftbl.GetCmdStatus)(res) : NULL; }
	Oid			GetOidValue(Result & res) { return this->ftbl.GetOidValue ? (this->*ftbl.GetOidValue)(res) : (Oid)-1; }

	int			GetLastVal(SqlLib::DataValue & val);
	int			GetLastVal(SqlLib::BigInt & bi);

public:
	static int	GetDateTimeValue(unsigned size, int * ar, const TCHAR * s);
	static int	EscapeString(SqlLib::VarBuffer & wbuf, unsigned size, const void * buffer, int quat = 0);
	static int	EscapeString(SqlLib::VarBuffer & wbuf, const TCHAR * text, int quat = 0);
	static int	EscapeBinary(SqlLib::VarBuffer & wbuf, unsigned size, const void * buffer, int quat = 0);
	static int	DeEscapeString(SqlLib::VarBuffer & wbuf, unsigned size, const void * buffer);
	static int	DeEscapeString(SqlLib::VarBuffer & wbuf, const TCHAR * text, int quat);
	static int	GetTimeParts(unsigned tblnum, const TCHAR * tbl[], const TCHAR * text);
	static int	GetTimeStamp(SqlLib::TimeStamp & ts, const TCHAR * text);

	int			Begin(void);
	int			Commit(void);
	int			Rollback(void);

	static Int64	atoi64(const TCHAR *s);

#if	defined(SQLLIB_SUPPORT_PGLIB)
	//////////////////////////////////////////////////////////////////////
	//	PgLib
private:
	int			pglib_init(void);
	int			pglib_term(void);

	BOOL		pglib_IsConnect(void);
	int			pglib_Connect(const ConnectInfo * info);
	int			pglib_Finish(void);

	int			pglib_Exec(SqlLib::Result & result, const TCHAR * query);
//	Result *	pglib_ExecRes(const TCHAR * query);
	void		pglib_ClearResult(SqlLib::Result & result);

	int			pglib_BeginWork(void);
	int			pglib_CommitWork(void);
	int			pglib_RollbackWork(void);

	TCHAR *		pglib_GetErrorMessage(void);

	TCHAR *		pglib_GetTypeName(SqlLib::VarBuffer & type, SqlLib::TableItem & item);
	int			pglib_SetSqlValue(SqlLib::VarBuffer & str, SqlLib::TableItem & item, SqlLib::DataValue & val);

	ResultStatus pglib_GetStatus(SqlLib::Result & res);
	int			pglib_GetTuples(SqlLib::Result & res);
	int			pglib_GetFields(SqlLib::Result & res);
	TCHAR *		pglib_GetFname(SqlLib::Result & res, int idx);
	int			pglib_GetFnumber(SqlLib::Result & res, const TCHAR * fname);
	DataType	pglib_GetFtype(SqlLib::Result & res, int idx);
	int			pglib_GetFmod(SqlLib::Result & res, int idx);
	int			pglib_GetFsize(SqlLib::Result & res, int idx);
	TCHAR *		pglib_GetValue(SqlLib::Result & res, int y, int x);
	int			pglib_GetDataValue(SqlLib::Result & res, int y, int x, SqlLib::DataValue & val);
	int			pglib_IsNull(SqlLib::Result & res, int y, int x);
	int			pglib_GetLength(SqlLib::Result & res, int y, int x);
	TCHAR *		pglib_GetCmdStatus(SqlLib::Result & res);
	SqlLib::Oid	pglib_GetOidValue(SqlLib::Result & res);

#if 0
	int			pglib_GetBooleanValue(SqlLib::Result & res, SqlLib::Boolean & val, int y, int x);
	int			pglib_GetSmallIntValue(SqlLib::Result & res, SqlLib::SmallInt & val, int y, int x);
	int			pglib_GetIntegerValue(SqlLib::Result & res, SqlLib::Integer & val, int y, int x);
	int			pglib_GetBigIntValue(SqlLib::Result & res, SqlLib::BigInt & val, int y, int x);
	int			pglib_GetFloatValue(SqlLib::Result & res, SqlLib::Float & val, int y, int x);
	int			pglib_GetDoubleValue(SqlLib::Result & res, SqlLib::Double & val, int y, int x);
	int			pglib_GetDateValue(SqlLib::Result & res, SqlLib::Date & val, int y, int x);
	int			pglib_GetTimeValue(SqlLib::Result & res, SqlLib::Time & val, int y, int x);
	int			pglib_GetTimeStampValue(SqlLib::Result & res, SqlLib::TimeStamp & val, int y, int x);
#endif
#endif // end of PgLib

};

#endif
