//////////////////////////////////////////////////////////////////////////
//	"sqltbldef.cpp"
//------------------------------------------------------------------------
//	$Id: sqltbldef.cpp 11 2006-03-03 14:20:05Z taro $
//////////////////////////////////////////////////////////////////////////

#include	"sqllib.h"
#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#if	defined(WIN32)
# include	<tchar.h>
#endif

#if 1 && defined(_DEBUG)
# if	defined(WIN32)
#  include	<crtdbg.h>
#  define	DBGREPORT(param_)	_CrtDbgReport param_
# endif
#else
# define	DBGREPORT(param_)
#endif

//////////////////////////////////////////////////////////////////////////
//	TableDef

SqlLib::TableDef::TableDef()
{
	this->list.num = 0;
	this->list.top = this->list.last = NULL;

	this->hash.num = 0;
	this->hash.tblmax = 0;
	this->hash.tbl = NULL;
}

SqlLib::TableDef::~TableDef()
{
	if ( this->list.top != NULL || this->hash.tbl != NULL )
		this->destroy();
}

int		SqlLib::TableDef::create_hashtbl(void)
{
	int				i;

	if ( this->hash.tbl != NULL )
		return 0;
	this->hash.tblmax = DEFAULT_HASHMAX;
	if ( (this->hash.tbl = new Item * [this->hash.tblmax]) == NULL )
		return -1;
	for ( i = 0; i < (int)this->hash.tblmax; i++ )
		this->hash.tbl[i] = NULL;

	return 0;
}

void	SqlLib::TableDef::destroy(void)
{
	Item		*item, *item_next;

	item = this->list.top;
	while ( item != NULL )
	{
		item_next = item->list.next;
		delete item;
		item = item_next;
	}
	this->list.num = 0;
	this->list.top = this->list.last = NULL;

	if ( this->hash.tbl != NULL )
	{
#if	0	// listitemdeleteς
		int			i;

		for ( i = 0; i < (int)this->hash.tblmax; i++ )
		{
			item = this->hash.tbl[i];
			while ( item != NULL )
			{
				item_next = item->hash.next;
				delete item;
				item = item_next;
			}
			this->hash.tbl[i] = NULL;
		}
#endif
		this->hash.num = 0;
		delete [] this->hash.tbl;
		this->hash.tbl = NULL;
	}
	return;
}


unsigned	SqlLib::TableDef::get_hash(SqlLib::TableDef::Item * item)
{
	unsigned	hash = 0;
	const TCHAR	*name = item->fname;

	while ( *name )
	{
		hash = hash + (*name * 31);
		name++;
	}
	return hash % this->hash.tblmax;
}

int		SqlLib::TableDef::add_item(SqlLib::TableDef::Item * item)
{
	Item	**pitem;

	if ( this->hash.tbl == NULL )
	{
		if ( this->create_hashtbl() < 0 )
			return -1;
	}
	item->list.next= NULL;
	if ( this->list.last != NULL )
		this->list.last->list.next = item;
	else
		this->list.top = item;
	this->list.last = item;
	this->list.num++;

	pitem = &this->hash.tbl[this->get_hash(item)];
	item->hash.next = *pitem;
	*pitem = item;
	this->hash.num++;

	return 0;
}


int		SqlLib::TableDef::Add(SqlLib::TableItem & item)
{
	Item *	newitem;

	if ( (newitem = new Item) == NULL )
		return -1;
	(TableItem &)*newitem = item;

	return this->add_item(newitem);
}

SqlLib::TableItem *	SqlLib::TableDef::GetDef(const TCHAR * fname)
{
	Item	**pitem, *item, itembuf;

	if ( fname == NULL || this->hash.tbl == NULL )
		return NULL;
	itembuf.fno   = 0;
	itembuf.fname = fname;
	itembuf.dtype = DTYPE_NONE;
	itembuf.tname = NULL;
	itembuf.size  = 0;
	itembuf.flags = 0;
	itembuf.text  = NULL;
	pitem = &this->hash.tbl[this->get_hash(&itembuf)];
	for ( item = *pitem; item != NULL; item = item->hash.next )
	{
		if ( item->fname != NULL && _tcscmp(fname, item->fname) == 0 )
			break;
	}
	return (TableItem *)item;
}


//////////////////////////////////////////////////////////////////////////
//	FieldsTbl

SqlLib::FieldsTbl::FieldsTbl()
{
	this->num = 0;
	this->top = this->last = NULL;
	this->ar = NULL;
}

SqlLib::FieldsTbl::~FieldsTbl()
{
	this->Clear();
}

void	SqlLib::FieldsTbl::Clear(void)
{
	Data	*data, *data_next;

	if ( this->ar != NULL )
	{
		delete [] this->ar;
		this->ar = NULL;
	}
	data = this->top;
	while ( data != NULL )
	{
		data_next = data->next;
		delete data;
		data = data_next;
	}
	this->num = 0;
	this->top = this->last = NULL;
	return;
}

int		SqlLib::FieldsTbl::Add(const TCHAR * fname, SqlLib::DataValue & val)
{
	Data	*data;
	int		len;

	if ( (data = new Data) == NULL )
		return -1;
	len = fname != NULL  ? (int)_tcslen(fname) : 0;
	if ( (data->fname = new TCHAR [len + 1]) == NULL )
	{
		delete data;
		return -1;
	}
	if ( fname != NULL )
		_tcscpy(data->fname, fname);
	else
		data->fname[0] = 0;
	data->val = val;

	data->next = NULL;
	if ( this->last != NULL )
		this->last->next = data;
	else
		this->top = data;
	this->last = data;
	this->num++;
	if ( this->ar != NULL )
	{
		delete [] this->ar;
		this->ar = NULL;
	}
	return 0;
}

int		SqlLib::FieldsTbl::AddText(const TCHAR * fname, const TCHAR * s, int size)
{
	DataValue		val;

	if ( s == NULL )
		s = TEXT("");
	val.SetText(s, size);
	return this->Add(fname, val);
}

int		SqlLib::FieldsTbl::AddVarChar(const TCHAR * fname, const TCHAR * s, int size)
{
	DataValue val;

	if ( s == NULL )
		s = TEXT("");
	val.SetVarChar(s,size);
	return this->Add(fname, val);
}

TCHAR *		SqlLib::FieldsTbl::GetValue(int n, SqlLib::DataValue & val)
{
	if ( this->ar == NULL )
	{
		Data *	data;
		int		i;

		if ( (data = this->top) == NULL || this->num == 0 )
			return NULL;
		if ( (this->ar = new Data * [this->num]) == NULL )
			return NULL;
		i = 0;
		while ( data != NULL && i < this->num )
		{
			this->ar[i++] = data;
			data = data->next;
		}
	}
	if ( n < 0 || n >= this->num )
		return NULL;
	val = this->ar[n]->val;
	return this->ar[n]->fname;
}

//////////////////////////////////////////////////////////////////////////
//	Function

int		SqlLib::CreateTable(const TCHAR * tblname, SqlLib::TableDef & tbldef, const TCHAR * addtbl)
{
	int					retcode;

	DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "SqlLib::CreateTable\n"));
	this->Begin();
	try
	{
		VarBuffer			sql, primary, index;
		TableDef::Item *	item;
		Result				res;

		sql.AddText(TEXT("CREATE TABLE "));
		sql.AddText(tblname);
		sql.AddText(TEXT(" ("));
		for ( item = tbldef.list.top; item != NULL; item = item->list.next )
		{
			VarBuffer	type;

			type.Clear();

			if ( item != tbldef.list.top )
				sql.AddText(TEXT(","));

			if ( this->GetTypeName(type, *item) < 0 )
				return -1;
			if ( item->flags & FF_UNIQUE )
				type.AddText(TEXT(" UNIQUE"));

			if ( item->flags & FF_PRIMARY )
			{
				if ( primary.GetTextLength() > 0 )
					primary.AddText(TEXT(","));
				primary.AddText(item->fname);
			}
			sql.AddText(item->fname);
			type.AddNulChar();
			sql.AddText(TEXT(" "));
			sql.AddText(type.GetText());
		}
		if ( primary.GetTextLength() > 0 )
		{
			primary.AddNulChar();
			sql.AddText(TEXT(",PRIMARY KEY ("));
			sql.AddText(primary.GetText());
			sql.AddText(TEXT(")"));
		}
		if ( addtbl != NULL && _tcslen(addtbl) > 0 )
		{
			sql.AddText(TEXT(","));
			sql.AddText(addtbl);
		}
		sql.AddText(TEXT(")"));
		sql.AddNulChar();

#if	0 && defined(_DEBUG)
		DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "  %s\n", sql.GetText()));
#else
		if ( this->Exec(res, sql.GetText()) < 0 || this->GetStatus(res) != RES_OK )
		{
			const TCHAR	* msg;

			msg = this->GetErrorMessage();
			throw -1;
		}
		this->ClearResult(res);
#endif

#if	defined(SQLLIB_SUPPORT_PGLIB)
		if ( this->dbtype == DBTYPE_PGLIB )
		{
			// CfbNX쐬
			for ( item = tbldef.list.top; item != NULL; item = item->list.next )
			{
				if ( item->flags & FF_INDEX )
				{
					sql.Clear();
					sql.AddText(TEXT("CREATE INDEX "));
					sql.AddText(tblname);
					sql.AddText(TEXT("_"));
					sql.AddText(item->fname);
					sql.AddText(TEXT("_idx ON "));
					sql.AddText(tblname);

					sql.AddText(TEXT(" ("));
					sql.AddText(item->fname);
					sql.AddText(TEXT(") "));
					sql.AddNulChar();
#if	0 && defined(_DEBUG)
					DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "  %s\n", sql.GetText()));
#else
					if ( this->Exec(res, sql.GetText()) < 0 || this->GetStatus(res) != RES_OK )
						throw -1;
					this->ClearResult(res);
#endif
				}
			}
		}
#endif	// end of PGLIB
		retcode = 0;
	} catch ( int ret) { retcode = ret; }
	if ( retcode == 0 )
		this->Commit();
	else
		this->Rollback();

	return 0;
}

int		SqlLib::DropTable(const TCHAR * tblname, SqlLib::TableDef & tbldef)
{
	int					retcode;

	DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "SqlLib::DropTable\n"));
	this->Begin();
	try
	{
		VarBuffer			sql;
		TableDef::Item *	item;
		Result				res;

		sql.Clear();
		sql.AddText(TEXT("DROP TABLE "));
		sql.AddText(tblname);
		sql.AddNulChar();
#if	0 && defined(_DEBUG)
		DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "  %s\n", sql.GetText()));
#else
		if ( this->Exec(res, sql.GetText()) < 0 && this->GetStatus(res) != RES_OK )
		{
			DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "  DROP TABLE ERROR: %s\n", this->GetErrorMessage()));
			throw -1;
		}
		this->ClearResult(res);
#endif

		for ( item = tbldef.list.top; item != NULL; item = item->list.next )
		{
#if	defined(SQLLIB_SUPPORT_PGLIB)
			if ( this->dbtype == DBTYPE_PGLIB && (item->flags & FF_SERIAL) )
			{
				VarBuffer	str;

				sql.Clear();
				sql.AddText(TEXT("SELECT * FROM pg_indexes WHERE indexname='"));
				SqlLib::EscapeString(sql, tblname, 0);
				sql.AddText(TEXT("_"));
				sql.AddText(item->fname);
				sql.AddText(TEXT("_seq"));
				sql.AddText(TEXT("'"));
				sql.AddNulChar();
				if ( this->Exec(res, sql.GetText()) >= 0 && this->GetStatus(res) == RES_OK && this->GetTuples(res) > 0 )
				{
					sql.Clear();
					sql.AddText(TEXT("DROP SEQUENCE "));
					sql.AddText(tblname);
					sql.AddText(TEXT("_"));
					sql.AddText(item->fname);
					sql.AddText(TEXT("_seq"));
					sql.AddNulChar();
#if	0 && defined(_DEBUG)
					DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "  %s\n", sql.GetText()));
#else
					if ( this->Exec(res, sql.GetText()) < 0 || this->GetStatus(res) != RES_OK )
					{
						DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "  DROP SEQUNCE ERROR: %s\n", this->GetErrorMessage()));
					}
					this->ClearResult(res);
#endif
				}
			}
			if ( this->dbtype == DBTYPE_PGLIB && (item->flags & FF_INDEX) )
			{
				sql.Clear();
				sql.AddText(TEXT("SELECT * FROM pg_indexes WHERE indexname='"));
				SqlLib::EscapeString(sql, tblname);
				sql.AddText(TEXT("_"));
				sql.AddText(item->fname);
				sql.AddText(TEXT("_seq"));
				sql.AddText(TEXT("'"));
				sql.AddNulChar();
				if ( this->Exec(res, sql.GetText()) >= 0 && this->GetStatus(res) == RES_OK && this->GetTuples(res) > 0 )
				{
					sql.Clear();
					sql.AddText(TEXT("DROP INDEX "));
					sql.AddText(tblname);
					sql.AddText(TEXT("_"));
					sql.AddText(item->fname);
					sql.AddText(TEXT("_idx"));
					sql.AddNulChar();
#if	0 && defined(_DEBUG)
					DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "  %s\n", sql.GetText()));
#else
					if ( this->Exec(res, sql.GetText()) < 0 || this->GetStatus(res) != RES_OK )
					{
						DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "  DROP INDEX ERROR: %s\n", this->GetErrorMessage()));
					}
					this->ClearResult(res);
#endif
				}
			}
#endif // end of DTYPE_PGLIB
		}
		retcode = 0;
	} catch ( int ret) { retcode = ret; }
	if ( retcode == 0 )
		this->Commit();
	else
		this->Rollback();

	return retcode;
}

int		SqlLib::GetSql_Insert(SqlLib::VarBuffer & sql, const TCHAR * tblname, SqlLib::TableDef & tbldef, SqlLib::FieldsTbl & data)
{
	VarBuffer			flds, vals, str;
	Result				res;
	int					i;

	DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "SqlLib::GetSql_Insert tblname=%s\n", tblname));
	sql.Clear();
	for ( i = 0; i < data.GetCount(); i++ )
	{
		const TCHAR *	fname;
		DataValue		val;
		TableItem *		titem;

		if ( (fname = data.GetValue(i, val)) < 0 )
			return -1;
		if ( (titem = tbldef.GetDef(fname)) == NULL )
			return -1;
		if ( titem->flags & FF_SERIAL )
			continue;
		str.Clear();
		if ( this->SetSqlValue(str, *titem, val) < 0 )
			return -1;

		if ( flds.GetTextLength() > 0 )
			flds.AddText(TEXT(","));
		flds.AddText(fname);

		if ( vals.GetTextLength() > 0 )
			vals.AddText(TEXT(","));
		vals.AddText(str.GetText());
	}
	if ( flds.GetTextLength() == 0 || vals.GetTextLength() == 0 )
		return -1;
	flds.AddNulChar();
	vals.AddNulChar();

	sql.AddText(TEXT("INSERT INTO "));
	sql.AddText(tblname);
	sql.AddText(TEXT(" ("));
	sql.AddText(flds.GetText());
	sql.AddText(TEXT(") VALUES ("));
	sql.AddText(vals.GetText());
	sql.AddText(TEXT(")"));
	sql.AddNulChar();

	DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "  sql = %s\n", sql.GetText()));

	return 0;
}

int		SqlLib::GetSql_Update(SqlLib::VarBuffer & sql, const TCHAR * tblname, SqlLib::TableDef & tbldef, SqlLib::FieldsTbl & data, const TCHAR * where)
{
	VarBuffer			sets, str;
	Result				res;
	int					i;

	DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "SqlLib::GetSql_Update tblname=%s\n", tblname));
	sql.Clear();
	for ( i = 0; i < data.GetCount(); i++ )
	{
		const TCHAR *	fname;
		DataValue		val;
		TableItem *		titem;

		if ( (fname = data.GetValue(i, val)) < 0 )
			return -1;
		if ( (titem = tbldef.GetDef(fname)) == NULL )
			return -1;
		if ( titem->flags & FF_SERIAL )
			continue;
		str.Clear();
		if ( this->SetSqlValue(str, *titem, val) < 0 )
			return -1;

		if ( sets.GetTextLength() > 0 )
			sets.AddText(TEXT(","));
		sets.AddText(fname);
		sets.AddText(TEXT("="));
		sets.AddText(str.GetText());
	}
	if ( sets.GetTextLength() == 0 )
		return -1;
	sets.AddNulChar();

	sql.AddText(TEXT("UPDATE "));
	sql.AddText(tblname);
	sql.AddText(TEXT(" SET "));
	sql.AddText(sets.GetText());
	if ( where != NULL && *where != TEXT('\0') )
	{
		sql.AddText(TEXT(" WHERE "));
		sql.AddText(where);
	}
	sql.AddNulChar();
	DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "  sql = %s\n", sql.GetText()));

	return 0;
}
