//////////////////////////////////////////////////////////////////////////
//	"sqstr.h" -- string class header file
//------------------------------------------------------------------------
//	$Id: sqstr.h 17 2006-03-04 03:33:26Z taro $
//////////////////////////////////////////////////////////////////////////
#ifndef	SQSTR_H_INCLUDED
#define	SQSTR_H_INCLUDED

#if	defined(WIN32)
# include	<tchar.h>
# include	<windows.h>
#elif	defined(LINUX)
# include	"../linux/defs.h"
# include	<stdlib.h>
#endif

#define	SQCHCODE_UNICODE	(0)
#define	SQCHCODE_ANSI		(1)
#define	SQCHCODE_SJIS		SQCHCODE_ANSI
#if	defined(_UNICODE)
# define	SQCHCODE_DEFAULT	SQCHCODE_UNICODE
#else
# define	SQCHCODE_DEFAULT	SQCHCODE_ANSI
#endif

enum sqerr_enum
{
	SQERR_NOERR           = 0,
	SQERR_OUT_OF_MEMORY   = -1,
	SQERR_CHCODE_MISMATCH = -2,
};

class SQStr
{
public:
	enum mode_enum
	{
		MODE_FLAG_ALLOC = 0x0001,
	};

private:
	class srep
	{
	public:
		union
		{
			void *		buffer;
			char *		abuf;
			wchar_t	*	wbuf;
		};
		unsigned		length;
		int				chcode;
		unsigned		mode;
		int				ref;
		srep ()
		{
			this->buffer = NULL;
			this->length = 0;
			this->mode = 0;
			this->ref = 1;
			this->chcode = SQCHCODE_DEFAULT;
		};
	};
	srep *				p;

protected:
	int					error;

public:
	SQStr();
	SQStr(const SQStr & str);
	SQStr(const TCHAR *s);
	~SQStr();

	SQStr & 		operator = (const SQStr & str);
	SQStr &			operator = (const TCHAR * s);

	friend int		operator == (const SQStr & s1, const SQStr & s2);
	friend int		operator != (const SQStr & s1, const SQStr & s2) { return !(s1 == s2); }
	friend int		operator == (const SQStr & s1, const TCHAR * s2) { SQStr t2(s2); return s1 == t2; }
	friend int		operator != (const SQStr & s1, const TCHAR * s2) { SQStr t2(s2); return s1 != t2; }

	void			SetChcode(int code) { if ( this->p == NULL ) {this->p = new srep;} this->p->chcode = code; }
	int				GetChcode(void) const { return this->p == NULL ? SQCHCODE_DEFAULT : this->p->chcode; }
	void			ClearText(void);
	TCHAR *			SetText(const TCHAR * s);
	SQStr &			SetText(const SQStr & s);
	TCHAR *			GetText(void) const   { return this->p == NULL ? NULL : (TCHAR *)this->p->buffer; }
	TCHAR *			GetDisplayText(void) const {  return this->GetLength() > 0 ? this->GetText() : (TCHAR *)TEXT(""); }
	const void *	GetBuffer(void) const { return this->p == NULL ? NULL : this->p->buffer; }
	unsigned		GetLength(void) const { return this->p == NULL ?    0 : this->p->length; };
	unsigned		GetBufferSize(void);
	void *			SetBuffer(const void * buffer, unsigned length, int chcode = SQCHCODE_DEFAULT);
	void * 			CreateBuffer(unsigned size, unsigned length, int chcode = SQCHCODE_DEFAULT);
	int				TransferToBuffer(unsigned byte, const void * buffer, int length, int chcode = SQCHCODE_DEFAULT);

	SQStr & 		AddText(const SQStr & str);
	SQStr & 		AddText(const TCHAR * str) { SQStr s; s.SetText(str); return this->AddText(s); }

	void *			GetPtrTop(void) const { return (void *)this->GetBuffer(); }
	int				GetPtrChar(void * ptr) const;
	void *			GetPtrNext(const void * ptr) const;
	void *			PutPtrChar(void * ptr, int ch, unsigned * psize = NULL, unsigned * plen = NULL);

	SQStr &			ChangeChcode(const SQStr & text, int chcode);
	SQStr &			ChangeChcode(int chcode);

	// unicode
	wchar_t *		SetTextWide(const wchar_t * s);
	wchar_t *		SetTextWide(const char * s);
	wchar_t *		SetTextWide(const SQStr & s);
	wchar_t *		GetTextWide(void);
	unsigned		GetLengthWide(void);

	// ansi
	char *			SetTextAnsi(const char * s);
	char *			SetTextAnsi(const wchar_t * s);
	char *			SetTextAnsi(const SQStr & s);
	char *			GetTextAnsi(void);
	unsigned		GetLengthAnsi(void);

	// error 
	void			ClearError(void)   { this->error = 0; }
	int				GetError(void)     { return this->error; }
	void			SetError(int code) { this->error = code; }

	// lib
	static int		StrCmp(const TCHAR * s1, const TCHAR * s2, unsigned n);
	static int		StrCmp(const TCHAR * s1, const TCHAR * s2);
	static int		StrCmpi(const TCHAR * s1, const TCHAR * s2, unsigned n);
	static int		StrCmpi(const TCHAR * s1, const TCHAR * s2);
	int				Trim(const TCHAR * list = NULL);
};

#endif
