//////////////////////////////////////////////////////////////////////////
//	"sdbctl.cpp"
//------------------------------------------------------------------------
//	$Id: sdbctl.cpp 161 2006-05-05 10:39:26Z taro $
//////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#	include "config.h"
#	undef	HAVE_CONFIG_H
#endif
#if	defined(WIN32)	// libpq*.h  winsock.hCN[hwinsock2.hCN[h
# include	<winsock2.h>
#endif
#include	"sdbctl.h"
#include	<sdbhttp.h>
#include	"../lib/getopt.h"
#include	"../lib/tform.h"
#if	defined(WIN32)
# include	"../win32/devctl.h"
#elif	defined(LINUX)
# include	"../linux/devctl.h"
# include	<ctype.h>
# include	<assert.h>
# if TIME_WITH_SYS_TIME
#  include <sys/time.h>
#  include <time.h>
# else
#  if HAVE_SYS_TIME_H
#   include <sys/time.h>
#  else
#   include <time.h>
#  endif
# endif
# ifdef	HAVE_UNISTD_H
#  include	<unistd.h>
# endif
#endif
#include	<stdlib.h>
#include	<stdio.h>
#if defined(WIN32)
# include	<crtdbg.h>
#endif

#if	defined(_DEBUG) && defined(WIN32)
#	define	DBGREPORT(param_)	_CrtDbgReport param_
#elif	defined(DEBUG) && defined(LINUX)
#	define	_ASSERT	assert
#	define	DBGREPORT(param_)	DbgReport param_
#	define	_CRT_WARN			0
	extern int		DbgReport(int type, const char *filename, int linenumber, const char *moduleName, const char *format, ... );
#else
#	define	DBGREPORT(param_)
# ifndef	_ASSERT
#		define	_ASSERT
# endif
#endif

#if	defined(LINUX)
extern int		InputConsole(void);
extern int		SelectConsole(const TCHAR * keys);
#endif

SdbCtl	sdbctl;

enum	CmdOption
{
	OPT_DB_HOST    ,
	OPT_DB_PASS    ,
	OPT_DB_PORT    ,
	OPT_DB_USER    ,
	OPT_DB_NAME    ,
	OPT_DEFFILE    ,
	OPT_DEV        ,
	OPT_CSVFILE    ,
	OPT_GROUPNAME  ,
	OPT_HELP       ,
	OPT_HOSTNAME   ,
	OPT_HTTP       ,
	OPT_HTTP_AUTH  ,
	OPT_HTTP_PASS  ,
	OPT_HTTP_USER  ,
	OPT_MEMO       ,
	OPT_RANDOM_WAIT,
	OPT_SMART      ,
	OPT_SMART_ATTR ,
	OPT_USER_PASS  ,
};

static void	help(GetOpt & getopt);

int	_tmain(int argc, const TCHAR ** argv)
{
	int		retcode = EXIT_SUCCESS;

	DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"_tmail(argc=%d)\n",argc));
	_putts(SDBCTL_TITLE);

	if ( sdbctl.Init() < 0 )
	{
		DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"sdbctl.Init() error\n"));
		return EXIT_FAILURE;
	}
	if ( sdbctl.CommandLine(argc, argv) < 0 )
		retcode = EXIT_FAILURE;
#if	defined(_DEBUG) && defined(WIN32)
	{
		TCHAR	tmp[1024];

		_tprintf(TEXT("*** EXIT ***\n"));
		_getts(tmp);
	}
#endif
	return retcode;
}

static void	help(GetOpt & getopt)
{
	int			i;

	static const TCHAR *msg1[] =
	{
	TEXT("usage: sdbctl [<command>] [<option>...]"),
	TEXT("  <command>"),
	TEXT("    show     .. Show device information"),
	TEXT("    createdb .. Create database"),
	TEXT("    dropdb   .. Drop database"),
	TEXT("    add      .. Add device"),
	TEXT("    delete   .. Delete device"),
	TEXT("    load     .. Load csv file"),
	TEXT("    update   .. Update S.M.A.R.T. data"),
//	TEXT("    watch    .. Watch S.M.A.R.T. database"),
	TEXT("  <option>"),
	};
	for ( i = 0; i < sizeof(msg1)/sizeof(*msg1); i++)
		_putts(msg1[i]);

	getopt.coltbl[GetOpt::COL_1] =  2;
	getopt.coltbl[GetOpt::COL_2] =  4;
	getopt.coltbl[GetOpt::COL_3] = 28;
	getopt.coltbl[GetOpt::COL_4] = 31;

	static const GetOpt::HelpItem	opt1[] =
	{
	{ NULL, OPT_CSVFILE    , TEXT("<path>"    ), TEXT("CSV file          (effective load" )},
	{ NULL, OPT_DB_HOST    , TEXT("<hostname>"), TEXT("Database hostname    (Default ") SDB_DEFAULT_HOST   TEXT(")")},
	{ NULL, OPT_DB_PASS    , TEXT("<password>"), TEXT("Database password    (Default ") SDB_DEFAULT_PASS   TEXT(")")},
	{ NULL, OPT_DB_PORT    , TEXT("<port#>"   ), TEXT("Database port-number (Default ") SDB_DEFAULT_PORT   TEXT(")")},
	{ NULL, OPT_DB_USER    , TEXT("<username>"), TEXT("Database username    (Default ") SDB_DEFAULT_USER   TEXT(")")},
	{ NULL, OPT_DB_NAME    , TEXT("<name>"    ), TEXT("Database name        (Default ") SDB_DEFAULT_DBNAME TEXT(")")},
	{ NULL, OPT_DEFFILE    , TEXT("<path>"    ), TEXT("Deffile path"     )},
	{ NULL, OPT_DEV        , TEXT("<device>"  ), TEXT("Device"           )},
	{ NULL, OPT_GROUPNAME  , TEXT("<name>"    ), TEXT("Groupname         (effective --dev")},
	{ NULL, OPT_HOSTNAME   , TEXT("<hostname>"), TEXT("Local PC hostname (effective --dev")},
	{ NULL, OPT_HTTP       , TEXT(""          ), TEXT("HTTP mode (add, update)")},
	{ NULL, OPT_HTTP_AUTH  , TEXT("<authtype>"), TEXT("HTTP Authorization mode\n<authtype>: {none|basic}")},
	{ NULL, OPT_HTTP_PASS  , TEXT("<password>"), TEXT("HTTP password"    )},
	{ NULL, OPT_HTTP_USER  , TEXT("<username>"), TEXT("HTTP username"    )},
	{ NULL, OPT_MEMO       , TEXT("<text>"    ), TEXT("Memo              (effective --dev")},
	{ NULL, OPT_RANDOM_WAIT, TEXT("<sec>"     ), TEXT("Random wait       (effective update")},
//	{ NULL, OPT_SMART      , TEXT(""          ), TEXT("S.M.A.R.T."       )},	// Default
	{ NULL, OPT_SMART_ATTR , TEXT("<id#>:<type>[:<text>]"), TEXT("S.M.A.R.T. Attribute\n<type> {normal|word0|word1|word2}" )},
	{ NULL, OPT_USER_PASS  , TEXT("<passcode>"), TEXT("User passcode     (effective --dev")},
	};
	getopt.Help(sizeof(opt1)/sizeof(*opt1), opt1);

	return;
}

//////////////////////////////////////////////////////////////////////////
//	SdbCtl
//////////////////////////////////////////////////////////////////////////

SdbCtl::SdbCtl()
{
	this->sdblib = new SdbLib;	_ASSERT(this->sdblib != NULL);

	this->sdbhttp = new SdbHttp;	_ASSERT(this->sdbhttp != NULL);

	this->opt.postmode = POSTMODE_DB;
	this->opt.random_wait_sec = -1;
	this->opt.smart = TRUE;

	return;
}

SdbCtl::~SdbCtl()
{
	if ( this->sdbhttp != NULL )
		delete this->sdbhttp;
	if ( this->sdblib != NULL )
		delete this->sdblib;
	return;
}

int		SdbCtl::Init(void)
{
	int		retcode;

	try
	{
		if ( this->sdblib  == NULL 
		  || this->sdbhttp == NULL )
			return -1;

#if	defined(WIN32)
		// NpX̎擾
		{
			TCHAR	buf[1024];
			int		len;

			if ( (len = ::GetModuleFileName(NULL, buf, sizeof(buf)/sizeof(*buf))) > 0 )
				this->progPath.SetText(buf);
			this->progDir.GetDnam(this->progPath);
		}
#elif	defined(LINUX)
		{
			this->progDir.SetText(SDB_LOADDEF_DIR);
		}
#endif

		// deft@C
		{
#if	defined(SDB_DEFFILENAME)
			this->defFile.SetText(this->progDir);
			this->defFile.AddPath(SDB_DEFFILENAME);
#else
			this->defFile.SetText(this->progPath);
			this->defFile.GetFn(path);
			this->defFile.AddText(TEXT(".def"));
#endif
		}

		retcode = 0;
	} catch(int ret) { retcode = ret; };

	return retcode;
}

int		SdbCtl::LoadDef(void)
{
	if ( this->defFile.GetLength() == 0 )
		return 0;	// no def file

	if ( this->sdblib->LoadScr(this->defFile.GetText()) < 0 )
	{
		DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"sdblib->LoadScr(%s) error\n", this->defFile.GetDisplayText()));
		_ftprintf(stderr, TEXT("Load %s error!\n"), this->defFile.GetDisplayText());
		return -1;
	}

	return 0;
}

int		SdbCtl::CommandLine(int argc, const TCHAR ** argv)
{
	GetOpt							getopt;
	enum	ArgError
	{
		ARG_SUCCESS     ,
		ARG_HELP        ,
		ARG_ERROR_PARAM ,
		ARG_ERROR_GETOPT,
		ARG_ERROR       ,
	};
	ArgError		argError;
	SQStr			errmsg;
	CtlCmd			cmd = CMD_HELP;
	SdbLib			*sdblib = this->sdblib;
	SdbDev			*sdbdev = NULL;
	int				retcode = 0;

	{
		static GetOpt::AddItemParam	tbl[] =
		{
			{ OPT_CSVFILE           , TEXT("csvfile")       , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_DB_HOST           , TEXT("db_host")       , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_DB_PASS           , TEXT("db_pass")       , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_DB_PORT           , TEXT("db_port")       , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_DB_USER           , TEXT("db_user")       , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_DB_NAME           , TEXT("db_name")       , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_DEFFILE           , TEXT("deffile")       , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_DEV               , TEXT("dev")           , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_GROUPNAME         , TEXT("groupname")     , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_HELP              , TEXT("help")          , NULL, GetOpt::VAR_FLAG  , FALSE, { NULL,0,0} },
			{ OPT_HOSTNAME          , TEXT("hostname")      , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_HTTP              , TEXT("http")          , NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_HTTP_AUTH         , TEXT("http_auth")     , NULL, GetOpt::VAR_TEXT  , TRUE , { TEXT("basic"),0,0} },
			{ OPT_HTTP_PASS         , TEXT("http_pass")     , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_HTTP_USER         , TEXT("http_user")     , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_MEMO              , TEXT("memo")          , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_RANDOM_WAIT       , TEXT("random_wait")   , NULL, GetOpt::VAR_INT   , FALSE, { NULL,0,0} },
			{ OPT_SMART             , TEXT("smart")         , NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_SMART_ATTR        , TEXT("smart_attr")    , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_USER_PASS         , TEXT("user_pass")     , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
		};

		if ( getopt.AddItem(sizeof(tbl)/sizeof(*tbl), tbl) < 0 )
		{
			_ftprintf(stderr, TEXT("Getopt AddItem Error!\n"));
			return -1;
		}
	}

	try
	{
		int			an;

		an = 1;
		getopt.SetCurPos(an);
		while ( getopt.GetCurPos() < argc )
		{
			GetOpt::Item	*item;
			const char		*s;

			if ( (item = getopt.ParseArg(an, argc, argv)) == NULL )
			{
				if ( (s = getopt.last_param) != NULL )
				{
					if ( _tcsicmp(s, TEXT("help")) == 0 )
						throw ARG_HELP;
					else if ( _tcsicmp(s, TEXT("show")) == 0 )
						cmd = CMD_SHOW;
					else if ( _tcsicmp(s, TEXT("createdb")) == 0 )
						cmd = CMD_DB_CREATE;
					else if ( _tcsicmp(s, TEXT("dropdb")) == 0 )
						cmd = CMD_DB_DROP;
					else if ( _tcsicmp(s, TEXT("add")) == 0 )
						cmd = CMD_DEV_ADD;
					else if ( _tcsicmp(s, TEXT("delete")) == 0 )
						cmd = CMD_DEV_DEL;
					else if ( _tcsicmp(s, TEXT("load")) == 0 )
						cmd = CMD_LOADCSV;
					else if ( _tcsicmp(s, TEXT("update")) == 0 )
						cmd = CMD_SMART_UPDATE;
					else if ( _tcsicmp(s, TEXT("watch")) == 0 )
						cmd = CMD_WATCH;
					else
						throw ARG_ERROR_GETOPT;
				} else if ( getopt.err_option != NULL || getopt.err_param != NULL )
				{
					throw ARG_ERROR_GETOPT;
				} // end of Arg
			} else
			{
				switch ( item->usertype )
				{
				case OPT_DB_HOST:
					this->opt.postmode = SdbCtl::POSTMODE_DB;
					sdblib->dbinfo.host.SetText(item->var.text);
					break;

				case OPT_DB_PASS:
					sdblib->dbinfo.pass.SetText(item->var.text);
					break;

				case OPT_DB_PORT:
					sdblib->dbinfo.port.SetText(item->var.text);
					break;

				case OPT_DB_USER:
					sdblib->dbinfo.user.SetText(item->var.text);
					break;

				case OPT_DB_NAME:
					sdblib->dbinfo.dbname.SetText(item->var.text);
					break;

				case OPT_DEFFILE:
					this->defFile.SetText(item->var.text);
					break;

				case OPT_DEV:
				{
					const TCHAR		*desc;
					TCHAR			tmp[1024];

					if ( (desc = item->var.text) == NULL )
						throw ARG_ERROR_PARAM;
					if ( sdblib->dev_search(item->var.text) != NULL )
					{
						_stprintf(tmp, TEXT("Device alreay use! (%)\n"), desc);
						errmsg.SetText(tmp);
						throw ARG_ERROR;
					}
					if ( (sdbdev = sdblib->CreateSdbDev(desc)) == NULL )
						throw ARG_ERROR;
					sdblib->dev_push_back(sdbdev);
					break;
				}

				case OPT_CSVFILE:
					this->opt.csvfile.SetText(item->var.text);
					break;

				case OPT_GROUPNAME:
					if ( sdbdev == NULL )
						throw ARG_ERROR_PARAM;
					sdbdev->groupname.SetText(item->var.text);
					break;

				case OPT_HELP:
					throw ARG_HELP;
					break;

				case OPT_HOSTNAME:
					if ( sdbdev == NULL )
						throw ARG_ERROR_PARAM;
					sdbdev->hostname.SetText(item->var.text);
					break;

				case OPT_HTTP:
					this->opt.postmode = item->var.flag ? SdbCtl::POSTMODE_HTTP : SdbCtl::POSTMODE_DB;
					break;

				case OPT_HTTP_AUTH:
					this->sdbhttp->authType = SdbHttp::GetAuthType(item->var.text);
					break;

				case OPT_HTTP_PASS:
					this->sdbhttp->authPass.SetText(item->var.text);
					break;

				case OPT_HTTP_USER:
					this->sdbhttp->authUser.SetText(item->var.text);
					break;

				case OPT_MEMO:
					if ( sdbdev == NULL )
						throw ARG_ERROR_PARAM;
					sdbdev->memo.SetText(item->var.text);
					break;

				case OPT_RANDOM_WAIT:
					this->opt.random_wait_sec = item->var.n;
					break;

				case OPT_SMART:
					this->opt.smart = item->var.flag;
					break;

				case OPT_SMART_ATTR:
					if ( sdbdev == NULL )
						throw ARG_ERROR_PARAM;
					sdbdev->smart.attrList.AddDef(item->var.text);
					break;

				case OPT_USER_PASS:
					if ( sdbdev == NULL )
						throw ARG_ERROR_PARAM;
					sdbdev->user_pass.SetText(item->var.text);
					break;

				default:
					_ASSERT(0);
					break;
				} // end of switch
			}
		} // end of while
		argError = ARG_SUCCESS;
	} catch(ArgError code) { argError = code; }

	if ( argError != ARG_SUCCESS )
	{
		switch ( argError )
		{
		case ARG_HELP:
			help(getopt);
			break;

		case ARG_ERROR_PARAM:
			_ftprintf(stderr, TEXT("Parametr error! (%s)\n"), getopt.GetLastPos() >= 0  ? argv[getopt.GetLastPos()] : TEXT("") );
			break;

		case ARG_ERROR_GETOPT:
			_ftprintf(stderr, TEXT("Option parameter error!\n"));
			if ( getopt.err_option != NULL )
				_ftprintf(stderr, TEXT("  option   =\"%s\"\n"), getopt.err_option);
			if ( getopt.err_param != NULL )
				_ftprintf(stderr, TEXT("  parameter=\"%s\"\n"), getopt.err_param);
			break;

		case ARG_ERROR:
		default:
			if ( errmsg.GetLength() == 0 )
				errmsg.SetText(TEXT("Internal error!\n"));
			_fputts(errmsg.GetDisplayText(), stderr);
			break;
		}
		return -1;
	}
	if ( this->LoadDef() < 0 )
		return -1;

#if	0 // defined(_DEBUG)
		_ftprintf(stderr, TEXT("db_host = %s\n"), sdb::db_host);
		_ftprintf(stderr, TEXT("db_port = %s\n"), sdb::db_port);
		_ftprintf(stderr, TEXT("db_user = %s\n"), sdb::db_user);
		_ftprintf(stderr, TEXT("db_pass = %s\n"), sdb::db_pass);
		_ftprintf(stderr, TEXT("db_name = %s\n"), sdb::db_name);
#endif
	{
		_ASSERT(sdblib != NULL);
		sdblib->conn_info.host   = sdblib->dbinfo.host.GetLength()   > 0 ? sdblib->dbinfo.host.GetText()  : (TCHAR *)sdb::db_host;
		sdblib->conn_info.port   = sdblib->dbinfo.port.GetLength()   > 0 ? sdblib->dbinfo.port.GetText()  : (TCHAR *)sdb::db_port;
		sdblib->conn_info.user   = sdblib->dbinfo.user.GetLength()   > 0 ? sdblib->dbinfo.user.GetText()  : (TCHAR *)sdb::db_user;
		sdblib->conn_info.pass   = sdblib->dbinfo.pass.GetLength()   > 0 ? sdblib->dbinfo.pass.GetText()  : (TCHAR *)sdb::db_pass;
		sdblib->conn_info.dbname = sdblib->dbinfo.dbname.GetLength() > 0 ? sdblib->dbinfo.dbname.GetText(): (TCHAR *)sdb::db_name;
		if ( sdblib->DBOpen() < 0 )
		{
			_ftprintf(stderr, TEXT("DBOpen error!\n"));
		}
	}
	if ( this->opt.postmode == SdbCtl::POSTMODE_HTTP )
	{
		_ASSERT(this->sdbhttp != NULL);

		if ( this->sdbhttp->authType == SdbHttp::AUTHTYPE_UNDECIDED && sdb::http_auth_type != NULL )
			this->sdbhttp->authType = SdbHttp::GetAuthType(sdb::http_auth_type);
		if ( this->sdbhttp->authUser.GetLength() == 0 && sdb::http_auth_user != NULL )
			this->sdbhttp->authUser.SetText(sdb::http_auth_user);
		if ( this->sdbhttp->authPass.GetLength() == 0 && sdb::http_auth_pass != NULL )
			this->sdbhttp->authPass.SetText(sdb::http_auth_pass);

		this->sdbhttp->host.SetText(sdb::http_post_host);
		this->sdbhttp->port.SetText(sdb::http_post_port);
		this->sdbhttp->path.SetText(sdb::http_post_path);
		DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"  host=%s, port=%s, path=%s\n"
		  , this->sdbhttp->host.GetDisplayText()
		  , this->sdbhttp->port.GetDisplayText()
		  , this->sdbhttp->path.GetDisplayText() ));
		if ( this->sdbhttp->host.GetLength() == 0 
		  || this->sdbhttp->port.GetLength() == 0 
		  || this->sdbhttp->path.GetLength() == 0 )
		{
			_ftprintf(stderr, TEXT("HTTP error!\n"));
			return -1;
		}
	}

	switch ( cmd )
	{
	case CMD_SHOW:
		this->DeviceWork(CMD_SHOW);
		break;

	case CMD_DB_CREATE:
		if ( this->CreateDatabase() < 0 )
			retcode = -1;
		break;
	case CMD_DB_DROP:
		if ( this->DropDatabase() < 0 )
			retcode = -1;
		break;

	case CMD_DEV_ADD:
		this->DeviceWork(CMD_DEV_ADD);
		break;

	case CMD_DEV_DEL:
		this->DeviceWork(CMD_DEV_DEL);
		break;

	case CMD_LOADCSV:
		if ( this->sdblib->dev_list.size() == 0 )
			this->LoadCsv(NULL);
		else
			this->DeviceWork(CMD_LOADCSV);
		break;

	case CMD_SMART_UPDATE:
		this->DeviceWork(CMD_SMART_UPDATE);
		break;

	case CMD_WATCH:
		this->WatchSmartDatabase();
		break;
	}

	sdblib->DBClose();

	return retcode;
}


//////////////////////////////////////////////////////////////////////////
//	Create new database

int		SdbCtl::CreateDatabase(void)
{
	_ASSERT(this->sdblib != NULL);
	_ASSERT(this->sdblib->conn_info.dbname != NULL);
	_ftprintf(stdout, TEXT("Create %s database\n"), this->sdblib->conn_info.dbname);
	if ( this->sdblib->DBIsConnect() == FALSE )
	{
		_ftprintf(stderr, TEXT("Offline database!\n"));
		return -1;
	}
	if ( this->sdblib->DBCreate() < 0 )
	{
		_ftprintf(stderr, TEXT("Create database error!\n"));
		return -1;
	}
	_ftprintf(stdout, TEXT("Create database success\n"));
	return 0;
}

//////////////////////////////////////////////////////////////////////////
//	Delete database

int		SdbCtl::DropDatabase(void)
{
	SqlLib				*sqllib;

	_ASSERT(this->sdblib != NULL);
	sqllib = this->sdblib->sqllib;
	_ASSERT(sqllib != NULL);
	_ASSERT(this->sdblib->conn_info.dbname != NULL);
	_ftprintf(stdout, TEXT("Drop %s database\n"), this->sdblib->conn_info.dbname);
	if ( this->sdblib->DBIsConnect() == FALSE )
	{
		_ftprintf(stderr, TEXT("Offline database!\n"));
		return -1;
	}

#if	defined(WIN32)
	{
		TCHAR	msg[1024];
		int		res;

		_stprintf(msg, TEXT("f[^x[X폜܂B"));
		res = ::MessageBox(NULL, msg, TEXT("mF"), MB_YESNO);
		if ( res != IDYES )
			return -1;
	}
#else
	{
		int		ret;

		_ftprintf(stderr, TEXT("Delete database (Y/N) ?\n"));
		ret = SelectConsole(TEXT("YN"));
		if ( ret != TEXT('Y') )
			return -1;
	}
#endif

	if ( this->sdblib->DBDrop() < 0 )
	{
		_ftprintf(stderr, TEXT("DBDrop error!\n"));
		return -1;
	}
	_ftprintf(stdout, TEXT("Drop database success\n"));
	return 0;
}


int		SdbCtl::DeviceWork(CtlCmd cmd)
{
	if ( cmd == CMD_SMART_UPDATE && (this->opt.random_wait_sec > 0 || sdb::random_wait_sec> 0) )
	{
		unsigned	sec;

		if ( this->opt.random_wait_sec > 0 )
			sec = this->opt.random_wait_sec;
		else
			sec = sdb::random_wait_sec;
		if ( sec > 0 )
		{
#if	defined(WIN32)
			srand(::GetTickCount());
#elif	defined(HAVE_GETTIMEOFDAY)
			{
				struct timeval	tv;

				gettimeofday(&tv, NULL);
				srand(tv.tv_usec);
			}
#endif
			sec = (rand() % sec) + 1;
			_ftprintf(stderr, TEXT("Random wait %u sec\n"), sec);
#if	defined(WIN32)
			::Sleep(sec * 1000);
#else
			sleep(sec);
#endif
		}
	}
	if ( this->sdblib->dev_list.size() == 0 )
	{	// 
		int			devno;
		SdbDev		*sdbdev;

		for ( devno = 0; ; devno++)
		{
			TCHAR	desc[80];

			_stprintf(desc, TEXT("disk%d"), devno);
			if ( (sdbdev = this->sdblib->CreateSdbDev(desc)) == NULL )
				return -1;
			this->sdblib->dev_push_back(sdbdev);
			if ( this->DeviceWork(cmd, sdbdev) < 0 )
				break;
		}
	} else
	{
		SdbLib::Dev_list::iterator	it;
		SdbDev						*sdbdev;

		for ( it = this->sdblib->dev_list.begin(); it != this->sdblib->dev_list.end(); it++ )
		{
			sdbdev = *it;
			_ASSERT(sdbdev != NULL);
			this->DeviceWork(cmd, sdbdev);
		}
	}

	return 0;
}

int		SdbCtl::DeviceWork(CtlCmd cmd, SdbDev * dev)
{
	unsigned	rwmode;
	TCHAR		t1[80];
	BOOL		smart_flag = FALSE;

	_ASSERT(dev != NULL);
	_ftprintf(stdout, TEXT("Device %s\n"), dev->desc.GetDisplayText());
	DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"SDbCtl::WorkDevice(%s)\n",dev->desc.GetDisplayText()));
	rwmode = DevCtl::RW_READ | DevCtl::RW_WRITE + DevCtl::RW_SHARE_READ + DevCtl::RW_SHARE_WRITE + DevCtl::RW_NOBUFFER;
	if ( dev->OpenDevice(rwmode) < 0 )
		return -1;
	if ( dev->GetSmart() == 0 && dev->GetDeviceData() == 0 )
	{
	//	SqlLib::FieldsTbl	data;

		smart_flag = TRUE;
		_ftprintf(stdout, TEXT("Model   : %s\n"), dev->model.GetDisplayText() );
		_ftprintf(stdout, TEXT("Firmware: %s\n"), dev->firm.GetDisplayText() );
		_ftprintf(stdout, TEXT("Serial  : %s\n"), dev->serial.GetDisplayText() );
		TForm::SPrintf(t1, TEXT("%I64k"), (TForm::Int64)dev->blockSize * dev->blockNumber);
		_ftprintf(stdout, TEXT("Capacity: %s\n"), t1);
	//	this->sdblib->dbdev.SearchDevice(data, dev->model.GetText(), dev->serial.GetText());
	}
	switch ( cmd )
	{
	case CMD_SHOW:
		if ( this->opt.smart && smart_flag )
		{
			this->PrintSmart(dev);
		}
		break;

	case CMD_DEV_ADD:
	{
		if ( smart_flag == FALSE
		  || dev->smart.identify == NULL || dev->model.GetLength() == 0 )
		{
			_ftprintf(stderr, TEXT("Device error!\n") );
			break;
		}
		this->AddDevice(dev);
		break;
	}

	case CMD_DEV_DEL:
		this->RemoveDevice(dev);
		break;

	case CMD_LOADCSV:
		this->LoadCsv(dev);
		break;

	case CMD_SMART_UPDATE:
		this->SmartUpdate(dev);
		break;
	}

	dev->CloseDevice();
	return 0;
}

//////////////////////////////////////////////////////////////////////////
//	Add Device

int		SdbCtl::AddDevice(SdbDev * dev)
{
	SqlLib::FieldsTbl	fdata;
	sdb::DBDevData		devdata;

	_ASSERT(dev->dbdev != NULL);
	switch ( this->opt.postmode )
	{
	case POSTMODE_DB:
		if ( dev->dbdev->SearchDevice(fdata, dev->model.GetText(), dev->serial.GetText()) > 0  )
		{
			_ftprintf(stderr, TEXT("Device already exist!\n") );
			return -1;
		}
		break;
	case POSTMODE_HTTP:
		break;
	}
#if	defined(WIN32)
	{
		TCHAR	msg[1024];
		int		res;

		_stprintf(msg, TEXT("̃foCXf[^x[Xɒǉ܂B\r\n  Model: %s\r\n  Serial: %s")
		  , dev->model.GetDisplayText()
		  , dev->serial.GetDisplayText()
		);
		res = ::MessageBox(NULL, msg, TEXT("mF"), MB_YESNO);
		if ( res != IDYES )
			return -1;
	}
#else
	{
		int		ret;

		_ftprintf(stderr, TEXT("Add database (Y/N) ?\n  Model: %s\n  Serial: %s\n")
		  , dev->model.GetDisplayText()
		  , dev->serial.GetDisplayText()
		);
		ret = SelectConsole(TEXT("YN"));
		if ( ret != TEXT('Y') )
			return -1;
	}
#endif
	devdata.dev_model       = dev->model;
	devdata.dev_serial      = dev->serial;
	devdata.dev_hostname    = dev->hostname;
	devdata.dev_groupname   = dev->groupname;
	devdata.dev_smt_tblname.SetText(TEXT(""));
	devdata.dev_f_disable   = FALSE;
	devdata.dev_blockSize   = dev->blockSize;
	devdata.dev_blockNumber = dev->blockNumber;
	sdb::GetSystemTime(devdata.dev_ts_begin);
	devdata.dev_ts_last     = devdata.dev_ts_begin;
	devdata.dev_user_stt    = 0;
	devdata.dev_user_id     = 0;
	devdata.dev_user_pass   = dev->user_pass.GetLength() > 0 ? dev->user_pass : TEXT("");
	devdata.dev_user_ts     = devdata.dev_ts_begin;
	devdata.dev_memo        = dev->memo;

	switch ( this->opt.postmode )
	{
	case POSTMODE_DB:
		if ( dev->dbdev->AddDevice(fdata, devdata) < 0 )
		{
			_ftprintf(stderr, TEXT("Add device error!\n"));
			return -1;
		}
		fdata.Clear();
		break;
	case POSTMODE_HTTP:
	{
		int		ret;

		this->sdbhttp->recvdata.ClearText();
		ret = this->sdbhttp->PostAddDevice(devdata);
		if ( this->sdbhttp->recvdata.GetLength() > 0 )
			_fputts(this->sdbhttp->recvdata.GetText(), stdout);
		if ( ret < 0 || this->sdbhttp->rescode < 200 || this->sdbhttp->rescode >= 300 )
		{
			_ftprintf(stdout, TEXT("Add device error!\n"));
			if ( ret >= 0 )
				_ftprintf(stdout, TEXT("http result code %d\n"), this->sdbhttp->rescode);
			return -1;
		}
		break;
	}
	}
	_ftprintf(stdout, TEXT("Add device success\n"));
	return 0;
}

//////////////////////////////////////////////////////////////////////////
//	Remove Device

int		SdbCtl::RemoveDevice(SdbDev * dev)
{
	SqlLib::FieldsTbl	fdata;
	sdb::DBDevData		devdata;

	_ftprintf(stdout, TEXT("Remove device (%s %s)\n"), dev->model.GetDisplayText(), dev->serial.GetDisplayText());
	_ASSERT(dev->dbdev != NULL);
	if ( dev->dbdev->SearchDevice(fdata, dev->model.GetText(), dev->serial.GetText()) <= 0  )
	{
		_ftprintf(stderr, TEXT("Device not found!\n") );
		return -1;
	}
#if	defined(WIN32)
	{
		TCHAR	msg[1024];
		int		res;

		_stprintf(msg, TEXT("̃foCXf[^x[X폜܂B\r\n  Model: %s\r\n  Serial: %s")
		  , dev->model.GetDisplayText()
		  , dev->serial.GetDisplayText()
		);
		res = ::MessageBox(NULL, msg, TEXT("mF"), MB_YESNO);
		if ( res != IDYES )
			return -1;
	}
#else
	{
		int		ret;

		_ftprintf(stderr, TEXT("Remove database (Y/N) ?\n  Model: %s\n  Serial: %s\n")
		  , dev->model.GetDisplayText()
		  , dev->serial.GetDisplayText()
		);
		ret = SelectConsole(TEXT("YN"));
		if ( ret != TEXT('Y') )
			return -1;
	}
#endif
	dev->dbdev->GetDevData(devdata, fdata);
	if ( dev->RemoveSmartDevice(devdata) < 0 )
	{
		_ftprintf(stderr, TEXT("Remove device error!\n"));
		return -1;
	}
	_ftprintf(stdout, TEXT("Remove device success\n"));

	return 0;
}

//////////////////////////////////////////////////////////////////////////
//	Print SMART information

void		SdbCtl::PrintSmart(SdbDev * dev)
{
	int					i;
	TCHAR				t1[80];
	sdb::SmartVendor	vendor = sdb::SMART_VENDOR_GENERAL;

	if ( dev == NULL )
		return;
	if ( dev->smart.val == NULL )
		return;
	PutPos(4     , TEXT("ID# S.M.A.R.T. Attribute"),
	       4 + 4+sdb::SMART_ATTRIBUTE_MAX_LENGHTH+1, TEXT("Val Wst Ths Flags"),
	       COLUMN_MAX - 1 - 9                      , TEXT("RAW Value"),
	       0);
	for ( i = 0; i < DevCtl::NUMBER_ATA_SMART_ATTRIBUTES; i++ )
	{
		DevCtl::AtaSmartAttribute			*cur;
		DevCtl::AtaSmartThresholdAttribute	*thr;
		TCHAR								att_name[80], att_msg[80], raw_msg[80], stt_msg[80];
		sdb::SmartAttrItem					*attr_item;
		sdb::SmartAttrRawType				rawtype = sdb::SMART_RAW_NORMAL;

		cur = NULL;
		thr = NULL;
		if ( dev->smart.val->device_attribute[i].id == 0 )
			continue;	// not use
		cur = &dev->smart.val->device_attribute[i];
		attr_item = dev->smart.attrList.GetSmartAttrItem(cur->id);
		if ( dev->smart.thr != NULL )
		{
			int			j;

			// search threshhold
			for ( j = 0; j < DevCtl::NUMBER_ATA_SMART_ATTRIBUTES; j++ )
			{
				if ( dev->smart.thr->threshhold_attribute[j].id == cur->id )
				{
					thr = &dev->smart.thr->threshhold_attribute[j];
					break;
				}
			}
		}
		if ( attr_item != NULL && attr_item->text.GetLength() > 0 )
		{
			_tcsncpy(att_name, attr_item->text.GetDisplayText(), sizeof(att_name)/sizeof(*att_name) - 1);
		} else
		{
			_tcsncpy(att_name, sdb::GetSmartAttributeIdName(vendor, cur->id), sizeof(att_name)/sizeof(*att_name) - 1);
		}

		sdb::Fmt(att_msg, TEXT("%3u %3u %3s %s")
		           , cur->value
		           , cur->worst
		           , thr != NULL ? sdb::Fmt(t1,TEXT("%3u"),thr->threshold) : TEXT("???")
		           , SdbDev::GetSmartFlagsText(stt_msg, cur->flags)
		);

		dev->smart.attrList.GetSmartRawValue(*cur, sizeof(raw_msg)/sizeof(*raw_msg), raw_msg);

		PutPos(4     , sdb::Fmt(t1, TEXT("%3u %s"), cur->id, att_name),
		       4 + 4+sdb::SMART_ATTRIBUTE_MAX_LENGHTH+1, att_msg,
		       COLUMN_MAX - 1 - _tcslen(raw_msg)       , raw_msg,
		       0);
	}
	return;
}

//////////////////////////////////////////////////////////////////////////
//	Update SMART data

int		SdbCtl::SmartUpdate(SdbDev * dev)
{
	SqlLib				*sqllib = NULL;
	sdb::DBDevData		devdata;
	BOOL				update_file = FALSE;
	int					ret = 0;
	SqlLib::TimeStamp	ts;

	if ( dev == NULL )
		return -1;
	switch ( this->opt.postmode )
	{
	case POSTMODE_DB:
		if ( this->sdblib->DBIsConnect() != FALSE )
			sqllib = dev->dbdev->GetSqlLib();
		else
			update_file = TRUE;
		break;
	case POSTMODE_HTTP:
		_ASSERT(this->sdbhttp != NULL );
		DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"  host=%s, port=%s, path=%s\n"
		  , this->sdbhttp->host.GetDisplayText()
		  , this->sdbhttp->port.GetDisplayText()
		  , this->sdbhttp->path.GetDisplayText() ));
		if ( this->sdbhttp->host.GetLength() == 0 
		  || this->sdbhttp->port.GetLength() == 0 
		  || this->sdbhttp->path.GetLength() == 0 )
		{
			_ftprintf(stderr, TEXT("HTTP error!\n"));
			return -1;
		}
		break;
	default:
		_ASSERT(0);
		break;
	}
	if ( dev->smart.identify == NULL || dev->smart.val == NULL || dev->smart.thr == NULL )
	{
		if ( dev->GetSmart() < 0 || dev->GetDeviceData() < 0 )
		{
			_ftprintf(stderr, TEXT("S.M.A.R.T. error!\n"));
			return -1;
		}
	}
	if ( dev->smart.identify == NULL || dev->smart.val == NULL || dev->smart.thr == NULL )
		return -1;
	sdb::GetSystemTime(ts);
//	sqllib = NULL; update_file = TRUE;	// for DEBUG
	if ( sqllib != NULL )
		sqllib->Begin();

	{
		SqlLib::FieldsTbl	dev_fdata;
		SqlLib::FieldsTbl	smt_fdata;
		sdb::DBSmtData		smtdata;

		_ASSERT(dev->dbdev != NULL);
		if ( sqllib != NULL )
		{
			if ( dev->dbdev->SearchDevice(dev_fdata, dev->model.GetText(), dev->serial.GetText()) < 1  )
			{
				_ftprintf(stderr, TEXT("Device not found! (%s %s)\n"), dev->model.GetDisplayText(), dev->serial.GetDisplayText() );
				ret = -1;
			}
			dev->dbdev->GetDevData(devdata, dev_fdata);
			DBGREPORT((_CRT_WARN,NULL,NULL,NULL,"  dev_id=%d\n", devdata.dev_id));
		} else
		{
			devdata.dev_id     = 0;
			devdata.dev_model  = dev->model.GetText();
			devdata.dev_serial = dev->serial.GetText();
			if ( dev->user_pass.GetLength() > 0 )
				devdata.dev_user_pass = dev->user_pass;
		}
		if ( this->opt.postmode == POSTMODE_HTTP )
		{
			this->sdbhttp->SetDevData(devdata);
		}

		smtdata.smt_ts = ts;
		smtdata.smt_dev_id = devdata.dev_id;
		if ( ret == 0 && update_file == FALSE )
		{
			int			i, j;

			for ( i = 0; i < DevCtl::NUMBER_ATA_SMART_ATTRIBUTES; i++ )
			{
				if ( dev->smart.val->device_attribute[i].id == 0 )
					continue;	// not use
				smtdata.smt_attr_id   = dev->smart.val->device_attribute[i].id; 
				smtdata.smt_value     = dev->smart.val->device_attribute[i].value;
				smtdata.smt_worst     = dev->smart.val->device_attribute[i].worst;
				smtdata.smt_threshold = dev->GetThreshold(smtdata.smt_attr_id);
				smtdata.smt_flags     = sdb::ATA_GET16(dev->smart.val->device_attribute[i].flags);
				for ( j = 0; j < sizeof(smtdata.smt_raw_data)/sizeof(*smtdata.smt_raw_data); j++ )
					smtdata.smt_raw_data[j] = dev->smart.val->device_attribute[i].raw[j];
				smtdata.smt_raw_value = dev->smart.attrList.GetSmartRawValue(dev->smart.val->device_attribute[i]);

				if ( this->opt.postmode == POSTMODE_HTTP )
				{
					this->sdbhttp->AddSmtData(smtdata);
				} else if ( sqllib != NULL )
				{
					_ASSERT(dev->dbsmt != NULL);
					if ( dev->dbsmt->SetSmtData(smt_fdata, smtdata) < 0 )
					{	// error
						ret = -1;
						break;
					}
					if ( dev->SmartUpdate(devdata, smtdata) < 0 )
					{	// error
						ret = -1;
						break;
					}
				}
			} // end of for(i;;)
		}
	}
	if ( ret < 0 )
	{
		_ftprintf(stderr, TEXT("Update error!\n"));
		if ( sqllib != NULL )
			sqllib->Rollback();
		return ret;
	}
	if ( this->opt.postmode == POSTMODE_HTTP )
	{
		if ( this->sdbhttp->PostUpdate() < 0 )
			update_file = TRUE;
		if ( this->sdbhttp->recvdata.GetLength() > 0 )
			_fputts(this->sdbhttp->recvdata.GetText(), stdout);
	}
	if ( sqllib != NULL )
		sqllib->Commit();
	if ( update_file )
	{
		SQStr	path;

		dev->SmartUpdateFile(devdata, sdblib->GetSubPath(path));
	}

	return 0;
}

//////////////////////////////////////////////////////////////////////////
//	LoadCsv file

int		SdbCtl::LoadCsv(SdbDev * target_dev)
{
	SqlLib				*sqllib = NULL;
	sdb::DBDevDataList	devdata_list;
	SQPath				path;
	sdb::IOBuffer		iobuf;
	enum	DataOrder
	{
	N_TimeStamp,
	N_Model    ,
	N_Serial   ,
	N_Attr_ID  ,
	N_Value    ,
	N_Worst    ,
	N_Threshold,
	N_Flags    ,
	N_RAW_Value,
	N_RAW_Data ,
	N_Num
	};
	int					retcode = 0;

	if ( this->opt.csvfile.GetLength() > 0 )
	{
		path.SetText(this->opt.csvfile);
	} else
	{
		this->sdblib->GetSubPath(path);
	}
	if ( path.GetLength() == 0 )
	{
		_ftprintf(stderr, TEXT("CSV file error!\n"));
		return -1;
	}

	_ASSERT(this->sdblib != NULL);
	sqllib = this->sdblib->sqllib;
	_ASSERT(sqllib != NULL);
	_ASSERT(this->sdblib->conn_info.dbname != NULL);
	_ftprintf(stdout, TEXT("Laod csv file %s\n"), path.GetDisplayText());
	if ( this->sdblib->DBIsConnect() == FALSE )
	{
		_ftprintf(stderr, TEXT("Offline database!\n"));
		return -1;
	}

	if ( iobuf.OpenFile(path.GetText(), sdb::FOM_READ) < 0 )
	{
		_ftprintf(stderr, TEXT("CSV file open error! (%s)\n"), path.GetText());
		return -1;
	}

	sqllib->Begin();
	try
	{
		SQStr				dev_model, dev_serial;
		int					ln = 0;

		while ( iobuf.IsEnd() == FALSE )
		{
			VarBuffer		lbuf;
			sdb::DBSmtData	smtdata;
			int				data_num, len;
			const TCHAR		*top;
			VarBuffer		vbuf;

			lbuf.Clear();
			if ( iobuf.GetLine(lbuf) < 0 )
				break;
			ln++;
		//	_ftprintf(stdout, TEXT("%s\n"), lbuf.GetText());	// DEBUG
			if ( lbuf.GetLength() == 0 )
				continue;
			lbuf.SetBuffer();
			lbuf.SkipSpace();
			top = lbuf.GetCurPtr();
			len = 0;
			vbuf.Clear();
			data_num = 0;
			for (;;)
			{
				int			ch;

				ch = lbuf.GetChar();
				if ( ch == TEXT(',') || ch < 0 || ch == TEXT('\0') || ch == TEXT('\n') || lbuf.IsEnd() )
				{
					const 				TCHAR	*s;
					SqlLib::VarBuffer	tmpbuf;

					vbuf.AddChar(TEXT('\0'));
					vbuf.SetBuffer();
					s = vbuf.GetText();
					if ( *s == TEXT('\'') )
					{
						tmpbuf.Clear();
						SqlLib::DeEscapeString(tmpbuf, s, TEXT('\''));
						tmpbuf.AddChar(TEXT('\0'));
						s = tmpbuf.GetText();
					}
					{
					//	if ( data_num > 0 )	// DEBUG
					//		_ftprintf(stdout, TEXT(","));
					//	_ftprintf(stdout, TEXT("%d:%s"), data_num, s );
						switch ( data_num )
						{
						case N_TimeStamp:	SqlLib::GetTimeStamp(smtdata.smt_ts, s);	break;
						case N_Model    :	dev_model.SetText(s);	break;
						case N_Serial   :	dev_serial.SetText(s);	break;
						case N_Attr_ID  :	smtdata.smt_attr_id   = _ttoi(s);	break;
						case N_Value    :	smtdata.smt_value     = _ttoi(s);	break;
						case N_Worst    :	smtdata.smt_worst     = _ttoi(s);	break;
						case N_Threshold:	smtdata.smt_threshold = _ttoi(s);	break;
						case N_Flags    :	smtdata.smt_flags     = _ttoi(s);	break;
						case N_RAW_Value:	smtdata.smt_raw_value = SqlLib::atoi64(s);	break;
						case N_RAW_Data :	smtdata.smt_raw_value = SqlLib::atoi64(s);	break;
						}
						data_num++;
					}
					if ( ch < 0 || ch == TEXT('\0') || ch == TEXT('\n') || lbuf.IsEnd() )
						break;
					lbuf.SkipSpace();
					top = lbuf.GetCurPtr();
					len = 0;
					vbuf.Clear();
				} else if ( ch == TEXT('\'') )
				{
					len++;
					vbuf.AddChar(ch);
					while ( lbuf.IsEnd() == FALSE )
					{
						ch = lbuf.GetChar();
						len++;
						vbuf.AddChar(ch);
						if ( ch == TEXT('\\') && lbuf.IsEnd() == FALSE )
						{
							lbuf.GetChar();
							len++;
							vbuf.AddChar(ch);
						} else if ( ch == TEXT('\'') )
							break;
					}
				} else
				{
					len++;
					vbuf.AddChar(ch);
				}
			} // end of for(;;)
		//	_ftprintf(stdout, TEXT("\n"));	// DEBUG
			if ( data_num < N_Num )
				continue;	// next line
			{
				SdbDev						*sdbdev;
				SdbLib::Dev_list::iterator	it;

				if ( smtdata.smt_attr_id <= 0  || smtdata.smt_attr_id   > 255
				  || smtdata.smt_value < 0     || smtdata.smt_value     > 255 
				  || smtdata.smt_worst < 0     || smtdata.smt_worst     > 255 
				  || smtdata.smt_threshold < 0 || smtdata.smt_threshold > 255 
				)
				{	// data error
					_ftprintf(stderr, TEXT("S.M.A.R.T. data error! (line no. %d)\n"), ln);
					throw -1;
				}
				if ( target_dev == NULL )
				{	// Oɓo^Ă邷ׂẴfoCX
					BOOL		found_dev;

					found_dev = FALSE;
					for ( it = this->sdblib->dev_list.begin(); it != this->sdblib->dev_list.end(); it++)
					{
						sdbdev = *it;
						if ( SQStr::StrCmp(sdbdev->model.GetText() , dev_model.GetText()) == 0 
						  && SQStr::StrCmp(sdbdev->serial.GetText(), dev_serial.GetText()) == 0 )
						{
							found_dev = TRUE;
							break;
						}
					}
					if ( found_dev == FALSE )
					{
						if ( (sdbdev = this->sdblib->CreateSdbDev(TEXT("."))) == NULL )
						{
							_ftprintf(stderr, TEXT("CreateSDbDev: Internal error! (line no. %d)\n"), ln);
							throw -1;
						}
						sdbdev->model.SetText(dev_model);
						sdbdev->serial.SetText(dev_serial);
						this->sdblib->dev_list.push_back(sdbdev);
					}
				} else
				{	// p[^Ɏw肳ꂽfoCX̂
					sdbdev = target_dev;
					if ( SQStr::StrCmp(sdbdev->model.GetText() , dev_model.GetText()) == 0 
					  && SQStr::StrCmp(sdbdev->serial.GetText(), dev_serial.GetText()) == 0 )
					{	// found
					} else
					{
						continue;	// next line
					}
				}
				_ASSERT(sdbdev != NULL);
				{
					sdb::DBDevDataList::iterator	it;
					sdb::DBDevData					devdata, *pdev;
					BOOL							push_back;

					pdev = NULL;
					push_back = FALSE;
					for ( it = devdata_list.begin(); it != devdata_list.end(); it++ )
					{
						if ( SQStr::StrCmp(sdbdev->model.GetText() , (*it).dev_model.GetText()) == 0 
						  && SQStr::StrCmp(sdbdev->serial.GetText(), (*it).dev_serial.GetText()) == 0 )
						{	// found
							pdev = &(*it);
							break;
						}
					}
					if ( pdev == NULL )
					{
						if ( this->sdblib->dbdev.SearchDevice(devdata, sdbdev->model.GetText(), sdbdev->serial.GetText()) > 0 )
						{	// found
						//	_ftprintf(stdout, TEXT("DEBUG: Found dev_id=%d\n") , devdata.dev_id );
							pdev = &devdata;
							push_back = TRUE;
						//	sdbdev->dev_id = devdata.dev_id;
						} else
						{
							_ftprintf(stdout, TEXT("Device not found! (mode %s, serial %s)\n")
							  , sdbdev->model.GetDisplayText()
							  , sdbdev->serial.GetDisplayText() );
							throw -1;
						}
					}
					_ASSERT(pdev != NULL);
					_ftprintf(stderr, TEXT("%5u: %4u-%02u-%02u %02u:%02u:%02u %s %s %3u %3u %3u %3u %04X %"FI64"u %"FI64"u\n")
					  , ln
					  , smtdata.smt_ts.year, smtdata.smt_ts.month , smtdata.smt_ts.day
					  , smtdata.smt_ts.hour, smtdata.smt_ts.minute, smtdata.smt_ts.second
					  , dev_model.GetDisplayText()
					  , dev_serial.GetDisplayText()
					  , smtdata.smt_attr_id
					  , smtdata.smt_value
					  , smtdata.smt_worst
					  , smtdata.smt_threshold
					  , smtdata.smt_flags
					  , (sdb::Int64)smtdata.smt_raw_value
					  , (sdb::Int64)smtdata.smt_raw_data
					);
					if ( sdbdev->SmartUpdate(*pdev, smtdata) < 0 )
					{
						_ftprintf(stderr, TEXT("S.M.A.R.T. update error!\n"));
						throw -1;
					}
					if ( push_back )
						devdata_list.push_back(*pdev);
				}
			}
		} // end of while
		retcode = 0;
	} catch(int ret) { retcode = ret; }
	if ( retcode == 0 )
		sqllib->Commit();
	else
		sqllib->Rollback();
	iobuf.CloseFile();

	return retcode;
}

//////////////////////////////////////////////////////////////////////////
//	Watch Smart Database

int		SdbCtl::WatchSmartDatabase(void)
{
	SqlLib				*sqllib = NULL;
	SqlLib::VarBuffer	sql;
	SqlLib::Result		sqlres;
	int					xmax, ymax;
	sdb::DBDevDataList	devdataList;


	_ASSERT(this->sdblib != NULL);
	if ( this->sdblib->DBIsConnect() == FALSE )
	{
		_ftprintf(stderr, TEXT("Offline database!\n"));
		return -1;
	}
	sqllib = this->sdblib->dbdev.GetSqlLib();
	_ASSERT(sqllib != NULL);

	sql.AddText(TEXT("SELECT * FROM "));
	sql.AddText(this->sdblib->dbdev.tblname.GetText());
	sql.AddText(TEXT(" WHERE "));
	sql.AddText(sdb::dev_tbl[sdb::DEV_F_DISABLE].fname );
	sql.AddText(TEXT("<>true AND "));
	sql.AddText(sdb::dev_tbl[sdb::DEV_SMT_TBLNAME].fname );
	sql.AddText(TEXT("<>''"));
	sql.AddText(TEXT(" ORDER BY "));
	sql.AddText(sdb::dev_tbl[sdb::DEV_GROUPNAME].fname );
	sql.AddText(TEXT(","));
	sql.AddText(sdb::dev_tbl[sdb::DEV_HOSTNAME].fname );
	sql.AddText(TEXT(","));
	sql.AddText(sdb::dev_tbl[sdb::DEV_MODEL].fname );
	sql.AddChar(0);

	if ( this->sdblib->dbdev.SelectData(sqlres, sql.GetText()) < 0 )
	{
		return -1;
	}
	ymax = sqllib->GetTuples(sqlres);
	xmax = sqllib->GetFields(sqlres);
	if ( xmax > 0 && ymax > 0 )
	{
		int		y;

		for ( y = 0; y < ymax; y++ )
		{
			sdb::DBDevData		devdata;
			SqlLib::FieldsTbl	fdata;
			int					x;

			fdata.Clear();
			for ( x = 0; x < xmax; x++ )
			{
				const TCHAR			*fname;
				SqlLib::DataValue	val;

				if ( (fname  = sqllib->GetFname(sqlres, x)) != NULL )
				{
					if ( sqllib->GetDataValue(sqlres, y, x, val) >= 0 )
						fdata.Add(fname, val);
				}
			}
			this->sdblib->dbdev.GetDevData(devdata, fdata);
			devdataList.push_back(devdata);
		}
	}
	sqllib->ClearResult(sqlres);

	{
		sdb::DBDevDataList::iterator	it;

		for ( it = devdataList.begin(); it != devdataList.end(); it++ )
		{
			sdb::DBDevData	devdata;

			devdata = *it;
			_ftprintf(stdout, TEXT("%s %s\n"), devdata.dev_model.GetDisplayText(), devdata.dev_serial.GetDisplayText() );
		}
	}

	return 0;
}

void	SdbCtl::PutPos(int pos, ... ) // static
{
	va_list	arg;
	TCHAR *	text;
	int		x = 0, len;

	va_start(arg, pos);
	while ( pos > 0 )
	{
		text = va_arg(arg, char*);
		len = (int)_tcslen(text);
		if ( pos > x )
		{
			_ftprintf(stdout, TEXT("%*s"), pos - x, "");
			x = pos;
		}
		_ftprintf(stdout, TEXT("%s"), text);
		x += len;
		pos  = va_arg(arg, int);
	}
	if ( pos == 0 )
		_ftprintf(stdout,TEXT("\n"));
	va_end(arg);
}

#if	defined(LINUX) && defined(DEBUG)
int		DbgReport(int type, const char *filename, int linenumber, const char *moduleName, const char *format, ... )
{
	char	buf[4096];
	va_list	arg;

	va_start(arg, format);
	vsprintf(buf, format, arg);
	va_end(arg);
	fputs(buf, stderr);
	return 0;
}
#endif

#if	defined(LINUX)
int		InputConsole(void)
{
//	while ( kbhit() != 0 )
//			gettch();
	return (int) getchar();
}

int		SelectConsole(const TCHAR * keys)
{
	int			ch;
	const TCHAR	*s;

	do
	{
		ch = InputConsole();
		ch = toupper(ch);
		for ( s = keys; *s != TEXT('\0'); s++ )
		{
			if ( toupper(*s) == ch )
				break;
		}
	} while ( *s == TEXT('\0') );
//	putchar(ch);
//	putchar(TEXT('\n'));
	return ch;
}
#endif

