//////////////////////////////////////////////////////////////////////////
//	"sdb_scr.cpp"
//------------------------------------------------------------------------
//	$Id: sdb_scr.cpp 100 2006-04-19 18:26:45Z taro $
//////////////////////////////////////////////////////////////////////////

#include	"../include/sdblib.h"
#include	"../lib/scrlib.h"
#if	defined(WIN32)
# include	<crtdbg.h>
#endif

#if	defined(_DEBUG) && defined(WIN32)
#	define	DBGREPORT(param_)	_CrtDbgReport param_
#else
#	define	DBGREPORT(param_)
# ifndef	_ASSERT
#	define	_ASSERT
# endif
#endif

struct MembInfo
{
	TCHAR		   *name;
	unsigned		type;
	union
	{
		void	   *data;
		TCHAR	 **pstr;
		int		   *pint;
	};
	BOOL			set;
};
static int	get_struct_member(ScrLib::ScValue & scValue, unsigned n, MembInfo * tbl);

int		SdbLib::LoadScr(const TCHAR * path)
{
	int		ret = 0;

	_ASSERT(this->scrlib != NULL);
	this->scrlib->Init();
	if ( this->scrlib->OpenFile(path) < 0 )
		return -1;
	ret = this->scrlib->Parse();
	this->scrlib->CloseFile();
	if ( ret < 0 )
		return -1;

	return this->InitScr();
}

int		SdbLib::InitScr(void)
{
	BOOL				progfile_set = FALSE;
	BOOL				fileinfo_set = FALSE;

	_ASSERT(this->scrlib != NULL);

	// ݒl
	{
		static MembInfo tbl[] =
		{
		{ TEXT("sdb_db_host"                     ), ScrLib::T_STRING, &sdb::db_host                       },
		{ TEXT("sdb_db_port"                     ), ScrLib::T_STRING, &sdb::db_port                       },
		{ TEXT("sdb_db_user"                     ), ScrLib::T_STRING, &sdb::db_user                       },
		{ TEXT("sdb_db_pass"                     ), ScrLib::T_STRING, &sdb::db_pass                       },
		{ TEXT("sdb_db_name"                     ), ScrLib::T_STRING, &sdb::db_name                       },

		{ TEXT("sdb_sub_dir"                     ), ScrLib::T_STRING, &sdb::sub_dir                       },

		{ TEXT("sdb_http_post_host"              ), ScrLib::T_STRING, &sdb::http_post_host                },
		{ TEXT("sdb_http_post_port"              ), ScrLib::T_STRING, &sdb::http_post_port                },
		{ TEXT("sdb_http_post_path"              ), ScrLib::T_STRING, &sdb::http_post_path                },
		{ TEXT("sdb_http_auth_type"              ), ScrLib::T_STRING, &sdb::http_auth_type                },
		{ TEXT("sdb_http_auth_user"              ), ScrLib::T_STRING, &sdb::http_auth_user                },
		{ TEXT("sdb_http_auth_pass"              ), ScrLib::T_STRING, &sdb::http_auth_pass                },

		{ TEXT("sdb_mail_warnlevel"              ), ScrLib::T_INT   , &sdb::mail_warnlevel                },
		{ TEXT("sdb_mail_from"                   ), ScrLib::T_STRING, &sdb::mail_from                     },
		{ TEXT("sdb_mail_to"                     ), ScrLib::T_STRING, &sdb::mail_to                       },
		{ TEXT("sdb_mail_smtp_server"            ), ScrLib::T_STRING, &sdb::mail_smtp_server              },

		{ TEXT("sdb_warn_high_estimate"          ), ScrLib::T_INT   , &sdb::warn_high_estimate            },
		{ TEXT("sdb_warn_mid_estimate"           ), ScrLib::T_INT   , &sdb::warn_mid_estimate             },
		{ TEXT("sdb_warn_low_estimate"           ), ScrLib::T_INT   , &sdb::warn_low_estimate             },
		{ TEXT("sdb_watch_offline_margin_sec"    ), ScrLib::T_INT   , &sdb::watch_offline_margin_sec      },

		{ TEXT("sdb_auto_raw_word_id"            ), ScrLib::T_INT   , &sdb::auto_raw_word_id              },
		{ TEXT("sdb_random_wait_sec"             ), ScrLib::T_INT   , &sdb::random_wait_sec               },
		{ TEXT("sdb_dialog_auto_close_sec"       ), ScrLib::T_INT   , &sdb::dialog_auto_close_sec         },
		};
		int					i;

		for ( i = 0; i < sizeof(tbl)/sizeof(*tbl); i++)
		{
			SQStr					name;
			ScrLib::ScValue		   *scValue;

			name.SetText(tbl[i].name);
			if ( (scValue = this->scrlib->GetValue(name)) != NULL && (scValue->type & (ScrLib::T_NAMEMASK)) == tbl[i].type )
			{
				tbl[i].set = TRUE;
				switch ( tbl[i].type )
				{
					case ScrLib::T_STRING:
						*(tbl[i].pstr) = scValue->val.str->GetText();
						break;
					case ScrLib::T_INT:
						*(tbl[i].pint) = scValue->val.si;
						break;
				}
			}
		}
	}
	if ( sdb::db_host != NULL ) this->dbinfo.host.SetText(sdb::db_host);
	if ( sdb::db_port != NULL ) this->dbinfo.port.SetText(sdb::db_port);
	if ( sdb::db_user != NULL ) this->dbinfo.user.SetText(sdb::db_user);
	if ( sdb::db_pass != NULL ) this->dbinfo.pass.SetText(sdb::db_pass);
	if ( sdb::db_name != NULL ) this->dbinfo.dbname.SetText(sdb::db_name);

	this->scr_GetDevOpt();

	return 0;
}

static int		get_struct_member(ScrLib::ScValue & scValue, unsigned n, MembInfo * tbl)
{
	unsigned			i;
	ScrLib::ScValue	   *memb;

	memb = NULL;
	for ( i = 0; i < n; i++)
		tbl[i].set = FALSE;
	for ( i = 0; i < n; i++)
	{
		SQStr	membname;

		membname.SetText(tbl[i].name);
		if ( (memb = scValue.GetMemb(membname)) != NULL && (memb->type & (ScrLib::T_NAMEMASK)) == tbl[i].type )
		{
			tbl[i].set = TRUE;
			switch ( tbl[i].type )
			{
				case ScrLib::T_STRING:
					*(tbl[i].pstr) = memb->val.str->GetText();
					break;
				case ScrLib::T_INT:
					*(tbl[i].pint) = memb->val.si;
					break;
			}
		}
	}
	for ( i = 0; i < n; i++)
	{
		if ( tbl[i].set == FALSE )
			break;
	}
	return i == n ? 0 : -1;
}

#define	SDBDEFOPTTBL_NAME	TEXT("sdbDevOptTbl")
#define	SDBDEFOPT_TYPE		TEXT("SdbDevOpt")

int			SdbLib::scr_GetDevOpt(void)
{
	ScrLib::ScValue		   *scArray, *scValue;
	SQStr					name;

	name.SetText(SDBDEFOPTTBL_NAME);
	if ( (scArray = this->scrlib->GetValue(name)) != NULL
	  && scArray->IsArray() && scArray->IsStruct() )
	{
		int			a;
		TCHAR	   *memb_model, *memb_serial, *memb_option;
		MembInfo	membtbl[] =
		{
			{ TEXT("model"    ), ScrLib::T_STRING, &memb_model    },
			{ TEXT("serial"   ), ScrLib::T_STRING, &memb_serial   },
			{ TEXT("option"   ), ScrLib::T_STRING, &memb_option   },
		};
		SQStr		text, url;

		a = 0;
		for ( a = 0; (scValue = scArray->GetArrayValue(a)) != NULL; a++ )
		{
			sdb::DeviceOption		*devopt;

			if ( get_struct_member(*scValue, sizeof(membtbl)/sizeof(*membtbl), membtbl) < 0 )
				continue;
			if ( (devopt = new sdb::DeviceOption) == NULL )
				return -1;	// error
			devopt->model.SetText(memb_model);
			devopt->serial.SetText(memb_serial);
			devopt->option.SetText(memb_option);

			this->devoptList.push_back(devopt);
		}
	}

	return 0;
}

