//////////////////////////////////////////////////////////////////////////
//	"sdbdev.h"
//------------------------------------------------------------------------
//	$Id: sdbdev.cpp 150 2006-04-30 23:03:17Z taro $
//////////////////////////////////////////////////////////////////////////

#include	"../include/sdblib.h"
#include	"../lib/getopt.h"
#if	defined(WIN32)
# include	"../win32/devctl.h"
# include	<sqarg.h>
# include	<crtdbg.h>
#else
# include	"../lib/sqarg.h"
# include	<ctype.h>
#endif

#if	defined(_DEBUG) && defined(WIN32)
#	define	DBGREPORT(param_)	_CrtDbgReport param_
#else
#	define	DBGREPORT(param_)
# ifndef	_ASSERT
#	define	_ASSERT
# endif
#endif

SdbDev::SdbDev()
{
	this->dbdev = NULL;
	this->dbsmt = new sdb::DBSmtLib;
	_ASSERT(this->dbsmt != NULL);

#if	defined(SDB_USE_DEVCTL)
	this->devCtl = new DevCtl;
	_ASSERT(this->devCtl != NULL);
#endif

	this->smart.identify = NULL;
	this->smart.val = NULL;
	this->smart.thr = NULL;

	this->getopt = NULL;
//	this->dev_id = -1;
	this->blockSize = 0;
	this->blockNumber = I64(0);
	this->user_id = 0;

	return;
}

SdbDev::~SdbDev()
{
	if ( this->getopt != NULL )
		delete this->getopt;

	if ( this->smart.identify != NULL )
		delete this->smart.identify;
	if ( this->smart.val != NULL )
		delete this->smart.val;
	if ( this->smart.thr != NULL )
		delete this->smart.thr;

#if	defined(SDB_USE_DEVCTL)
	if ( this->devCtl != NULL )
		delete this->devCtl;
#endif

	if ( this->dbsmt != NULL )
		delete this->dbsmt;
}

int		SdbDev::GetDeviceName(SQStr & devname, const TCHAR * desc)	// static
{
	TCHAR	tmp[80];

	if ( desc == NULL )
		return -1;
	if ( _tcsnicmp(desc, TEXT("disk"), 4) == 0 && _istdigit(desc[4]) )
	{
		_stprintf(tmp, TEXT("\\\\.\\PhysicalDrive%d"), (int)_ttoi(&desc[4]));
		devname.SetText(tmp);
		return 0;
	}
	return -1;
}

#if	defined(SDB_USE_DEVCTL)
int		SdbDev::OpenDevice(unsigned rwmode)
{
	SQStr	devname;

	if ( SdbDev::GetDeviceName(devname, this->desc.GetText()) < 0 )
		devname = this->desc;

	return this->devCtl->OpenDevice(devname.GetText(), rwmode);
}

int		SdbDev::CloseDevice(void)
{
	return this->devCtl->CloseDevice();
}

int		SdbDev::GetSmart(void)
{
	int		drive = 0;

	DBGREPORT((_CRT_WARN,NULL,NULL,NULL,TEXT("SdbDev::GetSmart()\n")));
	if ( SdbDev::IsHandle(this->devCtl->GetHandle()) == FALSE )
		return -1;

	{
		if ( this->smart.identify == NULL )
		{
			if ( (this->smart.identify = new DevCtl::AtaIdentifyDevice) == NULL )
				return -1;
		}
		memset(this->smart.identify, 0, sizeof(DevCtl::AtaIdentifyDevice));
		_ASSERT(this->devCtl != NULL);
		if ( this->devCtl->AtaReadIndentify(drive, *(this->smart.identify)) < 0 )
		{
			delete this->smart.identify;
			this->smart.identify = NULL;
		}
	}

	{
		if ( this->smart.val == NULL )
		{
			if ( (this->smart.val = new DevCtl::AtaSmartValues) == NULL )
				return -1;
		}
		memset(this->smart.val, 0, sizeof(DevCtl::AtaSmartValues));
		if ( this->devCtl->SmartReadValues(drive, *(this->smart.val)) < 0 )
		{
			delete this->smart.val;
			this->smart.val = NULL;
		}
	}

	{
		if ( this->smart.thr == NULL )
		{
			if ( (this->smart.thr = new DevCtl::AtaSmartThresholds) == NULL )
				return -1;
		}
		memset(this->smart.thr, 0, sizeof(DevCtl::AtaSmartThresholds));
		if ( this->devCtl->SmartReadThresholds(drive, *(this->smart.thr)) < 0 )
		{
			delete this->smart.thr;
			this->smart.thr = NULL;
		}
	}

	return 0;
}

int		SdbDev::GetThreshold(unsigned id, DevCtl::AtaSmartThresholds *thr) // static
{
	int		threshold = 0;

	if ( thr != NULL )
	{	// search threshhold
		int		i;

		for ( i = 0; i < DevCtl::NUMBER_ATA_SMART_ATTRIBUTES; i++ )
		{
			int		n;

			n = (id + i) % DevCtl::NUMBER_ATA_SMART_ATTRIBUTES;
			if ( id == thr->threshhold_attribute[n].id )
			{
				threshold = thr->threshhold_attribute[n].threshold;
				break;
			}
		}
	}
	return threshold;
}

int		SdbDev::GetDeviceData(void)
{
	if ( this->smart.identify == NULL )
	{
		if ( this->GetSmart() < 0 || this->smart.identify == NULL )
			return -1;
	}
	{
		TCHAR		model[40+1], firm[8+1], serial[20+1];

		FormatAtaString(model, this->smart.identify->model_number, sizeof(this->smart.identify->model_number));
		FormatAtaString(firm, this->smart.identify->micro_code_version, sizeof(this->smart.identify->micro_code_version));
		FormatAtaString(serial, this->smart.identify->serial_number, sizeof(this->smart.identify->serial_number));
		this->model.SetText(model);
		this->firm.SetText(firm);
		this->serial.SetText(serial);

		this->blockSize   = 512;
		this->blockNumber = SdbDev::GetAtaCapacity(*(this->smart.identify));
	}

	{
		sdb::DBDevData		devdata;
		sdb::DeviceOption	*devopt;
		SdbLib *			sdblib;

		devdata.dev_model  = this->model;
		devdata.dev_serial = this->serial;
		_ASSERT(this->dbdev != NULL);
		sdblib = this->dbdev->GetSdbLib();
		_ASSERT(sdblib != NULL);
		if ( (devopt = sdblib->GetDeviceOption(devdata)) != NULL && devopt->option.GetLength() > 0 )
		{
			this->ParseOption(devopt->option.GetText());
		}
	}

	return 0;
}

sdb::Uint64	SdbDev::GetAtaCapacity(DevCtl::AtaIdentifyDevice & identify) // static
{
	WORD	command_set_2  = identify.command_set083.GetData();
	WORD 	capabilities_0 = identify.capabilities049.GetData();
	WORD	sects_16       = identify.total_user_address_sectors[0].GetData();
	WORD	sects_32       = identify.total_user_address_sectors[1].GetData();
	WORD	lba_16         = identify.min_user_LBA_address_48bit[0].GetData();
	WORD	lba_32         = identify.min_user_LBA_address_48bit[1].GetData();
	WORD	lba_48         = identify.min_user_LBA_address_48bit[2].GetData();
	WORD	lba_64         = identify.min_user_LBA_address_48bit[3].GetData();
	sdb::Uint64	capacity = I64(0), capacity_short = I64(0);
  
	if ( capabilities_0 & 0x0200 )
	{
		capacity_short = (DWORD)sects_32 << 16 | 
	                     (DWORD)sects_16 << 0  ;

		if ( (command_set_2 & 0xC000) == 0x4000 && (command_set_2 & 0x0400) )
		{	// support 48bit LBA
			capacity = (sdb::Uint64)lba_64 << 48 | 
					   (sdb::Uint64)lba_48 << 32 |
					   (sdb::Uint64)lba_32 << 16 | 
					   (sdb::Uint64)lba_16 << 0  ;
		}

		if ( capacity_short > capacity )
			capacity = capacity_short;
	}

	return capacity;
}

TCHAR * SdbDev::FormatAtaString(TCHAR * out, const DevCtl::ATA_WORD * inp, int n) // static
{
	TCHAR	tmp[80];
	int		i;

	n = n > sizeof(tmp)/sizeof(*tmp) - 1 ? sizeof(tmp)/sizeof(*tmp) - 1 : n;
	for ( i = 0; i < n; i += 2 )
	{
		tmp[i  ] = inp[i/2].b[1];
		tmp[i+1] = inp[i/2].b[0];
	}
	tmp[n] = TEXT('\0');

	if ( _tcslen(tmp) > 0 )
	{
		int			top = 0, end;
		char *		s = tmp;

		for ( i = 0; tmp[i]; i++ )
		{
			if ( tmp[i] != TEXT(' ') && tmp[i] != TEXT('\t') )
			{
				top = i;
				break;
			}
		}
		i = (int)_tcslen(tmp) - 1;
		while ( i >= top )
		{
			if ( tmp[i] != TEXT(' ') && tmp[i] != TEXT('\t') )
				break;
			i--;
		}
		end = i;

		_tcsncpy(out, tmp + top, end - top + 1);
		out[end - top + 1] = TEXT('\0');
	} else
		out[0] = TEXT('\0');
	return out;
}
#endif

//	Needs: devdata.dev_id
//	Option: devdata.dev_smt_tblname
int		SdbDev::RemoveSmartDevice(sdb::DBDevData & devdata)
{
	SqlLib		*sqllib;

	_ASSERT(this->dbdev != NULL);
	if ( (sqllib = this->dbdev->GetSqlLib()) == NULL )
		return -1;
#if	defined(_DEBUG) && defined(SDBCTL)
	_ftprintf(stderr, TEXT("dev_id = %d\n"), devdata.dev_id);
	_ftprintf(stderr, TEXT("dev_smt_tblname = %s\n"), devdata.dev_smt_tblname.GetDisplayText());
#endif
	sqllib->Begin();
	if ( devdata.dev_smt_tblname.GetLength() > 0 )
	{
		_ASSERT(this->dbsmt);
		this->dbsmt->tblname.SetText(devdata.dev_smt_tblname);
		if ( this->dbsmt->DropTable() < 0 )
		{
			sqllib->Rollback();
			return -1;
		}
	}
	if ( this->dbdev->DeleteDevice(devdata) < 0 )
	{
		sqllib->Rollback();
		return -1;
	}
	sqllib->Commit();

	return 0;
}

int		SdbDev::SmartUpdate(sdb::DBDevData & devdata, sdb::DBSmtData & smtdata)
{
	SqlLib::FieldsTbl	dev_fdata, dev_update;
	SqlLib::FieldsTbl	smt_fdata;
	SdbLib *			sdblib = this->dbdev->GetSdbLib();
	SqlLib::Oid			oid;

	_ASSERT(this->dbsmt != NULL);
	if ( devdata.dev_smt_tblname.GetLength() == 0 )
	{	// e[u
		_ASSERT(sdblib != NULL);
		if ( sdblib->GenerateSmtTblName(devdata.dev_smt_tblname, devdata) < 0 )
			return -1;
		this->dbsmt->tblname.SetText(devdata.dev_smt_tblname);
		this->dbsmt->CreateTable();
		dev_update.AddVarChar(sdb::dev_tbl[sdb::DEV_SMT_TBLNAME].fname, devdata.dev_smt_tblname.GetText());
	} else
	{
		if ( SQStr::StrCmp(this->dbsmt->tblname.GetText(), devdata.dev_smt_tblname.GetText()) != 0 )
			this->dbsmt->tblname.SetText(devdata.dev_smt_tblname);
	}

	if ( this->dbsmt->SetSmtData(smt_fdata, smtdata) < 0 )
		return -1;
	if ( (oid = this->dbsmt->InsertData(smt_fdata)) < 0 )
		return -1;

	if ( sdb::CompTimeStamp(devdata.dev_ts_begin, smtdata.smt_ts) > 0 )
	{
		devdata.dev_ts_begin = smtdata.smt_ts;
		dev_update.AddTimeStamp(sdb::dev_tbl[sdb::DEV_TS_BEGIN].fname, devdata.dev_ts_begin);
	}
	if ( sdb::CompTimeStamp(devdata.dev_ts_last , smtdata.smt_ts) < 0 )
	{
		devdata.dev_ts_last = smtdata.smt_ts;
		dev_update.AddTimeStamp(sdb::dev_tbl[sdb::DEV_TS_LAST].fname, devdata.dev_ts_last);
	}

	if ( this->hostname.GetText() != NULL && SQStr::StrCmp(devdata.dev_hostname.GetText(), this->hostname.GetText()) != 0 )
	{
		devdata.dev_hostname = this->hostname;
		dev_update.AddVarChar(sdb::dev_tbl[sdb::DEV_HOSTNAME].fname, devdata.dev_hostname.GetText());
	}
	if ( this->groupname.GetText() != NULL && SQStr::StrCmp(devdata.dev_groupname.GetText(), this->groupname.GetText()) != 0 )
	{
		devdata.dev_groupname = this->groupname;
		dev_update.AddVarChar(sdb::dev_tbl[sdb::DEV_GROUPNAME].fname, devdata.dev_groupname.GetText());
	}
	if ( this->user_pass.GetText() != NULL && SQStr::StrCmp(devdata.dev_user_pass.GetText(), this->user_pass.GetText()) != 0 )
	{
		devdata.dev_user_pass = this->user_pass;
		dev_update.AddVarChar(sdb::dev_tbl[sdb::DEV_USER_PASS].fname, devdata.dev_user_pass.GetText());
	}
	if ( this->memo.GetText() != NULL && SQStr::StrCmp(devdata.dev_memo.GetText(), this->memo.GetText()) != 0 )
	{
		devdata.dev_memo = this->memo;
		dev_update.AddText(sdb::dev_tbl[sdb::DEV_MEMO].fname, devdata.dev_memo.GetText());
	}

	if ( dev_update.GetCount() > 0 )
	{	// UPDATE
		TCHAR	where[256];

		_ASSERT(devdata.dev_id > 0 );
		_stprintf(where, TEXT("%s=%d"), sdb::dev_tbl[sdb::DEV_ID].fname, devdata.dev_id);
		if ( this->dbdev->UpdateData(dev_update, where) < 0 )
			return -1;
	}

	return 0;
}


int		SdbDev::ParseOption(const TCHAR * option)
{
	SQArg							arg;
	int								argc;
	TCHAR **						argv;
	int								retcode;

	if ( this->getopt == NULL )
	{
		static GetOpt::AddItemParam	tbl[] =
		{
			{ OPT_GROUPNAME         , TEXT("groupname")     , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_HOSTNAME          , TEXT("hostname")      , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_SMART_ATTR        , TEXT("smart_attr")    , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_USER_PASS         , TEXT("user_pass")     , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
		};

		if ( (this->getopt = new GetOpt) == NULL )
			return -1;
		if ( this->getopt->AddItem(sizeof(tbl)/sizeof(*tbl), tbl) < 0 )
			return -1;
	}

	arg.setCmdline(option);
	arg.getArgs();
	argc = arg.getArgc();
	argv = arg.getArgv();

	try
	{
		int			an;

		an = 0;
		this->getopt->SetCurPos(an);
		while ( this->getopt->GetCurPos() < argc )
		{
			GetOpt::Item	*item;
			const char		*s;

			if ( (item = this->getopt->ParseArg(an, argc, (const TCHAR **)argv)) == NULL )
			{
				if ( (s = this->getopt->last_param) != NULL )
				{
				} else if ( this->getopt->err_option != NULL || this->getopt->err_param != NULL )
				{
				} // end of Arg
			} else
			{
				switch ( item->usertype )
				{
				case OPT_GROUPNAME:
					this->groupname.SetText(item->var.text);
					break;

				case OPT_HOSTNAME:
					this->hostname.SetText(item->var.text);
					break;

				case OPT_SMART_ATTR:
					this->smart.attrList.AddDef(item->var.text);
					break;

				case OPT_USER_PASS:
					this->user_pass.SetText(item->var.text);
					break;

				default:
					_ASSERT(0);
					break;
				} // end of switch
			}
		} // end of while
		retcode = 0;
	} catch(int ret) { retcode = ret; }

	return retcode;
}

TCHAR *	SdbDev::GetSmartFlagsText(TCHAR * buffer, unsigned stt) // static
{
	return sdb::Fmt(buffer, TEXT("%c%c%c%c%c%c")
	  , stt & 0x0020 ? TEXT('1') : TEXT('-')
	  , stt & 0x0010 ? TEXT('1') : TEXT('-')
	  , stt & 0x0008 ? TEXT('1') : TEXT('-')
	  , stt & 0x0004 ? TEXT('1') : TEXT('-')
	  , stt & 0x0002 ? TEXT('O') : TEXT('-')
	  , stt & 0x0001 ? TEXT('P') : TEXT('-')
	);
}
